/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * _range_gist_ops1.c
 *
 *    This file contains GiST method implementation for range array
 *    datatype. 
 *
 *    For now, we have two implementations for range array Gist index.
 *    they are gist__range_ops1, gist__range_ops2
 *    gist__range_ops1 (this file) differs from gist__range_ops2 in that
 *    its leaf nodes contain the whole range array while those in gist__range_ops2
 *    only contains the minimum bounding box of that range array. More experiments
 *    have to be done to see which is more efficient. Obviously, for storage 
 *    concerns, gist__range_ops2 uses MUCH LESS storage while gist__range_ops1
 *    MIGHT run a little bit faster.
 *
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */

#include "postgres.h"
#include "access/gist.h"
#include "access/rtree.h"
#include "utils/builtins.h"
#include "_range_gist_ops1.h"
#include "global.h"

/*
 *    We are using PG_FUNCTION VERSION 1 calling convention here because
 *    it makes further modification and debugging easier.
 */
PG_FUNCTION_INFO_V1(gist__range_ops1_consistent);
PG_FUNCTION_INFO_V1(gist__range_ops1_compress);
PG_FUNCTION_INFO_V1(gist__range_ops1_decompress);
PG_FUNCTION_INFO_V1(gist__range_ops1_penalty);
PG_FUNCTION_INFO_V1(gist__range_ops1_picksplit);
PG_FUNCTION_INFO_V1(gist__range_ops1_union);
PG_FUNCTION_INFO_V1(gist__range_ops1_same);

Datum gist__range_ops1_consistent(PG_FUNCTION_ARGS);
Datum gist__range_ops1_compress(PG_FUNCTION_ARGS);
Datum gist__range_ops1_decompress(PG_FUNCTION_ARGS);
Datum gist__range_ops1_penalty(PG_FUNCTION_ARGS);
Datum gist__range_ops1_picksplit(PG_FUNCTION_ARGS);
Datum gist__range_ops1_union(PG_FUNCTION_ARGS);
Datum gist__range_ops1_same(PG_FUNCTION_ARGS);

/*
 *     These two functions are needed for internal reasons, users
 *     are not supposed to (and can't ) call them.
 */
PG_FUNCTION_INFO_V1(_range_in);
Datum _range_in(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(_range_out);
Datum _range_out(PG_FUNCTION_ARGS);

Datum _range_in(PG_FUNCTION_ARGS)
{
   ereport(ERROR,
           (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
           errmsg("_range_in() not implemented")));
   PG_RETURN_DATUM(0);
}

Datum _range_out(PG_FUNCTION_ARGS)
{
   ereport(ERROR,
           (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
            errmsg("_range_out() not implemented")));
   PG_RETURN_DATUM(0);
}

/* ----------------
 *   internal_range_inter_size
 *
 *   This is the routine called internally in this module. It calculates the
 *   minimum bounding box of two range arrays and then calculate the size
 *   of their intersection.
 *
 *   parameters:
 *     a - range array
 *     b - range array
 *
 *   return:
 *     return the size of the intersection of the minimum bounding boxes of
 *     the given two range array.
 *
 * ----------------
 */

RANGE_int internal__range_inter_size( ArrayType *a, ArrayType *b )
{
   RANGE range1, range2, *pa, *pb;
   RANGE_int min, max, size;

   if ( ISLEAFKEY(a) )
   {
     range1.lower = -1;
     _range_min_bounding_box(a, &(range1.lower), &(range1.upper) );
     range2.lower = -1;
     _range_min_bounding_box(b, &(range2.lower), &(range2.upper) );
     pa = &range1;
     pb = &range2;
   }
   else
   {
     pa = (RANGE *) ARRPTR(a);
     pb = (RANGE *) ARRPTR(b);
   }

   min = MAX(pa->lower,pb->lower);
   max = MIN(pa->upper,pb->upper);
   size = max - min + 1;
   return (size>0) ? size : 0;
}

/* ----------------
 *   internal_range_union
 *
 *   This is the routine called internally in this module. It generates the
 *   union of two array object.
 *
 *   parameters:
 *     a - range array
 *     b - range array
 *
 *   return:
 *     return the union of the given two range array.
 *
 * ----------------
 */

RANGE* internal__range_union( ArrayType *a, ArrayType *b )
{
   RANGE *range, *pa, *pb;

   range = (RANGE *) palloc(sizeof(RANGE));
   if ( ISLEAFKEY(a) )
   {
      range->lower = -1;
      _range_min_bounding_box(a, &(range->lower), &(range->upper) );
      _range_min_bounding_box(b, &(range->lower), &(range->upper) );
   }
   else
   {
       /* internal node */
       pa = (RANGE *) ARRPTR(a);
       pb = (RANGE *) ARRPTR(b);

       range->lower = MIN(pa->lower,pb->lower);
       range->upper = MAX(pa->upper,pb->upper);
   }

   return range;
}


/*
   The following routines are for GiST index implementation
   GiST requires us to provide six methods: (according to the GiST paper)

   1. Consistent(E,q): determine if datum in the sub tree of the given entry E can be
      guaranteed un-satisfiable to query q or not.
   2. Union(P): given a set P of entries, form a predicate that holds true for all entries.
   3. Compress(E): return the compressed representation of the given entry E.
   4. Decompress(E): return the decompressed representation of the given entry E.
   5. Penalty(E1,E2):  return a domain-specific penalty for inserting E2 into the subtree of E1.
      GiST inserts the entry with least penalty into a sub tree.
   6. PickSplit(P): split a set P of entries into two sets of entries.

   In Postgresql, query type is specified by StrategyNumber so we are able to support
   as many predicate as we could as long as the index structure supports it efficiently.
*/


/* ----------------
 *   gist__range_ops1_consistent
 *
 *   This is one of the 6 GiST routines. Given a GiST entry and a query (in this case,
 *   RANGE), this routine should determine whether this particular entry is consistent
 *   with the query with respect to the given strategy. StrategyNumber is specified in
 *   the header file and should be provided when defining the GiST access method
 *   in Postgresql.
 *
 *   Since the determination of consistency differs between leaf nodes and internal
 *   nodes, so this routine is implemented as a wrapper to two other routines. (for
 *   leaf node, internal node respectively)
 *
 *   parameters:
 *     entry - a given GiST node
 *     query - user-specified query (a range array, or range object)
 *     strategy - the strategy number (defined in range.h)
 *
 *   return:
 *     return true if this query is consistent with the given GiST entry; otherwise,
 *     return false.
 *
 *   note:
 *     If you're very sure that the entries under this given GiST node is not consistent
 *     with the query, you should always return FALSE. Returning TRUE is not harmful if
 *     subnodes are not consistent with the query, but the performance degradation is the
 *     penalty.
 *
 * ----------------
 */

Datum gist__range_ops1_consistent(PG_FUNCTION_ARGS)
{
   GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
   StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
   bool	retval;
   ArrayType *array_query;
   RANGE *tmp, *range_query;

   switch (strategy)
   {
      case GIST__RANGE_OPS1_ContainsStrategyNumber:
           array_query = (ArrayType *) PG_GETARG_POINTER(1);
           if ( GIST_LEAF(entry) )
              retval = inner_range_contains( (ArrayType *) DatumGetPointer(entry->key), array_query );
           else
           {
              tmp = (RANGE *) ARRPTR((ArrayType *) DatumGetPointer(entry->key));
              retval = internal_range_contains__range( tmp, array_query );
           }
           break;
      case GIST__RANGE_OPS1_ContainedStrategyNumber:
           array_query = (ArrayType *) PG_GETARG_POINTER(1);
           if ( GIST_LEAF(entry) )
              retval = inner_range_contains( array_query, (ArrayType *) DatumGetPointer(entry->key) );
           else
           {
              tmp = (RANGE *) ARRPTR((ArrayType *) DatumGetPointer(entry->key));
              retval = internal_range_overlaps__range( tmp, array_query );
           } 
           break;
      case GIST__RANGE_OPS1_SameStrategyNumber:
           array_query = (ArrayType *) PG_GETARG_POINTER(1);
           if ( GIST_LEAF(entry) )
             retval = (bool) DatumGetBool( DirectFunctionCall2(
                                           _range_eq,
                                           PointerGetDatum(DatumGetPointer(entry->key)),
                                           PointerGetDatum(array_query)));
           else
           {
              tmp = (RANGE *) ARRPTR((ArrayType *) DatumGetPointer(entry->key));
              retval = internal_range_contains__range( tmp, array_query);
           } 
           break;
      case GIST__RANGE_OPS1_OverlapsStrategyNumber:
           array_query = (ArrayType *) PG_GETARG_POINTER(1);
           if ( GIST_LEAF(entry) )
              retval = inner_range_overlaps( (ArrayType *) DatumGetPointer(entry->key), array_query);
           else
           {
              tmp = (RANGE *) ARRPTR((ArrayType *) DatumGetPointer(entry->key));
              retval = internal_range_overlaps__range( tmp, array_query);
           }
           break;
      case GIST__RANGE_OPS1_ContainsRangeStrategyNumber:
	   range_query = (RANGE *) PG_GETARG_POINTER(1);
           if ( GIST_LEAF(entry) )
              retval = ( _range_match_first( (ArrayType *) DatumGetPointer(entry->key), range_query ) !=0 );
           else
           {
              tmp = (RANGE *) ARRPTR((ArrayType *) DatumGetPointer(entry->key));
              retval = range_contains(tmp, range_query );
           }
           break;
       case GIST__RANGE_OPS1_ContainsIntervalStrategyNumber:
	   range_query = (RANGE *) PG_GETARG_POINTER(1);
           if ( GIST_LEAF(entry) )
              retval = inner__range_contains_interval_any( (ArrayType *) DatumGetPointer(entry->key), range_query );
           else
           {
              tmp = (RANGE *) ARRPTR((ArrayType *) DatumGetPointer(entry->key));
              retval = range_contains(tmp, range_query );
           }
           break;
       case GIST__RANGE_OPS1_ContainedIntervalStrategyNumber:
       case GIST__RANGE_OPS1_OverlapsIntervalStrategyNumber:
           range_query = (RANGE *) PG_GETARG_POINTER(1);
           if ( GIST_LEAF(entry) )
           {
              if (strategy==GIST__RANGE_OPS1_ContainedIntervalStrategyNumber)
                retval = inner__range_contained_interval_any( (ArrayType *) DatumGetPointer(entry->key), range_query );
              else
                retval = inner__range_overlaps_interval_any( (ArrayType *) DatumGetPointer(entry->key), range_query );
           }
           else
           {
              tmp = (RANGE *) ARRPTR((ArrayType *) DatumGetPointer(entry->key));
              retval = range_overlaps(tmp, range_query );
           }
           break;

      default:
      	   retval = FALSE;
   }
   PG_RETURN_BOOL(retval);
}

/* ----------------
 *   gist__range_ops1_union
 *
 *   This is one of the six required GiST methods.
 *   It will be invoked with a number of GiST entries and should return a GiST
 *   entry which holds true for any of these entries.
 *
 *   In our implementation, we returns the minimal bounding range that encloses
 *   all the entries in the entryvec.
 *
 *   parameters:
 *     entryvec - GiST entries
 *     sizep - the size of a single range object.
 *
 *   return:
 *     return a range data object which holds true for all the given entries.
 *
 * ----------------
 */
Datum gist__range_ops1_union(PG_FUNCTION_ARGS)
{
   GistEntryVector *entryvec = (GistEntryVector *) PG_GETARG_POINTER(0);
   int *size = (int *) PG_GETARG_POINTER(1);
   int4	i=0;
   ArrayType *res;
   RANGE range, *r;


   res = (ArrayType *) GETENTRY(entryvec,0);

   /* we need to know if we are going to union leaf nodes or internal nodes */
   if ( ISLEAFKEY(res) )
   {
      range.lower = -1;
      for (i=0; i<entryvec->n; i++)
         _range_min_bounding_box( (ArrayType*) GETENTRY(entryvec,i), &(range.lower), &(range.upper) );   

      res = range_to_rangeset(&range);
   }
   else
   {
      r = (RANGE*) ARRPTR((ArrayType *) GETENTRY(entryvec,i));
      range.lower = r->lower;
      range.upper = r->upper;
      for(i=1;i<entryvec->n;i++)
      {
         r = (RANGE *) ARRPTR((ArrayType*) GETENTRY(entryvec,i));
         if (r->lower<range.lower) range.lower = r->lower;
         if (r->upper>range.upper) range.upper = r->upper;
      }  
      res = range_to_rangeset(&range);
   }

   CLEARLEAFKEY(res);
   *size = VARSIZE(res);
   PG_RETURN_POINTER(res);
}

/* ----------------
 *   gist__range_ops1_compress
 *
 *   This is one of the six required GiST methods.
 *   Since we do not compress any GiST nodes, so it does nothing.
 *
 *   parameters:
 *     entry - GiST entries
 *
 *   return:
 *     return the compressed form of the given GiST entry (which is itself)
 *
 * ----------------
 */

Datum gist__range_ops1_compress(PG_FUNCTION_ARGS)
{
   GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
   GISTENTRY  *retval;
   ArrayType  *r;

   /* leaf nodes are represented as ArrayType objects, while internal
    * nodes are represented as a pair of bounding box, we want to make
    * a distinction between these two types of nodes.
    */

   if (entry->leafkey)
   {
      r = (ArrayType *) PG_DETOAST_DATUM_COPY(entry->key);
      PREPAREARR(r);
      r->flags |= LEAFKEY;
      retval = palloc(sizeof(GISTENTRY));
      gistentryinit(*retval, PointerGetDatum(r), entry->rel, entry->page, entry->offset, VARSIZE(r), FALSE);
      PG_RETURN_POINTER(retval);
   }

   r = (ArrayType *) PG_DETOAST_DATUM(entry->key);
   if (ISLEAFKEY(r) || ARRISVOID(r))
   {
      if (r != (ArrayType *) DatumGetPointer(entry->key)) pfree(r);
      PG_RETURN_POINTER(entry);
   }

   /* now, we know this is an internal node, no need to process it */
   PG_RETURN_POINTER(entry);
}

/* ----------------
 *   gist__range_ops1_decompress
 *
 *   This is one of the six required GiST methods.
 *   Since we do not compress any GiST nodes, so it does nothing.
 *
 *   parameters:
 *     entry - GiST entries
 *
 *   return:
 *     return the decompressed form of the given GiST entry (which is itself)
 *
 * ----------------
 */

Datum gist__range_ops1_decompress(PG_FUNCTION_ARGS)
{
   GISTENTRY *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
   GISTENTRY *retval;
   ArrayType *in;
   /* do nothing */

   in = (ArrayType*) PG_DETOAST_DATUM(entry->key);
   if (ARRISVOID(in))   PG_RETURN_POINTER(entry);

   if (ISLEAFKEY(in))
   {
     if ( in!=(ArrayType *) DatumGetPointer(entry->key) )
     {
        retval = palloc(sizeof(GISTENTRY));
        gistentryinit(*retval, PointerGetDatum(in), entry->rel, entry->page, entry->offset, VARSIZE(in), FALSE);
        PG_RETURN_POINTER(retval);
     }
     else 
       PG_RETURN_POINTER(entry);
  }

  PG_RETURN_POINTER(entry);
}

/* ----------------
 *   gist__range_ops1_penalty
 *
 *   This is one of the six required GiST methods.
 *   GiST kernel requires us to give it domain penalty for insertion an
 *   entry into a subtree. This routine tells GiST what penalty value should
 *   it receive if GiST change the original entry into a new entry.
 *
 *   In this implemenation, this penalty is determined by the change in area
 *   as described in the R-tree paper.
 *
 *   parameters:
 *     origentry - original GiST entry
 *     newentry  - new GiST entry
 *
 *   return:
 *     return the penalty value
 *
 * ----------------
 */

Datum gist__range_ops1_penalty(PG_FUNCTION_ARGS)
{
   GISTENTRY *origentry = (GISTENTRY *) PG_GETARG_POINTER(0);
   GISTENTRY *newentry = (GISTENTRY *) PG_GETARG_POINTER(1);
   float  *result = (float *) PG_GETARG_POINTER(2);
   ArrayType  *origarray, *newarray;
   RANGE r;
   int64 tmp1,tmp2;

   origarray = (ArrayType *) DatumGetPointer(origentry->key);
   newarray  = (ArrayType *) DatumGetPointer(newentry->key);

   r.lower = -1;
   _range_min_bounding_box( origarray, &(r.lower), &(r.upper) );
   tmp1 = r.upper - r.lower;

   _range_min_bounding_box( newarray, &(r.lower), &(r.upper) );
   tmp2 = r.upper - r.lower;

   *result = (float) (tmp2-tmp1);

   PG_RETURN_POINTER(result);
}

/* ----------------
 *   gist__range_ops1_same
 *
 *   This is one of the six required GiST methods.
 *   GiST kernel requires us to tell it whether two objects are the same.
 *   (since GiST does not know the internal representation of the user datatype)
 *
 *   In this implementation, if the interval defined by these two objects are
 *   equal, then they are the same
 *
 *   parameters:
 *     b1 - range datatype
 *     b2 - range datatype
 *     result - the result (pass by pointer )
 *
 *   return:
 *     return the boolean value (pass by pointer)
 *
 * ----------------
 */


Datum gist__range_ops1_same(PG_FUNCTION_ARGS)
{
  ArrayType *a = (ArrayType *) PointerGetDatum(PG_GETARG_POINTER(0));
  ArrayType *b = (ArrayType *) PointerGetDatum(PG_GETARG_POINTER(1));
  bool *result = (bool *) PG_GETARG_POINTER(2);
  int4 n = ARRNELEMS(a);
  RANGE *da, *db;

  if (n != ARRNELEMS(b))
  {
     *result = FALSE;
     PG_RETURN_POINTER(result);
  }
  *result = TRUE;
  da = ARRPTR(a);
  db = ARRPTR(b);
  while (n--)
     if ( range_ne(da,db) )
     {
       *result = FALSE;
        break;
     }
  PG_RETURN_POINTER(result);
}

/* 
 *   this is just a structure used internally for qsort()
 */
typedef struct
{
   OffsetNumber pos;
   int64 cost;
} SPLITCOST;

static int comparecost(const void *a, const void *b)
{
  if (((SPLITCOST *) a)->cost == ((SPLITCOST *) b)->cost)
     return 0;
  else
     return (((SPLITCOST *) a)->cost > ((SPLITCOST *) b)->cost) ? 1 : -1;
}

/* ----------------
 *   gst__range_ops1_picksplit
 *
 *   This is one of the six required GiST methods.
 *   When a node in the GiST search tree is full, we need to split this node into
 *   two. GiST requires us to dictate the split process.
 *
 *   In this implementation, we use Guttman's poly time split algorithm.
 *
 *   parameters:
 *     entryvec - entries in the old GiST node
 *     v        - the split vector. It has a rather complex structure, but the part
 *                that we need to know is
 *
 *     v->spl_ldatum, v->spl_rdatum  : pointers to left and right of the split nodes.
 *     v->spl_nleft,  v->spl_nright  : number of entries are there in each of the left, right node
 *     v->spl_left,   v->spl_right   : the entry numbers which belong to the left, right node respectively
 *
 *   return:
 *     the GiST split vector structure.
 *
 * ----------------
 */
Datum gist__range_ops1_picksplit(PG_FUNCTION_ARGS)
{
   GistEntryVector *entryvec = (GistEntryVector *) PG_GETARG_POINTER(0);
   GIST_SPLITVEC *v = (GIST_SPLITVEC *) PG_GETARG_POINTER(1);
   OffsetNumber i, j;
   ArrayType  *datum_alpha,  *datum_beta;
   ArrayType  *datum_l, *datum_r;
   //ArrayType  *union_d, *union_dl, *union_dr;
   RANGE *union_d, *union_dl, *union_dr;
   RANGE range;
   bool firsttime;

   RANGE_int size_alpha, size_beta, size_union, size_inter;
   RANGE_int size_waste, waste;
   RANGE_int size_l, size_r, *size;

   int nbytes;
   bool isLeaf;
   OffsetNumber seed_1 = 0,seed_2 = 0;
   OffsetNumber *left, *right;
   OffsetNumber maxoff;
   SPLITCOST  *costvector;

   size = 0;

   datum_l = (ArrayType*) GETENTRY(entryvec,FirstOffsetNumber);
   isLeaf = ISLEAFKEY(datum_l);

   maxoff = entryvec->n - 2;
   nbytes = (maxoff + 2) * sizeof(OffsetNumber);
   v->spl_left = (OffsetNumber *) palloc(nbytes);
   v->spl_right = (OffsetNumber *) palloc(nbytes);

   firsttime = true;
   waste = 0;

   for (i = FirstOffsetNumber; i < maxoff; i = OffsetNumberNext(i))
   {
      datum_alpha = (ArrayType *) GETENTRY(entryvec, i);

      for (j = OffsetNumberNext(i); j <= maxoff; j = OffsetNumberNext(j))
      {
  	 datum_beta = (ArrayType *) GETENTRY(entryvec, j);

	 /* compute the wasted space by unioning these guys */
	 /* size_waste = size_union - size_inter; */

	 union_d = internal__range_union(datum_alpha, datum_beta);
         size = range_size(union_d);
         size_union = *size;
	 size_inter = internal__range_inter_size(datum_alpha, datum_beta);
	 size_waste = size_union - size_inter;
	 if (union_d!=NULL)  pfree(union_d);

         /* are these a more promising split that what we've already seen? */
 	 
         if (size_waste > waste || firsttime)
	 {
	    waste = size_waste;
	    seed_1 = i;
	    seed_2 = j;
	    firsttime = false;
	 }
      }
   }


   left = v->spl_left;
   v->spl_nleft = 0;
   right = v->spl_right;
   v->spl_nright = 0;
   if (seed_1 == 0 || seed_2 == 0)
   {
      seed_1 = 1;
      seed_2 = 2;
   }

   datum_alpha = (ArrayType *) GETENTRY(entryvec, seed_1);
   datum_l = copy_rangeArray(datum_alpha);
   range.lower = -1;
   _range_min_bounding_box( datum_alpha, &(range.lower), &(range.upper) );
   size_l = range.upper-range.lower;


   datum_beta = (ArrayType *) GETENTRY(entryvec, seed_2);
   datum_r = copy_rangeArray(datum_beta);
   range.lower = -1;
   _range_min_bounding_box( datum_beta, &(range.lower), &(range.upper) );

   size = range_size(&range);
   size_r = *size;

   maxoff = OffsetNumberNext(maxoff);

   /* sort entries */
   costvector = (SPLITCOST *) palloc(sizeof(SPLITCOST) * maxoff);


   for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i))
   {
      costvector[i - 1].pos = i;
      datum_alpha = (ArrayType *) GETENTRY(entryvec, i);
      union_d = internal__range_union(datum_l, datum_alpha);

      size = range_size(union_d);
      size_alpha = *size;

      union_d = internal__range_union(datum_r, datum_alpha);

      size = range_size(union_d);
      size_beta = *size;

      costvector[i - 1].cost = ABS((size_alpha - size_l) - (size_beta - size_r));
   }

   qsort((void *) costvector, maxoff, sizeof(SPLITCOST), comparecost);

   /*
    * Now split up the regions between the two seeds.	An important
    * property of this split algorithm is that the split vector v has the
    * indices of items to be split in order in its left and right
    * vectors.  We exploit this property by doing a merge in the code
    * that actually splits the page.
    *
    * For efficiency, we also place the new index tuple in this loop. This
    * is handled at the very end, when we have placed all the existing
    * tuples and i == maxoff + 1.
    */


   for (j = 0; j < maxoff; j++)
   {
      i = costvector[j].pos;

      /*
       * If we've already decided where to place this item, just put it
       * on the right list.  Otherwise, we need to figure out which page
       * needs the least enlargement in order to store the item.
       */

      if (i == seed_1)
      {
  	 *left++ = i;
 	 v->spl_nleft++;
	 continue;
      }
      else if (i == seed_2)
      {
         *right++ = i;
	 v->spl_nright++;
	 continue;
      }

      /* okay, which page needs least enlargement? */
      datum_alpha = (ArrayType *) GETENTRY(entryvec, i);
      union_dl = internal__range_union(datum_l, datum_alpha);
      union_dr = internal__range_union(datum_r, datum_alpha);

      size_alpha = *(range_size(union_dl));
      size_beta  = *(range_size(union_dr));

      /* pick which page to add it to */
      //if (size_alpha - size_l < size_beta - size_r + WISH_F(v->spl_nleft, v->spl_nright, 0.01))
      if (size_alpha-size_l<size_beta-size_r)
      {
	 if (datum_l)   pfree(datum_l);
	 if (union_dr)	pfree(union_dr);
         datum_l = range_to_rangeset(union_dl);
	 size_l = size_alpha;
	 *left++ = i;
	 v->spl_nleft++;
      }
      else
      {
  	 if (datum_r) 	pfree(datum_r);
	 if (union_dl)	pfree(union_dl);
	 datum_r = range_to_rangeset(union_dr);
	 size_r = size_beta;
	 *right++ = i;
	 v->spl_nright++;
      }
   }
   pfree(costvector);
   *right = *left = FirstOffsetNumber;

   datum_l->flags &= ~LEAFKEY;
   datum_r->flags &= ~LEAFKEY;
   v->spl_ldatum = PointerGetDatum(datum_l);
   v->spl_rdatum = PointerGetDatum(datum_r);

   PG_RETURN_POINTER(v);
}
