/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU  General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * _range_op.c
 *
 *    This file contains routines for _range data type.
 *
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */


#include "postgres.h"
#include "executor/executor.h"
#include "catalog/pg_type.h"
#include "fmgr.h"
#include "funcapi.h"
#include "executor/spi.h"
#include "range.h"
#include "_range.h"
#include "global.h"

PG_FUNCTION_INFO_V1(_range_ne);
PG_FUNCTION_INFO_V1(_range_eq);
PG_FUNCTION_INFO_V1(_range_contains);
PG_FUNCTION_INFO_V1(_range_contained);
PG_FUNCTION_INFO_V1(_range_overlaps);
PG_FUNCTION_INFO_V1(_range_union);
PG_FUNCTION_INFO_V1(_range_inters);
PG_FUNCTION_INFO_V1(rangeset);
PG_FUNCTION_INFO_V1(rcount);
PG_FUNCTION_INFO_V1(sort);
PG_FUNCTION_INFO_V1(sort_asc);
PG_FUNCTION_INFO_V1(sort_desc);
PG_FUNCTION_INFO_V1(unique);
PG_FUNCTION_INFO_V1(idx);
PG_FUNCTION_INFO_V1(subarray);
PG_FUNCTION_INFO_V1(_range_push_elem);
PG_FUNCTION_INFO_V1(_range_push_array);
PG_FUNCTION_INFO_V1(_range_del_elem);
PG_FUNCTION_INFO_V1(_range_union_elem);
PG_FUNCTION_INFO_V1(_range_subtract);

PG_FUNCTION_INFO_V1(_range_contains_interval_any);
PG_FUNCTION_INFO_V1(_range_contains_interval_all);
PG_FUNCTION_INFO_V1(_range_contained_interval_any);
PG_FUNCTION_INFO_V1(_range_contained_interval_all);
PG_FUNCTION_INFO_V1(_range_overlaps_interval_any);
PG_FUNCTION_INFO_V1(_range_overlaps_interval_all);

/*
 *  We need the oid for range datatype occasionally. However, it is
 *  costly to do table lookup every time when we need it, so we
 *  cache it. BUT, this is based on the following assumptions:
 *  1. RANGE datatype has already existed!
 *  2. oid for range doesn't change during the execution
 *  Under normal circumstances, these assumptions are true.
 */

static uint range_oid = 0;    // oid will never be 0 for user-defined datatype


/* ----------------
 *   isort
 *
 *   This is the routine used to create internal representation of range
 *   data type in Postgresql. Any new value of user data types in Postgresql
 *   is provided in string and Postgresql server calls this routine to form
 *   the internal representation of this user data type.
 *
 *   It is our responsibility to parse the string (according to the format we
 *   specify) and form the internal representation of range object.
 *
 *   the supported syntax for range data type is as follows:
 *
 *   INTEGER    <- [1-9]*[0-9]
 *   RANGE      <- INTEGER .. INTEGER
 *   RANGE      <- INTEGER : INTEGER
 *
 *   in other words, it looks like
 *   INTEGER..INTEGER      -- perl slice syntax
 *   INTEGER:INTEGER       -- python slice syntax
 *
 *
 *   parameters:
 *     str - a C string from user input
 *
 *   return:
 *     construct the internal representation of the data type.
 *
 *   note:
 *
 *   if the given string does not follow the syntax, error must be returned by
 *   elog( ERROR, "error message" ). This routine will return immediately.
 *
 * ---------------
 */

bool isort( RANGE *a, int len )
{
   int4 index;
   RANGE *cur, *end, *tmp;
   bool r = FALSE;

   end = a + len;
   do
   {
      index = 0;
      cur = a+1;
      while (cur<end )
      {
        if ( range_gt(cur-1,cur) )
        {
          tmp = range_duplicate(cur-1);
          range_copy( cur-1, cur);
          range_copy( cur, tmp );
          pfree(tmp);
	  index = 1;  
        }
        else if ( !r && range_eq(cur-1,cur) )  r = TRUE;
        cur++;
      }
   }  while (index);
   return r;
}



/* ----------------
 *   new_rangeArray
 *
 *   This is the routine used to create internal representation of 1-D range
 *   array (number of elements is specified, however, the total number
 *   of elements are not limited to this initial value).
 *
 *   parameters:
 *     num - initial value for array size
 *
 *   return:
 *     construct the internal representation of the range array.
 *
 * ---------------
 */
ArrayType *new_rangeArray( int num )
{
   ArrayType *r;
   SPITupleTable *spi_tuptable;
   TupleDesc spi_tupdesc;
   HeapTuple spi_tuple;
   int nbytes, ret, proc;
   char *val;

   nbytes = ARR_OVERHEAD(NDIM) + sizeof(RANGE) * num;
   r = (ArrayType *) palloc0(nbytes);	// allocate enough memory and fill with 0

   ARR_SIZE(r) = nbytes;
   ARR_NDIM(r) = NDIM;

   if ( range_oid==0 )
   {
      if ((ret=SPI_connect())<0)
        elog( ERROR, "new_rangeArray: SPI_connect returns %d",ret);

      ret = SPI_execute("select oid from pg_type where typname='range'",true,0);
      proc = SPI_processed;
      if ((ret==SPI_OK_SELECT) && (proc>0))
      {
         spi_tuptable = SPI_tuptable;
         spi_tupdesc = spi_tuptable->tupdesc;
         spi_tuple = spi_tuptable->vals[0];
         val = SPI_getvalue(spi_tuple, spi_tupdesc,1);
         range_oid = atoi(val);
      }
      else
        elog(ERROR, "unable to find oid for range datatype!" );

      SPI_finish();
   }

   ARR_ELEMTYPE(r) = range_oid;
   *((int *) ARR_DIMS(r)) = num;
   *((int *) ARR_LBOUND(r)) = 1;

   return r;
}


/* ----------------
 *   resize_rangeArray
 *
 *   This is the routine used to resize the internal representation of 
 *   1-D range array
 *
 *   parameters:
 *     a   - range array
 *     num - initial value for array size
 *
 *   return:
 *     construct the internal representation of the range array.
 *
 * ---------------
 */

ArrayType *resize_rangeArray(ArrayType *a, int num)
{
   int nbytes;

   nbytes = ARR_OVERHEAD(NDIM) + sizeof(RANGE)*num;

   if (num==ARRNELEMS(a))  return a;

   a = (ArrayType *) repalloc(a, nbytes);
   a->size = nbytes;
   *((int *) ARR_DIMS(a)) = num;
   return a;
}

/* ----------------
 *   copy_rangeArray
 *
 *   This routine faithly duplicate an existing range array
 *
 *   parameters:
 *     a   - range array
 *
 *   return:
 *     construct a new range array.
 *
 * ---------------
 */

ArrayType *copy_rangeArray(ArrayType *a)
{
   ArrayType *r;
   r = new_rangeArray(ARRNELEMS(a));
   memmove(r, a, VARSIZE(a));
   return r;
}

/* ----------------
 *   range_to_rangeset
 *
 *   This routine is typically used to transform a range into 
 *   range array.
 *
 *   parameters:
 *     a   - range object
 *
 *   return:
 *     construct a range array containing the range object (ie. with only one
 *     element).
 *
 * ---------------
 */
ArrayType *range_to_rangeset(RANGE *r)
{
   ArrayType *result;
   RANGE *newr;

   result = new_rangeArray(1);
   newr = ARRPTR(result);
   memcpy( newr, r, sizeof(RANGE) );
   return result;
}

/* ----------------
 *   _range_concat
 *
 *   This routine is typically used to concatenate two 
 *   range arrays. NOTICE: this routine does not eliminate 
 *   duplicate elements in both array, it just faithly concatenate
 *   these two arrays.
 *
 *   parameters:
 *     a   - range array
 *     b   - range array
 *
 *   return:
 *     construct a range array containing the range object (ie. with only one
 *     element).
 *
 * ---------------
 */
ArrayType *_range_concat(ArrayType *a, ArrayType *b)
{
   ArrayType *result;
   int32 ac = (ARRISVOID(a)) ? 0 : ARRNELEMS(a);
   int32 bc = (ARRISVOID(b)) ? 0 : ARRNELEMS(b);

   result = new_rangeArray(ac+bc);
   if (ac)  memcpy(ARRPTR(result), ARRPTR(a), ac*sizeof(RANGE));
   if (bc)  memcpy(ARRPTR(result)+ac, ARRPTR(b), bc*sizeof(RANGE));
   return result;
}

/* r is sorted and size of r>1 */
ArrayType *_range_unique(ArrayType *r)
{
   RANGE *tmp, *dr, *data;
   int num = ARRNELEMS(r);

   if (num < 2) return r;

   data = tmp = dr = ARRPTR(r);
   while (tmp - data < num)
     if ( range_ne(tmp,dr) )
     {
        dr++;
        range_copy( dr, tmp );
        tmp++;
     }
     else
        tmp++;
   return resize_rangeArray(r, dr + 1 - ARRPTR(r));

}

int32 _range_match_first(ArrayType *a, RANGE *r)
{
  RANGE *b;
  int32 c, i;

  c = (ARRISVOID(a)) ? 0 : ARRNELEMS(a);
  b = ARRPTR(a);
  for(i=0;i<c;i++,b++)
     if (range_eq(b,r))  return (i+1);

  return 0;
}


/* ----------------
 *   _range_add_elem
 *
 *   We can use this routine to add one more element into the
 *   range array.
 *
 *   parameters:
 *     a   - range array
 *     r   - range object
 *
 *   return:
 *     construct a range array containing the content of the given range array and
 *     the range object element as well.
 *
 * ---------------
 */

ArrayType *_range_add_elem(ArrayType *a, RANGE *r)
{
   ArrayType *result;
   RANGE *newr;
   int32 c;

   c = (ARRISVOID(a)) ? 0 : ARRNELEMS(a);
   result = new_rangeArray(c+1);
   newr = ARRPTR(result);
   if (c>0)  memcpy(newr, ARRPTR(a), c*sizeof(RANGE));
   memcpy( newr+c, r, sizeof(RANGE));
   return result;
}

void _range_min_bounding_box( ArrayType *arr, RANGE_int *min, RANGE_int *max )
{
   int i, nelems = ARRNELEMS(arr);
   RANGE_int lmin, lmax;
   RANGE *rptr = ARRPTR(arr);


   /* this is a loop hole which enables us to continue enlarging
    * the minimum bounding box given another range array
    */
   i = 0;
   if ( *min>=0 )
   {
      lmin = *min;
      lmax = *max;
   }
   else
   {
      lmin = rptr->lower;
      lmax = rptr->upper;
      rptr++;
      i++;
   }

   for (;i<nelems;i++)
   {
      if ( rptr->lower < lmin )  lmin = rptr->lower;
      if ( rptr->upper > lmax )  lmax = rptr->upper;
      rptr++;
   }

   *min = lmin;
   *max = lmax;
}

int compASC(const void *a, const void *b)
{
   int v;

   v = range_cmp( (RANGE*) a, (RANGE*) b);
   return v;
}

int compDESC(const void *a, const void *b)
{
  int v;
  v = range_cmp( (RANGE*) a, (RANGE*) b);
  return (-1)*v;
}

// functions called externally

Datum rangeset(PG_FUNCTION_ARGS)
{
   PG_RETURN_POINTER(range_to_rangeset((RANGE*)PG_GETARG_POINTER(0)));
}

Datum rcount(PG_FUNCTION_ARGS)
{
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(0)));
   int32 count = (ARRISVOID(a))? 0 : ARRNELEMS(a);

   PG_FREE_IF_COPY(a,0);
   PG_RETURN_INT32(count);
}



Datum idx(PG_FUNCTION_ARGS)
{
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(0)));
   int32 result = (ARRISVOID(a)) ? 0 : ARRNELEMS(a);

   if (result)  result = _range_match_first(a, (RANGE*)PG_GETARG_POINTER(1));
   PG_FREE_IF_COPY(a,0);
   PG_RETURN_INT32(result);
}

Datum subarray(PG_FUNCTION_ARGS)
{
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(0)));
   ArrayType *result;
   int32 start = (PG_GETARG_INT32(1)>0) ? PG_GETARG_INT32(1)-1 : PG_GETARG_INT32(1);
   int32 len = (fcinfo->nargs == 3) ? PG_GETARG_INT32(2) : 0;
   int32 end = 0;
   int32 c;

   if (ARRISVOID(a))
   {
       PG_FREE_IF_COPY(a,0);
       PG_RETURN_POINTER(new_rangeArray(0));
   }

   c = ARRNELEMS(a);
   if (start<0)  start = c+start;
   if (len<0) end = c+len;
   else if (len==0)  end = c;
   else end = start + len;

   if (end>c)  end = c;
   if (start<0)  start = 0;
   if (start >= end || end <=0)
   {
      PG_FREE_IF_COPY(a,0);
      PG_RETURN_POINTER(new_rangeArray(0));
   }

   result = new_rangeArray(end-start);
   if ( end-start>0)  memcpy(ARRPTR(result), ARRPTR(a)+start,  (end-start)*sizeof(RANGE));
   PG_FREE_IF_COPY(a,0);
   PG_RETURN_POINTER(result);
}

Datum sort(PG_FUNCTION_ARGS)
{
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   text *dirstr = (fcinfo->nargs==2) ? PG_GETARG_TEXT_P(1): NULL;
   int32 dc = (dirstr)? VARSIZE(dirstr)-VARHDRSZ : 0;
   char *d = (dirstr)? VARDATA(dirstr):NULL;
   int dir = -1;

   if (ARRISVOID(a) || ARRNELEMS(a)<2)   PG_RETURN_POINTER(a);
   if (dirstr ==NULL || (dc==3 && (d[0]=='A' || d[0]=='a')
			       && (d[1]=='S' || d[1]=='s')
                               && (d[2]=='C' || d[2]=='c')))
	dir = 1;
   else if (dc==4 && (d[0]=='D' || d[0]=='d')
		  && (d[1]=='E' || d[1]=='e')
                  && (d[2]=='S' || d[2]=='s')
                  && (d[3]=='C' || d[3]=='c'))
	dir = 0;

   if (dir==-1)
	ereport(ERROR, 
                (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                 errmsg("second parameter must be \"ASC\" or \"DESC\"")));

   QSORT(a,dir);
   PG_RETURN_POINTER(a);
}

Datum sort_asc(PG_FUNCTION_ARGS)
{
   ArrayType  *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));

   if (ARRISVOID(a))  PG_RETURN_POINTER(a);
   QSORT(a, 1);
   PG_RETURN_POINTER(a);
}

Datum sort_desc(PG_FUNCTION_ARGS)
{
   ArrayType  *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));

   if (ARRISVOID(a))  PG_RETURN_POINTER(a);
   QSORT(a, 0);
   PG_RETURN_POINTER(a);
}

Datum unique(PG_FUNCTION_ARGS)
{
   ArrayType  *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));

   if (ARRISVOID(a) || ARRNELEMS(a) < 2)  PG_RETURN_POINTER(a);
   a = _range_unique(a);
   PG_RETURN_POINTER(a);

}

Datum _range_push_elem(PG_FUNCTION_ARGS)
{
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(0)));
   ArrayType *result;

   result = _range_add_elem( a, (RANGE*) PG_GETARG_POINTER(1) );	
   PG_FREE_IF_COPY(a,0);
   PG_RETURN_POINTER(result);
}

Datum _range_push_array(PG_FUNCTION_ARGS)
{
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(0)));
   ArrayType *b = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(1)));

   ArrayType *result;

   result = _range_concat(a,b);
   PG_FREE_IF_COPY(a,0);
   PG_FREE_IF_COPY(b,1);
   PG_RETURN_POINTER(result);
}


Datum _range_del_elem(PG_FUNCTION_ARGS)
{
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   int32 c = (ARRISVOID(a)) ? 0 : ARRNELEMS(a);
   int32 n = 0, i;
   RANGE *elem = (RANGE*) PG_GETARG_POINTER(1);
   RANGE *aa = ARRPTR(a);

   for(i=0;i<c;i++)
   {
     if ( range_ne(aa+i,elem) )
     {
        if (i>n) 
        { 
           range_copy( aa+n, aa+i );
           n++;
        } 
        else n++;

     }
   }
   if (c>0)  a = resize_rangeArray(a,n);
   PG_RETURN_POINTER(a);
}

Datum _range_union_elem(PG_FUNCTION_ARGS)
{
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   ArrayType *result;

   result = _range_add_elem(a, (RANGE*) PG_GETARG_POINTER(1));
   PG_FREE_IF_COPY(a,0);
   QSORT(result,1);
   PG_RETURN_POINTER(_range_unique(result));   
}

Datum _range_subtract(PG_FUNCTION_ARGS)
{
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   ArrayType *b = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(1)));
   ArrayType *result;
   int32 ca = ARRISVOID(a);
   int32 cb = ARRISVOID(b);
   RANGE *aa, *bb, *r;
   int32 n=0, i=0, k=0;

   QSORT(a,1);
   a = _range_unique(a);
   ca = ARRNELEMS(a);
   QSORT(b,1);
   b = _range_unique(b);
   cb = ARRNELEMS(b);
   result = new_rangeArray(ca);
   aa = ARRPTR(a);
   bb = ARRPTR(b);
   r = ARRPTR(result);
   while (i<ca)
   {
      if (k==cb || range_lt(aa+i,bb+k) )  
      {
         range_copy( r+n, aa+i);
         n++;
         i++;
      }
      else if ( range_eq(aa+i,bb+k) )
      {
         i++;   k++;
      }
      else k++;
   }

   result = resize_rangeArray(result,n);
   pfree(a);
   pfree(b);
   PG_RETURN_POINTER(result);
}

// again, called only internally


bool inner_range_contains(ArrayType *a, ArrayType *b)
{   
   int na, nb, i, j, n;     
   RANGE *ra, *rb;
   int4 res;

   if (ARRISVOID(a) || ARRISVOID(b))  return FALSE;

   na = ARRNELEMS(a);
   nb = ARRNELEMS(b);
   ra = ARRPTR(a);
   rb = ARRPTR(b);

   i = j = n = 0;
   while (i < na && j < nb)
   {
      res = range_cmp(ra+i,rb+j);
      if ( res==-1) i++;
      else if ( res==0 )
      {
         n++;
         i++;
         j++;
      }
      else
         j++;
   }

   return (n == nb) ? TRUE : FALSE;
}

/* if any of the inteval (range) in the range array contains the given range r */
bool inner__range_contains_interval_any(ArrayType *a, RANGE *r)
{
   RANGE *ra;
   int i, nelem;

   if (ARRISVOID(a)) return FALSE;

   nelem = ARRNELEMS(a);
   ra = ARRPTR(a);
   for (i=0;i<nelem;i++)
   {
      if ( range_contains(ra,r) )  return TRUE;
      ra++;
   }
   return FALSE;
}   

/* if all of the inteval (range) in the range array contains the given range r */
bool inner__range_contains_interval_all(ArrayType *a, RANGE *r)
{
   RANGE *ra;
   int i, nelem;

   if (ARRISVOID(a)) return FALSE;

   nelem = ARRNELEMS(a);
   ra = ARRPTR(a);
   for (i=0;i<nelem;i++)
   {
      if ( range_contains(ra,r)==FALSE )  return FALSE;
      ra ++;
   }
   return TRUE;
}

/* if any of the inteval (range) in the range array contained the given range r */
bool inner__range_contained_interval_any(ArrayType *a, RANGE *r)
{
   RANGE *ra;
   int i, nelem;

   if (ARRISVOID(a)) return FALSE;

   nelem = ARRNELEMS(a);
   ra = ARRPTR(a);
   for (i=0;i<nelem;i++)
   {
      if ( range_contained(ra,r) )  return TRUE;
      ra ++;
   }
   return FALSE;
}

/* if all of the inteval (range) in the range array contained the given range r */
bool inner__range_contained_interval_all(ArrayType *a, RANGE *r)
{
   RANGE *ra;
   int i, nelem;

   if (ARRISVOID(a)) return FALSE;

   nelem = ARRNELEMS(a);
   ra = ARRPTR(a);
   for (i=0;i<nelem;i++)
   {
      if ( range_contained(ra,r)==FALSE )  return FALSE;
      ra ++;
   }
   return TRUE;
}

/* if any of the inteval (range) in the range array overlaps the given range r */
bool inner__range_overlaps_interval_any(ArrayType *a, RANGE *r)
{
   RANGE *ra;
   int i, nelem;

   if (ARRISVOID(a)) return FALSE;

   nelem = ARRNELEMS(a);
   ra = ARRPTR(a);
   for (i=0;i<nelem;i++)
   {
      if ( range_overlaps(ra,r) )  return TRUE;
      ra ++;
   }
   return FALSE;
}

/* if all of the inteval (range) in the range array contained the given range r */
bool inner__range_overlaps_interval_all(ArrayType *a, RANGE *r)
{
   RANGE *ra;
   int i, nelem;

   if (ARRISVOID(a)) return FALSE;

   nelem = ARRNELEMS(a);
   ra = ARRPTR(a);
   for (i=0;i<nelem;i++)
   {
      if ( range_overlaps(ra,r)==FALSE )  return FALSE;
      ra ++;
   }
   return TRUE;
}


/* ----------------
 *   internal_range_overlaps__range
 *
 *   This is the routine called internally in this module to check if the given
 *   range overlaps the ranges in the second argument (which is a range array)
 *
 *   parameters:
 *     range - range
 *     arr   - range array
 *
 *   return:
 *     return true if this given range (1st argument) overlaps any of the ranges in
 *     the given range array.
 *
 * ----------------
 */

bool internal_range_overlaps__range( RANGE *range, ArrayType *arr )
{
   RANGE r;

   r.lower = -1;
   _range_min_bounding_box( arr, &(r.lower), &(r.upper) );

   return range_overlaps(range,&r);
}
/* ----------------
 *   internal_range_containss__range
 *
 *   This is the routine called internally in this module to check if the given
 *   range contains the minimum bounding box of the second argument (which is a range array) *
 *   parameters:
 *     range - range
 *     arr   - range array
 *
 *   return:
 *     return true if this given range (1st argument) contains the minimum bounding box of the
 *     ranges in the given range array.
 *
 * ----------------
 */

bool internal_range_contains__range( RANGE *range, ArrayType *arr )
{
   RANGE r;

   r.lower = -1;
   _range_min_bounding_box( arr, &(r.lower), &(r.upper) );
   return range_contains(range,&r);
}


bool _range_contains_range(ArrayType *a, RANGE *b)
{
   if (ARRISVOID(a))  return FALSE;

   return _range_match_first(a,b)==0 ? FALSE : TRUE; 
}

bool inner_range_overlaps(ArrayType *a, ArrayType *b)
{
   int na, nb, i, j;
   RANGE *ra, *rb;
   int4 res;  

   if (ARRISVOID(a) || ARRISVOID(b))  return FALSE;

   na = ARRNELEMS(a);
   nb = ARRNELEMS(b);
   ra = ARRPTR(a);
   rb = ARRPTR(b);

   i = j = 0;
   while (i < na && j < nb)
   {
      res = range_cmp(ra+i,rb+j);
      if ( res==-1)
          i++;
      else if ( res==0 )
          return TRUE;
      else
          j++;
   }

   return FALSE;
}

ArrayType *inner_range_union(ArrayType *a, ArrayType *b)
{                       
   ArrayType  *r = NULL;
   int na, nb, i, j;
   RANGE *ra, *rb, *rr;

   if (ARRISVOID(a) && ARRISVOID(b))  return new_rangeArray(0);
   if (ARRISVOID(a))  r = copy_rangeArray(b);
   if (ARRISVOID(b))  r = copy_rangeArray(a);

   if (r)
      rr = ARRPTR(r);
   else
   {
      na = ARRNELEMS(a);
      nb = ARRNELEMS(b); 
      ra = ARRPTR(a);
      rb = ARRPTR(b);

      r = new_rangeArray(na + nb);
      rr = ARRPTR(r);   

      /* union */
      i = j = 0;               
      while (i < na && j < nb)   
      if ( range_lt( ra+i, rb+j) )
      {
         range_copy( rr, ra+i );
         rr++;
         i++;
      }   
      else      
      {
         range_copy( rr, rb+j);
         rr++;
         j++;
      }                        

      while (i < na)
      {
         range_copy( rr, ra+i );
         rr++;
         i++;
      }

      while (j<nb)
      {
         range_copy( rr, rb+j);
         rr++;
         j++;
      }

   }

   if (ARRNELEMS(r) > 1)  
      r = _range_unique(r);

   return r;
}                                 


ArrayType *inner_range_inters(ArrayType *a, ArrayType *b)
{
   ArrayType  *r;
   int na, nb, i, j;
   RANGE *ra, *rb, *rr;
   bool res;

   if (ARRISVOID(a) || ARRISVOID(b))  return new_rangeArray(0);

   na = ARRNELEMS(a);
   nb = ARRNELEMS(b);
   ra = ARRPTR(a);
   rb = ARRPTR(b);           
   r = new_rangeArray(Min(na, nb));
   rr = ARRPTR(r);

   i = j = 0;
   while (i < na && j < nb)
   {
      res = range_lt( ra+i, rb+j);
      if ( res==-1 ) i++;
      else if (res==1) j++;
      else
      {
         if (i+j==0 || (i+j>0 && range_ne(rr-1,rb+j) ))
         {
           range_copy( rr, rb+j);
           rr++;
         }
         i++;
         j++;
      }   
   }

   if ( (rr - ARRPTR(r)) == 0)
   {
      pfree(r);         
      return new_rangeArray(0);
   }
   else
      return resize_rangeArray(r, rr - ARRPTR(r));
}




// called externally


Datum _range_contained(PG_FUNCTION_ARGS)
{
  PG_RETURN_BOOL(DatumGetBool( DirectFunctionCall2(
                                 _range_contains,
                                 PointerGetDatum(PG_GETARG_POINTER(1)),
                                 PointerGetDatum(PG_GETARG_POINTER(0)))));
}

Datum _range_contains(PG_FUNCTION_ARGS)
{
   bool res;
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   ArrayType *b = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(1)));

   if (ARRISVOID(a)||ARRISVOID(b))  return FALSE;

   PREPAREARR(a);
   PREPAREARR(b);
   res = inner_range_contains(a,b);
   pfree(a);
   pfree(b);
   PG_RETURN_BOOL(res);
}


Datum _range_ne(PG_FUNCTION_ARGS)
{
  PG_RETURN_BOOL(!DatumGetBool( DirectFunctionCall2(
                                 _range_eq,
                                 PointerGetDatum(PG_GETARG_POINTER(1)),
                                 PointerGetDatum(PG_GETARG_POINTER(0)))));
}

Datum _range_eq(PG_FUNCTION_ARGS)
{
   int na, nb, n;
   RANGE *ra, *rb;
   bool res;
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   ArrayType *b = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(1)));
   bool avoid = ARRISVOID(a), bvoid = ARRISVOID(b);


   if (avoid || bvoid)  return (avoid && bvoid) ? TRUE:FALSE;

   SORT(a);
   SORT(b);

   na = ARRNELEMS(a);
   nb = ARRNELEMS(b);
   ra = ARRPTR(a);
   rb = ARRPTR(b);

   res = FALSE;
   if ( na==nb)
   {
      res = TRUE;
      for(n=0;n<na;n++)
      {
         if (range_ne(ra,rb))
         {
            res = FALSE;
            break;
         }
      }
   }

   pfree(a);
   pfree(b);

   PG_RETURN_BOOL(res);
}

/* _range_overlaps -- does a overlap b? */
Datum _range_overlaps(PG_FUNCTION_ARGS)
{
   ArrayType  *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   ArrayType  *b = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(1)));
   bool            result;

   if (ARRISVOID(a) || ARRISVOID(b))  return FALSE;

   SORT(a);
   SORT(b);                                                                                          

   result = inner_range_overlaps(a, b);

   pfree(a);
   pfree(b);

   PG_RETURN_BOOL(result);  
}

Datum _range_union(PG_FUNCTION_ARGS)
{
   ArrayType  *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   ArrayType  *b = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(1)));
   ArrayType  *result;

   if (!ARRISVOID(a))   SORT(a);   
   if (!ARRISVOID(b))   SORT(b);                                                                                  

   result = inner_range_union(a, b);  

   if (a) pfree(a);
   if (b) pfree(b);

   PG_RETURN_POINTER(result);
}

Datum _range_inters(PG_FUNCTION_ARGS)
{
   ArrayType  *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   ArrayType  *b = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(1)));
   ArrayType * result;

   if (ARRISVOID(a) || ARRISVOID(b))  return FALSE;

   SORT(a);
   SORT(b);

   result = inner_range_inters(a, b);

   pfree(a);
   pfree(b);

   PG_RETURN_POINTER(result);
}


Datum _range_contains_interval_any(PG_FUNCTION_ARGS)
{
   bool res;
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   //RANGE     *b = (RANGE *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(1)));
   RANGE *b = (RANGE*) DatumGetPointer(PG_GETARG_DATUM(1));

   if (ARRISVOID(a))  return FALSE;

   PREPAREARR(a);
   res = inner__range_contains_interval_any(a,b);
   pfree(a);
   //pfree(b);
   PG_RETURN_BOOL(res);
}

Datum _range_contains_interval_all(PG_FUNCTION_ARGS)
{
   bool res;
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   //RANGE     *b = (RANGE *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(1)));
   RANGE *b = (RANGE *) DatumGetPointer(PG_GETARG_DATUM(1));

   if (ARRISVOID(a))  return FALSE;

   PREPAREARR(a);
   res = inner__range_contains_interval_all(a,b);
   pfree(a);
   //pfree(b);
   PG_RETURN_BOOL(res);
}

Datum _range_contained_interval_any(PG_FUNCTION_ARGS)
{
   bool res;
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   //RANGE     *b = (RANGE *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(1)));
   RANGE *b = (RANGE *) DatumGetPointer(PG_GETARG_DATUM(1));

   if (ARRISVOID(a))  return FALSE;

   PREPAREARR(a);
   res = inner__range_contained_interval_any(a,b);
   pfree(a);
   //pfree(b);
   PG_RETURN_BOOL(res);
}

Datum _range_contained_interval_all(PG_FUNCTION_ARGS)
{
   bool res;
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   //RANGE     *b = (RANGE *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(1)));
   RANGE *b = (RANGE *) DatumGetPointer(PG_GETARG_DATUM(1));

   if (ARRISVOID(a))  return FALSE;

   PREPAREARR(a);
   res = inner__range_contained_interval_all(a,b);
   pfree(a);
   //pfree(b);
   PG_RETURN_BOOL(res);
}
Datum _range_overlaps_interval_any(PG_FUNCTION_ARGS)
{
   bool res;
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   //RANGE     *b = (RANGE *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(1)));
   RANGE *b = (RANGE *) DatumGetPointer(PG_GETARG_DATUM(1));

   if (ARRISVOID(a))  return FALSE;

   PREPAREARR(a);
   res = inner__range_overlaps_interval_any(a,b);
   pfree(a);
   //pfree(b);
   PG_RETURN_BOOL(res);
}

Datum _range_overlaps_interval_all(PG_FUNCTION_ARGS)
{
   bool res;
   ArrayType *a = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(0)));
   //RANGE     *b = (RANGE *) DatumGetPointer(PG_DETOAST_DATUM_COPY(PG_GETARG_DATUM(1)));
   RANGE *b = (RANGE *) DatumGetPointer(PG_GETARG_DATUM(1));


   if (ARRISVOID(a))  return FALSE;

   PREPAREARR(a);
   res = inner__range_overlaps_interval_all(a,b);
   pfree(a);
   //pfree(b);
   PG_RETURN_BOOL(res);
}

