/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * blastsvc.c
 *    
 *    This file contains blast service routines used by blast.c
 *
 *
 * This is part of the BLASTgres project. 
 *
 *-------------------------------------------------------------------------
 */

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>

#include "postgres.h"
#include "util.h"
#include "network.h"
#include "blastsvc.h"


/* default settings for BLAST connections */
char	blast_host[] = "www.ncbi.nlm.nih.gov";
char 	blast_program[] = "/blast/Blast.cgi";
int 	blast_server_port = 80;


/* ----------------
 *   strip
 *
 *   This returns a pointer pointing at the first non-space
 *   character in the given string.
 * ----------------
 */

char *strip( char *string )
{
   char *p;

   if ( string==NULL ) return NULL;

   p = string;                                                                             
   while ( *p==' ' )  p++;

   return p;
}

/* ----------------
 *   set_default_blast_param
 *
 *   BLAST web service requires client program to set up
 *   a lot of parameters. We set a number of parameters
 *   by their default values. 
 *
 *   There are just some common parameters and can be 
 *   overwritten by user specified parameter values.
 *
 *   for more details about BLAST parameters, please refer
 *   to BLAST URL API:
 *   http://www.ncbi.nlm.nih.gov/BLAST/Doc/urlapi.html
 *   
 * ----------------
 */

//void set_default_blast_param( query_param* query )
void set_default_blast_param( HttpContext hc )
{
   /* since these parameters do not contain special URL characters,
      we can skip the url_code function safely here. 
      please notice that: this is just default parameters.
      some of these parameters would be overwritten for different
      situations
    */

   HttpContextSetParameter( hc, "CMD", "Put" );
   HttpContextSetParameter( hc, "DATABASE", "nr" );
   HttpContextSetParameter( hc, "SERVICE", "plain" );
   //HttpContextSetParameter( hc, "WORD_SIZE", "11" );
   HttpContextSetParameter( hc, "FILTER", "L" );
   HttpContextSetParameter( hc, "EXPECT", "10" );
   HttpContextSetParameter( hc, "AUTO_FORMAT", "FullAuto" );
   HttpContextSetParameter( hc, "CLIENT", "web" );
   HttpContextSetParameter( hc, "PROGRAM", "blastn" );
}

/* ----------------
 *   record_count  (replaced)
 *
 *   This is used merely to count the number of records 
 *   returned in BLAST hit table report.
 *
 *   Assumption: 
 *   this assumes that BLAST reports are always in the following
 *   form:
 *
 * BLASTN 2.2.10 [Oct-19-2004]
 * # Query: 
 * # Database: nr
 * # Fields: Query id, Subject id, % identity, alignment length, mismatches, gap openings, q. start, q. end, s. start, s. end, e-value, bit score
 * 1_24847	gi|55818351|gb|AY688983.1|	100.00	60	0	0	1	60	301	360	5e-25	 119
 * 1_24847	gi|55818471|gb|AY688964.1|	100.00	59	0	0	1	59	301	359	2e-24	 117
 *
 *
 * ----------------
 */


/* ----------------
 *   send_blast_request
 *
 *   This is the routine in charge of the initiation of BLAST request. 
 *   It sends the request to BLAST service with proper parameters
 *   and retrieve the request id, estimated time (specified in 
 *   BLAST result).
 *
 *   parameters:
 *     query - the BLAST cgi query string
 *     request_id - the address for request_id string
 *     estimated_time - the address for estimated_time string
 *
 *   return:
 *      returns 1 if this request succeeds. otherwise, returns 0.
 *
 * ----------------
 */

//int send_blast_request( HttpContext hc, char *request_id, char *estimated_time )
int SendBlastRequest( HttpContext hc, char *request_id, char *estimated_time )
{
   char *data, *p, *q;
   int ret;

   data = NULL;
   ret = 0;

   if ( HTTP_OK==HttpSendRequest( hc, blast_host, blast_server_port, blast_program ) )
   {
      data = HttpReceiveRequest(hc);

      if ( (p=strstr(data,BLAST_INFO_BEGIN))==NULL )
      {
         /*
          * can not find BLAST_INFO_BEGIN
          * this means two things: an BLAST error or return data is in
          * unknown format.
          */
         if ( (q=strstr(data, BLAST_REQUEST_ERROR)) != NULL )
         {
            /*
             *  if we can find BLAST_REQUEST_ERROR, then it's an
             *  BLAST error
             */
             if ( (p=strchr(q,'<'))!=NULL)
             {
                *p = 0;
                elog (ERROR, "%s", q);
             }
             else  
             {
                // we can not find this, which means BLAST changed format
                elog( ERROR, "BLAST changed format");
             }
         }
         else
         {
             // wrong BLAST format
             elog( ERROR, "BLAST service returned unrecognizable data stream");
         }
      }
      else
      {
         /*
          * we got the correct BLAST result format
          */
          q = strstr( p, BLAST_REQUEST_ID_STRING );
          q += strlen("RID = " );
          p = request_id;
          while ( *q != '\n' ) *p++=*q++;
          *p = 0;

          q = strstr( q, BLAST_ESTIMATED_TIME_STRING );
          q += strlen( "RTOE = " );
          p = estimated_time;
          while ( *q != '\n' ) *p++=*q++;
          *p = 0;

          elog( LOG, "[blast] request_id=[%s], estimated waiting time=%s seconds.", request_id, estimated_time);
          ret = 1;
      } 	
   }
   else
   {
      elog( ERROR, "TCP/IP error" );
   }

   if (data)  pfree(data);
   return ret;
}


/* ----------------
 *   retrieve_hit_table
 *
 *   This is the routine in charge of the retrieval of BLAST hit table
 *   result (with respect to a given request_id and estimated_time)
 *   It keeps polling BLAST server for result and return the 
 *   character pointer pointing to the hit table.
 *
 *   parameters:
 *     request_id - the request_id string
 *     estimated_time - the estimated_time string
 *
 *   return:
 *      returns 1 if this request succeeds. otherwise, returns 0.
 *
 * ----------------
 */

//char* retrieve_hit_table( char *request_id, char* estimated_time )
char *RetrieveBlastResult( char *request_id, char *estimated_time, int type )
{
   FILE *fp;
   HttpContext hc;   
   char *data, *p, *q, *result;
   int time_interval, quit, parsing_required;

   parsing_required = 0;
   hc = CreateHttpContext( POST, 20, NULL );

   HttpContextSetParameter( hc, "CMD", "Get" );
   HttpContextSetParameter( hc, "RID", request_id );
   HttpContextSetParameter( hc, "FORMAT_TYPE", "Text" );
   switch (type)
   {
      case FT_TABULAR:
      case FT_TABULAR_COMMENT:
	 HttpContextSetParameter( hc, "ALIGNMENT_VIEW", "Tabular" );
         parsing_required = 1;
	 break;
      case FT_PAIRWISE:
         HttpContextSetParameter( hc, "ALIGNMENT_VIEW", "Pairwise" );
         break;
      default:
	 HttpContextSetParameter( hc, "ALIGNMENT_VIEW", "Pairwise" );
   }

   HttpContextSetParameter( hc, "FORMAT_OBJECT", "Alignment" );
   //hc = NULL;

   // now, we need to poll BLAST server in fixed time interval. (specified by estimated_time)
   time_interval = atoi(estimated_time);
   result = NULL;
   data = NULL;

   do
   {
      //if (hc) FreeHttpContext(hc);

      //hc = CreateHttpContext( POST, 20, NULL );

      //HttpContextSetParameter( hc, "CMD", "Get" );
      //HttpContextSetParameter( hc, "RID", request_id );
      //HttpContextSetParameter( hc, "FORMAT_TYPE", "Text" );
      //HttpContextSetParameter( hc, "ALIGNMENT_VIEW", "Tabular" );
      //HttpContextSetParameter( hc, "FORMAT_OBJECT", "Alignment" );

      elog( LOG, "sleep for %d seconds", time_interval );
      sleep( time_interval );
      quit = 1;

      if ( HTTP_OK == HttpSendRequest( hc, blast_host, blast_server_port, blast_program ) )
      {
      	 data = HttpReceiveRequest( hc );
         p = strstr( data, "QBlastInfoBegin" );	 
	 if ( p==NULL )   elog( ERROR, "QBlastInfoBegin can not be found!" );
         else
         {
            if ( (p=strstr(p, "Status="))==NULL )  elog( ERROR, "Status= can not be found" );
            else
            {
               p += 7;
               if ( strncmp( p, "READY", 5)==0 )
               {
                  elog( LOG, "status=ready");
		  
                  if (parsing_required)
                  { 
                     p = strstr(data, "<PRE>");
                     if ( p==NULL )  elog(LOG, "ERROR: no <PRE>");
                     else
                     {
                        p=p+5;
                        q = strstr(p,"</PRE>");
                        if ( q==NULL )  elog(LOG, "ERROR: no </PRE>");
                        else
                        {
                          *q = 0;
                          result = pstrdup(p);
                        }
                     }
                  }
                  else
                  {
                     if ( (p=strstr(data,"<pre>"))!=NULL )      p = p+5;
                     else  p = data;

                     result = pstrdup(p);
                     fp = fopen("/tmp/result.plain", "wt");
                     fprintf(fp,"%s",result);
                     fclose(fp);
                  }
               }
               else if ( strncmp(p, "WAITING", 7)==0 )
               {
                  elog( LOG, "Result not ready...wait" );
                  if (  (p=strstr(data,"automatically updated in"))!=NULL)
                  {
                     p = strstr(p,"<b>")+3;
                     q = strstr(p,"</b>");
                     *q = 0;
                     time_interval = atoi(p);
                     quit = 0;
                  }
              }
	      else
              {
                  elog( ERROR, "Uknown blast server status!");
              }
            }            
         }
      }
      else 
      {
         elog( ERROR, "TCP/IP error: can not connect to BLAST server" );
      }
      pfree(data);
      data = NULL;
   } while ( !quit );

   if (data) pfree(data);
   FreeHttpContext(hc);
   return result;
}      


/* ----------------
 *   request_blast_service
 *
 *   This is the main routine to retrieve BLAST hit table. It does
 *   the following things:
 *   1) make a connection to BLAST service
 *   2) request the service with parameters provided by users
 *   3) return a character pointer pointing to the hit table result.
 *
 *   parameters:
 *     seq - the sequence user is blasting against the database
 *     db  - the database user chooses to use
 *     parameters - user specified BLAST parameters
 * 
 *   note:
 *     1. db could be NULL (or any value) and then specified (overwritten) 
 *        by parameters
 *     2. default BLAST parameters are effective if parameters not specified.
 *
 * ----------------
 */

char* request_blast_service( char *seq, char *program, char *db, char **parameters, int type )
{
   FILE *fp;
   HttpContext hc;
   char *result;
   char request_id[50], estimated_time[20];

   result = NULL;
   /* initialize the parameter structure for HTTP POST request */
   hc = CreateHttpContext( POST, 20, NULL );

   /* setup required blast parameters with default values and user specified parameters */
   set_default_blast_param( hc );
   HttpContextSetParameter( hc, "QUERY", seq );

   if ( db!=NULL )
   { 
      HttpContextSetParameter( hc, "DATABASE", db );
   }

   if ( program!=NULL )
   {
      HttpContextSetParameter( hc, "PROGRAM", program );

      if ( strcmp(program,"blastn")==0 )
	 HttpContextSetParameter( hc, "PAGE", "Nucleotides" );
      else if ( strcmp(program,"blastp")==0 )
	 HttpContextSetParameter( hc, "PAGE", "Proteins" );
      else if ( strcmp(program, "tblastx")==0 )
  	 HttpContextSetParameter( hc, "PAGE", "MegaBlast" );
      else 
	 HttpContextSetParameter( hc, "PAGE", "Translations" );

   }

   if ( parameters!=NULL )
   {
      while (*parameters != NULL)
      {
         HttpContextSetParameter( hc, *parameters, *(parameters+1) );
         parameters+=2;
      }     
   }

   CgiParametersList(hc->parameters);
   /* send request to blast server */
   if ( 1==SendBlastRequest(hc,request_id,estimated_time) )
   {
      FreeHttpContext(hc);
      result = RetrieveBlastResult(request_id, estimated_time, type);

      fp = fopen("/tmp/result.txt", "wt");
      fprintf(fp,"%s",result);
      fclose(fp);

   }

   return result;
}

/* ----------------
 *   remote_blast
 *
 *   This is a wrapper to the main routine to retrieve BLAST hit table. 
 *   It allows the user to specify sequence and database parameters.  
 *   Default BLAST parameters are applied automatically.
 * 
 *   parameters:
 *     seq - the sequence user is blasting against the database
 *     program - the blast program (ie. blastn, blastp, ...)
 *     db  - the database user chooses to use
 *
 * ----------------
 */

// used to be called query_blast
//char* query_blast( char *seq, char *program,  char *db )
char* remote_blast( char *seq, char *program, char *db, int type)
{
   return request_blast_service( seq, program, db, NULL, type );
}

/* ----------------
 *   remote_blast_v
 *
 *   This is a wrapper to the main routine to retrieve BLAST hit table.
 *   It allows the user to specify BLAST parameters through the parameters
 *   argument. Default BLAST parameters are applied automatically.
 *
 *   parameters:
 *     seq - the sequence user is blasting against the database
 *     parameters - user specified BLAST parameters
 *
 *   note:
 *     parameters argument must be restricted to the following form:
 *     - parameters point to even number of strings. strings in the
 *       odd number entries represent the BLAST parameter name and
 *       strings in the immediately following entries represent
 *       BLAST parameter values.
 *       
 * ----------------
 */

//char* query_blast_v( char *seq, char **parameters)
// used to be called:  query_blast_v

char* remote_blast_v( char *seq, char **parameters, int type )
{
   return request_blast_service( seq, NULL, NULL, parameters, type );
}


int HspPairCount( const char *report )
{
   char *index, *p;
   int count;

   count = 0;
   index = (char *) report;
   while ( (p=strstr(index, "Score = ")) != NULL)
   {
      index = p + strlen("Score = ");
      count ++;
   }

   return count;
}

int HitTablePairCount( const char *report )
{
   char *p, *prev;
   int count;

   if ( report==NULL ) return 0;
   count = 0;

   if ( (p=strstr( report, HIT_TABLE_PREFIX ))==NULL )  return 0;

   prev = strchr(p,'\n')+1;

   while ( (p=strchr(prev,'\n'))!=0 )
   {
      if ( (int) (p-prev) < 10 ) break;
      prev = p+1;
      count++;
   }

   return count;
}

/*
 *   This function counts the number of alignments reported from
 *   megablast program. megablast could generate other formats
 *   but this function is used to process the default report generated
 *   from megablast. It looks like
 *
 *   'AU108953'=='+AU108953' (1 1 322 322) 0
 *   'AU109042'=='+AU108953' (18 14 324 319) 5
 *   'AU109359'=='+AU108953' (12 58 282 322) 13
 *   'AU109359'=='+AU108953' (12 58 282 322) 13
 *   'AU110478'=='+AU108953' (25 21 240 236) 8
 *  
 *   this function is easy and obviously does not need a complicated coding
 *   like the following, but I just want to make sure I detect as many
 *   exceptions as possible. (C does not have a good exception handling mechanism)
 */

int MegaBlastPairCount( char *report )
{
   char *p, *q;
   int count;

   if ( report==NULL ) return 0;
   count = 0;

   p = report;
   while (1)
   {
      if ( strchr(p,'=')==NULL )  break;
      if ( (q=strchr(p,'\n'))==NULL )  break;
      p = q+1;
      count++;
   }

   return count;
}


/*
 *   call this function when you're expecting a new entry
 *   staring with " Score = ... bits "
 */
HspEntryPointer HspParseEntry( HspEntryPointer entry )
{
   char *line, *eol;
   int qstart, qend, sstart, send;
   int length;
   int ident1, ident2, percent;
   int start1, end1, start2, end2;
   char buffer1[1024], buffer2[256];
   float score;

   line = entry->tracker;
   qstart = -1;
   // just to shut the compiler up
   qend = sstart = send = -1;

   while (1)
   {
      eol = strchr(line, '\n');
      *eol = '\0';

      if ( (strncmp( line, " Score = ", 9)==0) && (qstart==-1) )
      {
         sscanf( line, " Score = %f bits (%d), Expect = %s ", &score, &length, buffer1);
         entry->score = score;
         entry->expect = atof(buffer1);
         entry->length = length;
      }
      else if ( strncmp( line, " Identities = ",14)==0 )
      {
         sscanf( line, " Identities = %d/%d (%d) ", &ident1, &ident2, &percent );
         entry->identity = percent /100.0;
      }
      else if ( strncmp( line, " Strand = ",10)==0 )
      {
         sscanf( line, " Strand = %s / %s ", buffer1, buffer2 );

         if ( buffer1[0]=='P' )
            entry->strands[0] = '+';
         else
            entry->strands[0] = '-';

         if ( buffer2[0]=='P' )
            entry->strands[1] = '+';
         else
            entry->strands[1] = '-';
      }
      else if ( strncmp( line, "Query:",6)==0 )
      {
         sscanf( line, "Query: %d  %s %d ", &start1, buffer1, &end1 );
         *eol='\n';
         line = strchr(eol+1,'\n')+1;
         eol = strchr(line,'\n');
         *eol = '\0';
         sscanf( line, "Sbjct: %d %s %d ", &start2, buffer2, &end2);
         if ( qstart==-1 )
         {
            qstart = start1;
            sstart = start2;
         }
         qend = end1;
         send = end2;
      }
      else if ( (line[0]=='>') ||
                (strncmp( line, " Score = ",9)==0) ||
                (strncmp( line, "Lambda", 6)==0) )
      {
         *eol = '\n';
         eol = line -1;
         break;
      }
      *eol = '\n';
      line = eol+1;
   }

   *eol = '\n';

   entry->tracker = eol+1;

   if ( entry->strands[0]=='+' )
   {
      entry->ranges[0][0] = qstart;
      entry->ranges[0][1] = qend;
   }
   else
   {
      entry->ranges[0][0] = qend;
      entry->ranges[0][1] = qstart;
   }

   if ( entry->strands[1]=='+' )
   {
      entry->ranges[1][0] = sstart;
      entry->ranges[1][1] = send;
   }
   else
   {
      entry->ranges[1][0] = send;
      entry->ranges[1][1] = sstart;
   }


   return entry;
}

HspEntryPointer HspFindNext( HspEntryPointer entry )
{
   char *line, *p;
   char buffer1[1024], buffer2[1024];
   line = entry->tracker;

   while (1)
   {
      if ( line[0]=='>' )
      {
         sscanf( line+1, "%s %s ", buffer1, buffer2 );
         strcpy( entry->subject, buffer1 );

         //printf("[DEBUG] subject=[%s]\n", entry->subject);

         p = strchr(line, '\n')+1;
         while (  strncmp(p, " Score =", 8)!=0 )
           p = strchr( p, '\n') + 1;
         line = p;
      }
      else if ( strncmp( line, " Score = ", 9)==0 )
      {
         break;
      }
      else
      {
         line = strchr(line,'\n');
         if (line==NULL)  return NULL;
         else line = line+1;
      }
   }

   entry->tracker = line;
   return HspParseEntry(entry);
}

HspEntryPointer HspFindFirst( char *report, char *seqid )
{
   char *p, *q, *r;
   HspEntryPointer entry;
   char buffer[1024];

   //FILE *fp;

   //fp = fopen("/tmp/result.txt", "wt");
   //fprintf(fp, "%s\n", report );
   //fclose(fp);

   p = strstr( report, "Query= " );
   if ( p==NULL )
       return NULL;

   r = strchr(p,'\n');
   *r = '\0';
   if ( (q=strchr(p+7,' '))!=NULL )  *q = '\0';

   /*
    * notice that if you submit a sequence to web interface, BLAST will never
    * give you an identifier (even temporary one), so this query identifier 
    * field will be empty. However, it is illegal to have an empty identifier
    * for BLASTgres, so we give it a fake name.
    */

   if ( seqid[0]!=0 )
      strcpy(buffer, seqid);
   else if ( strlen(p+7)==0 )
      strcpy(buffer,"none");
   else   
      strcpy(buffer, p+7);

   if ( q!=NULL ) *q = ' ';
   *r = '\n';


   p = strstr( report, "Sequences producing significant alignments:" );
   if ( p==NULL )
       return NULL;

   q = strchr(p+20, '\n');
   q = strchr(q+1, '\n');

   entry = (HspEntry *) palloc(sizeof(HspEntry));

   if ( entry==NULL )  return NULL;
   entry->source = report;
   entry->query = pstrdup(buffer);

   p = strstr(p+1, "ALIGNMENTS");
   if ( p==NULL )
   {
      // the reason why we need to do another string search is
      // because sometimes BLAST report consists of
      // ALIGNMENTS  ... without >identifier
      // or
      // [no ALIGNMENTS] ... with >identifier
      p = strchr(q,'>');
      if ( p==NULL ) return NULL;
      entry->tracker = p;
      return HspFindNext( entry );

   }
   else
   {
      p = strchr(p,'\n')+1;
      // actually the above is not true... sometimes (from the website), it has ALIGNMENTS and >identifier
      // so we do it here (the following line)
      //if ( *p=='>' ) p++;
      entry->tracker = p;

      //return HspParseEntry( entry );
      return HspFindNext(entry);
   }
}


void HspEntryFree( HspEntryPointer entry )
{
   if (entry->query)
      pfree(entry->query);
}

void HitTableEntryFree( HitTableEntryPointer entry )
{
}

HitTableEntryPointer HitTableParseEntry( HitTableEntryPointer entry )
{
   char *line, *eol;
   char buffer[50];
   int tmp;

   line = entry->tracker;

   eol = strchr( line, '\n');
   if ( eol==NULL ) return NULL;
   tmp = sscanf( line, "%s %s %f %d %d %d %d %d %d %d %f %f", buffer, entry->subject,
                 &(entry->identity), &(entry->length), &(entry->mismatches),
                 &(entry->gap), &(entry->ranges[0][0]), &(entry->ranges[0][1]),
                 &(entry->ranges[1][0]), &(entry->ranges[1][1]),
                 &(entry->expect), &(entry->score) );

   if ( tmp<4 ) return NULL;

   if ( entry->query[0]==0 )
     strcpy(entry->query,buffer);

   entry->strands[0] = entry->strands[1] = '+';
   if ( entry->ranges[0][0]>entry->ranges[0][1] )
   {
       tmp = entry->ranges[0][0];
       entry->ranges[0][0] = entry->ranges[0][1];
       entry->ranges[0][1] = tmp;
       entry->strands[0] = '-';
   }
   if ( entry->ranges[1][0]>entry->ranges[1][1] )
   {
       tmp = entry->ranges[1][0];
       entry->ranges[1][0] = entry->ranges[1][1];
       entry->ranges[1][1] = tmp;
       entry->strands[1] = '-';
   }

   entry->tracker = eol+1;

   return entry;
}

HitTableEntryPointer HitTableFindNext( HitTableEntryPointer entry )
{
   return HitTableParseEntry(entry);
}

HitTableEntryPointer HitTableFindFirst( char *report, char *seqid )
{
   char *p;
   HitTableEntryPointer entry;

   FILE *fp;

   fp = fopen("/tmp/result.txt", "wt");
   fprintf(fp, "%s\n", report );
   fclose(fp);


   p = strstr( report, HIT_TABLE_PREFIX );
   if ( p==NULL )  return NULL;

   p = strchr(p+20, '\n');
   entry = (HitTableEntry *) palloc(sizeof(HitTableEntry));


   if ( seqid[0]!=0 )
      strcpy( entry->query, seqid );
   else
      entry->query[0] = 0;

   entry->source = report;
   entry->tracker = p+1;

   return HitTableParseEntry( entry );
}

void MegaBlastEntryFree( MegaBlastEntryPointer entry )
{
}

MegaBlastEntryPointer MegaBlastFindNext( MegaBlastEntryPointer entry )
{
   return MegaBlastParseEntry(entry);
}

MegaBlastEntryPointer MegaBlastFindFirst( char *report, char *seqid )
{
   MegaBlastEntryPointer entry;

   if ( report==NULL )  return NULL;

   entry = (MegaBlastEntry *) palloc(sizeof(MegaBlastEntry));

   if ( seqid[0]!=0 )
      strcpy( entry->query, seqid );

   entry->source = report;
   entry->tracker = report;

   return MegaBlastParseEntry( entry );
}

MegaBlastEntryPointer MegaBlastParseEntry( MegaBlastEntryPointer entry )
{
   char *line, *eol, *p;
   int tmp;

   line = entry->tracker;
   eol = strchr(line,'\n');
   *eol = '\0';

   if ( (p=strstr(line, "\'=="))==NULL )
      elog( ERROR, "report for megablast has changed!" );

   *p = '\0';
   while ( (*line=='\'') || (*line=='>') ) line++;
   strcpy( entry->subject, line );

   if ( (line=strchr(p+1,'\''))==NULL )
      elog( ERROR, "report for megablast has changed!" );

   line++;
   entry->strands[1] = *line;
   entry->strands[0] = '+';

   if ( (p=strchr(line,'\''))==NULL )
      elog( ERROR, "report for megablast has changed!");

   *p = 0;

   if ( entry->query[0]==0 )
      strcpy( entry->query, line+1);

   tmp = sscanf( p+1, " ( %d %d %d %d ) %f ",
                 &(entry->ranges[1][0]), &(entry->ranges[0][0]),
                 &(entry->ranges[1][1]), &(entry->ranges[0][1]),
                 &(entry->score) );

   entry->tracker = eol+1;

   return entry;
}


/*
 *  blast_with_file is the ultimate function which actually calls a specific blast program
 *  to query the database. Ideally, 'blastall' takes care of everything. However, megablast
 *  kind of ruins the perfect dispatching of queries. It takes different parameter sets and
 *  it's not dispatched by blastall. So we have to separate megablast query apart from blastall
 */

char *blast_with_file(char *database, char *program, char *file, int type, char **parameters)
{
   char buffer[1024], tmp[40];
   char *ret, *p, *q, **iterator;

   ret = NULL;

   /*  
    *  annother misfortune is that the program can be specified in the parameter list instead
    *  of the argument - program, so we have to retrieve it. So the priority is:
    *  the argument "program" takes the highest priority.
    */
   if ( program==NULL )
   {
       /* if the program argument is NULL, it indicates that the "program" parameter has
        * be specified in the parameters; otherwise, we are in a deep trouble 
        */

       iterator = parameters;
       while (*iterator != NULL )
       {
          if ( strcasecmp(*iterator,"program")==0 || strcmp(*iterator,"p")==0 )
          {
             program = *(iterator+1);
             break;
          }
          iterator += 2;
       }
       if ( program==NULL )  elog( ERROR, "blast program not specified" );
       return NULL;
   }

   /* now, we prepare two different command lines for normal blast search and megablast */
   if ( strcasecmp(program, "megablast")==0 )
   {
      sprintf( buffer, "megablast -D %d -i %s ", type, file );
      if (database!=NULL)  sprintf( buffer, "%s -d %s ", buffer, database );
   }
   else
   {
      sprintf( buffer, "blastall -m %d -i %s ", type, file );
      if (database!=NULL)   sprintf( buffer, "%s -d %s ", buffer, database );
      sprintf( buffer, "%s -p %s ", buffer, program );
   }

   if (parameters!=NULL)
   {
      while (*parameters != NULL)
      {
         sprintf( tmp, "-%s %s ", *parameters, *(parameters+1));
         strcat( buffer, tmp );
         parameters+=2;
      }
   }


   ret = __exec_program(buffer);

   if ( (ret!=NULL) && (p=strstr(ret,"ERROR")) )
   {
      q = strchr(p,'\n');
      *q = 0;
      pfree(ret);
      ret = NULL;
   }


   return ret;
}

char *blast_with_seq(char *database, char *program, char *seq, char *seqid, int type, char **parameters)
{
   FILE *fp;
   char filename[1024], *ret;


   sprintf(filename, "/tmp/BLASTgres-sequence-%d", getpid() );

   fp = fopen( filename, "wt" );

   fprintf(fp, ">%s seq\n%s", seqid, seq );
   fclose(fp);

   ret = blast_with_file(database, program, filename, type, parameters);

   remove(filename);
   return ret;
}

FastaSequencePointer FormFastaRecord( char *data )
{
   FastaSequencePointer ret;
   char *p, *q;
   int length;

   ret = (FastaSequencePointer) palloc(sizeof(FastaSequence));
   ret->identifier = NULL;
   ret->description = NULL;
   ret->sequence = NULL;

   if ( (q=strchr(data, '\n'))==NULL )  return ret;
   if ( (p=strchr(data,' '))!=NULL )
   {
      *p = 0;
      ret->identifier = pstrdup(data+1);
      *q = 0;
      ret->description = pstrdup(p+1);
   }
   else
   {
      *q = 0;
      ret->identifier = pstrdup(data+1);
      ret->description = pstrdup(" ");
   }

   p = q;

   length = strlen(p+1);
   ret->sequence = palloc(length+1);
   q = ret->sequence;
   p = p+1;
   while ( *p!='\0' )
     if ( isalpha(*p) )   *q++=*p++;  
     else p++;

   *q = '\0';


   return ret;
}

FastaSequencePointer ReadFastaSequenceFromFile( char* filename )
{
   FILE *fp;
   FastaSequencePointer ret;
   int length;
   char *data;


   fp = fopen(filename,"rt");
   if ( fp==NULL)  
   {
      return NULL;
   }

   fseek( fp, 0, SEEK_END );
   length = ftell(fp);
   fseek( fp, 0, SEEK_SET );

   data = (char *) palloc(length+1);
   fread(data, sizeof(char), length, fp);
   data[length] = 0;
   ret = FormFastaRecord(data);
   pfree(data);

   return ret;
}

FastaSequencePointer ReadFastaSequenceFromDB( char *dbname, char *seqid, int start, int end )
{
   FastaSequencePointer ret;
   char buffer[1024];
   char *data, *p, *q;
   int length;

   sprintf( buffer, "fastacmd -d %s -s %s -L %d,%d", dbname, seqid, start, end);

   ret = (FastaSequencePointer) NULL;
   data = __exec_program( buffer);

   if ( data!=NULL )
   {
      /* handle error condition */
      if ( (p=strstr( data, "ERROR:" )) )
      {
         elog( NOTICE, "%s", p );
         pfree(data);
      }
      else
      {
         ret = (FastaSequencePointer) palloc(sizeof(FastaSequence));
         ret->identifier = NULL;
         ret->description = NULL;
         ret->sequence = NULL;

         p = strchr(data,' ');
         *p = '\0';
         ret->identifier = pstrdup(data+1);
         q = p+1;
         p = strchr(q,'\n');
         *p = '\0';

         ret->description = pstrdup(q);

         length = strlen(p+1);
         q = ret->sequence = palloc(length+1);
         p = p+1;
         while ( *p!='\0' )
           if ( *p!='\n' )   *q++=*p++;
           else p++;

         *q = '\0';

         //printf("[DEBUG] description=[%s]\n", ret->description );
         //printf("[DEBUG] sequence=[%s]\n", ret->sequence );

         pfree(data);
      }
   }
   else
   {
      elog( ERROR, "can not find fastacmd program.\n");
   }

   return ret;
}

void FreeFastaSequence( FastaSequencePointer v )
{
   if ( v->identifier ) pfree(v->identifier);
   if ( v->description )  pfree(v->description );
   if ( v->sequence )  pfree(v->sequence );
}


FastaDBInformation* FastaDBInfo( char *dbname )
{
   FastaDBInformation *ret;
   char buffer[1024], tmp[100];
   char *data, *p, *q, *r;
   int number; //number1;

   sprintf( buffer, "fastacmd -d %s -I", dbname );

   ret = (FastaDBInformation *) NULL;
   data = __exec_program( buffer);

   if ( data!=NULL )
   {
      /* handle error condition */
      if ( (p=strstr( data, "ERROR:" )) )
      {
         elog( NOTICE, "%s", p );
         pfree(data);
      }
      else
      {
         if ( (p=strstr(data, "Database:"))!=NULL )
         {
	    ret = (FastaDBInformation *) palloc(sizeof(FastaDBInformation));
            sscanf( p, "Database: %s\n %d sequences; %s total letters", buffer, &number, tmp );            
            ret->dbname = pstrdup(buffer);
            ret->nsequences = number;

            // processing the formatted number 
            q = tmp;
            r = buffer;            
            while (*q)
            {
              if ( isdigit(*q) )  *r++=*q;
              q++;
            }
            *r = 0;

            ret->nletters = atoi(buffer);

            p = strstr(p+1, "File name:");
            sscanf( p, "File name:\n%s\n", buffer);

	    ret->filepath = pstrdup(buffer);

            p = strstr(p+1,"Date:")+5;
            q = strstr(p,"Version");
            *(q-1)=0;
            strcpy(buffer,p);

	    ret->datetime = pstrdup( buffer);

            sscanf( q, "Version: %s  Longest sequence: %d bp", buffer, &number);

	    ret->version = pstrdup(buffer);
            ret->longest = number;
         }
         else
           elog( ERROR, "can not parse database information" );
      } 
   }
   else
   {
      elog( ERROR, "can not find fastacmd program.\n");
   }

   return ret;
}
