/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * blastsvc.h
 *
 *    Definitions of functions for interacting with BLAST server.
 *
 * This file is included in blast.c and all other files which request
 * BLAST service internally.
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */

#ifndef __BLASTSVC_H
#define __BLASTSVC_H

#include "network.h"

#define MAX_ID_LENGTH   1024


/*
 *   HspEntry structure
 *   this structure keeps the parsing context for pair-wise BLAST report.
 *   every HSP (high-scoring pair) is represented as an HspEntry, which
 *   retains the information of each matching aligned pairs.
 *   note: HspEntryFree() is required after the completion of parsing.
 */
typedef struct HspEntry
{
   char *source,                                /* blast data buffer, used internally */
        *tracker,                               /* tracker tracks the progress of the parsing */
        *query;                                 /* this retains the query identifier */
   char subject[MAX_ID_LENGTH];                 /* subject identifier */
   int length;                                  /* the length of the aligned sequence */
   float score,                                 /* e-score */
         identity,                              /* percent identity */
         expect;                                /* expectation value */
   char strands[2];                             /* the strand of each sequence */
   int  ranges[2][2];                           /* the interval of each sequence */
} HspEntry;


typedef struct HitTableEntry
{
   char *source,
        *tracker;
   char query[MAX_ID_LENGTH],
        subject[MAX_ID_LENGTH];
   int length,
       mismatches,
       gap;
   float identity,
         expect,
         score;
   char strands[2];
   int ranges[2][2];
} HitTableEntry;

typedef struct MegaBlastEntry
{
   char *source,
        *tracker;
   char query[MAX_ID_LENGTH],
        subject[MAX_ID_LENGTH];
   float score;
   char strands[2];
   int ranges[2][2];
} MegaBlastEntry;

typedef struct FastaSequence
{
   char *identifier,
        *description,
        *sequence;
} FastaSequence;

typedef struct FastaDBInformation
{
   char *dbname;
   int nsequences, nletters, longest;
   char *filepath, *datetime, *version;
} FastaDBInformation;

typedef HspEntry *HspEntryPointer;
typedef HitTableEntry *HitTableEntryPointer;
typedef MegaBlastEntry *MegaBlastEntryPointer;
typedef FastaSequence *FastaSequencePointer;


#define HspEntryQueryId(entry)      ((entry)->query)
#define HspEntrySbjctId(entry)      ((entry)->subject)
#define HspEntryScore(entry)        ((entry)->score)
#define HspEntryIdentity(entry)     ((entry)->identity)
#define HspEntryExpect(entry)       ((entry)->expect)
#define HspEntryQueryStrand(entry)  ((entry)->strands[0])
#define HspEntrySbjctStrand(entry)  ((entry)->strands[1])
#define HspEntryQueryLower(entry)   ((entry)->ranges[0][0])
#define HspEntryQueryUpper(entry)   ((entry)->ranges[0][1])
#define HspEntrySbjctLower(entry)   ((entry)->ranges[1][0])
#define HspEntrySbjctUpper(entry)   ((entry)->ranges[1][1])

#define HitTableEntryQueryId(entry)    ((entry)->query)
#define HitTableEntrySbjctId(entry)    ((entry)->subject)
#define HitTableEntryScore(entry)      ((entry)->score)
#define HitTableEntryIdentity(entry)   ((entry)->identity)
#define HitTableEntryExpect(entry)     ((entry)->expect)
#define HitTableEntryMismatches(entry) ((entry)->mismatches)
#define HitTableEntryGap(entry)        ((entry)->gap)
#define HitTableEntryLength(entry)     ((entry)->length)
#define HitTableEntryQueryStrand(entry)  ((entry)->strands[0])
#define HitTableEntrySbjctStrand(entry)  ((entry)->strands[1])
#define HitTableEntryQueryLower(entry)   ((entry)->ranges[0][0])
#define HitTableEntryQueryUpper(entry)   ((entry)->ranges[0][1])
#define HitTableEntrySbjctLower(entry)   ((entry)->ranges[1][0])
#define HitTableEntrySbjctUpper(entry)   ((entry)->ranges[1][1])

#define MegaBlastEntryQueryId(entry)    ((entry)->query)
#define MegaBlastEntrySbjctId(entry)    ((entry)->subject)
#define MegaBlastEntryScore(entry)      ((entry)->score)
#define MegaBlastEntryQueryStrand(entry)  ((entry)->strands[0])
#define MegaBlastEntrySbjctStrand(entry)  ((entry)->strands[1])
#define MegaBlastEntryQueryLower(entry)   ((entry)->ranges[0][0])
#define MegaBlastEntryQueryUpper(entry)   ((entry)->ranges[0][1])
#define MegaBlastEntrySbjctLower(entry)   ((entry)->ranges[1][0])
#define MegaBlastEntrySbjctUpper(entry)   ((entry)->ranges[1][1])

#define HIT_TABLE_PREFIX   "# Fields" 
#define BLAST_INFO_BEGIN   "<!--QBlastInfoBegin" 
#define BLAST_REQUEST_ERROR "ERROR:"
#define BLAST_REQUEST_ID_STRING  "RID"
#define BLAST_ESTIMATED_TIME_STRING "RTOE" 


/* exported functions */
int	record_count( char* );
char*	remote_blast( char *, char *, char*, int);
char*   remote_blast_v( char *, char **, int );
char*	request_blast_service( char*, char*, char*, char**, int );



/* functions used internally */
char*	strip( char * );
void 	set_default_blast_param( HttpContext hc );
int SendBlastRequest( HttpContext hc, char *request_id, char *estimated_time);
char *RetrieveBlastResult(char *, char *, int);

/* new functions */
int HspPairCount( const char *report );
int HitTablePairCount( const char *report );
int MegaBlastPairCount( char *report );

HspEntryPointer HspParseEntry( HspEntryPointer entry );
HspEntryPointer HspFindNext( HspEntryPointer entry );
HspEntryPointer HspFindFirst( char *report, char *seqid );
void HspEntryFree( HspEntryPointer entry );

HitTableEntryPointer HitTableParseEntry( HitTableEntryPointer entry );
HitTableEntryPointer HitTableFindFirst( char *report, char *seqid );
HitTableEntryPointer HitTableFindNext( HitTableEntryPointer entry );
void HitTableEntryFree( HitTableEntryPointer entry );

MegaBlastEntryPointer MegaBlastParseEntry( MegaBlastEntryPointer entry );
MegaBlastEntryPointer MegaBlastFindFirst( char *report, char *seqid );
MegaBlastEntryPointer MegaBlastFindNext( MegaBlastEntryPointer entry );
void MegaBlastEntryFree( MegaBlastEntryPointer entry );

FastaSequencePointer FormFastaRecord( char *data );
FastaSequencePointer ReadFastaSequenceFromFile( char *file);
FastaSequencePointer ReadFastaSequenceFromDB( char *dbname, char *seqid, int start, int end );
void FreeFastaSequence( FastaSequencePointer seq );
FastaDBInformation *FastaDBInfo( char *dbname );
char *__exec_program( const char *cmd );
char *blast_with_file( char *database, char *program, char *file, int type, char **parameters);
char *blast_with_seq( char *database, char *program, char *seq, char *seqid, int type, char **parameters);

int CheckFileExistence( char *file );

#define FT_PAIRWISE                 			0
#define FT_QUERY_ANCHORED_WITH_IDENTITY    		1
#define FT_QUERY_ANCHORED_WITHOUT_IDENTITY      	2
#define FT_FLAT_QUERY_ANCHORED_WITH_IDENTITY 		3
#define FT_FLAT_QUERY_ANCHORED_WITHOUT_IDENTITY 	4
#define FT_QUERY_ANCHORED_WITHOUT_IDENTITY_BLUNT  	5
#define FT_FLAT_QUERY_ANCHORED_WITHOUT_IDENTITY_BLUNT	6
#define FT_XML						7
#define FT_TABULAR					8
#define FT_TABULAR_COMMENT				9
#define FT_ASN_TEXT					10
#define FT_ASN_BINARY					11
#define MEGABLAST					0



#endif
