/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

#include "postgres.h"

#include <math.h>
#include <string.h>
#include <stdio.h>

#include "fmgr.h"
#include "funcapi.h"
#include "executor/spi.h"
#include "miscadmin.h"
#include "utils/builtins.h"
#include "utils/guc.h"
#include "utils/lsyscache.h"
#include "lib/stringinfo.h"

#include "coalesce.h"
#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))

static interval* create_list();
static void create_node(interval*, int, int);
static void expand(interval*, int32);
static void add_interval(interval*, int, int, int32);
static void free_list(interval*);

static interval *tail = 0;

static interval* create_list(void)
{
  interval *head;

  head = (interval*) malloc( sizeof(interval) );

  /* lazy initilization */
  if ( tail== 0 ) 
    tail = (interval*) malloc( sizeof(interval) );

  head->lower = head->upper = -20000;  /* yet...another magic number */
  head->next = tail;
  tail->next = tail;

  return head;
}

static void create_node( interval *p, int lower, int upper )
{
  interval *q;

  q = (interval*) malloc( sizeof(interval) );
  q->lower = lower;
  q->upper = upper;
  q->next = p->next;
  p->next = q;
}

static void expand( interval *p, int32 gap_size )
{
  interval *q = p->next;

  while ( q!=tail )
  {
    if ( q->lower <= (p->upper+gap_size) )
    {
      p->upper = p->upper>q->upper ? p->upper : q->upper;
      p->next = q->next;
      free(q);
      q = p->next;
    }
    else break;
  }
}

//void add_interval( interval* list, int64 lower, int64 upper, int32 gap_size )
static void add_interval( interval* list, int lower, int upper, int32 gap_size)
{
  interval *p = list->next, *q = list;

  for(;p!=tail;q=p,p=p->next)
  {
    if ( lower > (p->upper+gap_size) ) continue;
    if ( upper < (p->lower-gap_size) ) break;

    p->lower = lower<p->lower ? lower : p->lower;
    p->upper = upper<p->upper ? p->upper : upper;
    expand(p, gap_size);
    return;

  }

  create_node(q, lower, upper);
}

static void free_list( interval* head )
{
  interval *q, *p = head;

  while ( p!=tail )
  {
    q = p->next;
    free(p);
    p = q;
  }
}


PG_FUNCTION_INFO_V1(coalesce);
Datum coalesce(PG_FUNCTION_ARGS)
{
  ReturnSetInfo *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
  TupleDesc tupdesc;
  AttInMetadata *attinmeta;
  MemoryContext per_query_ctx;
  MemoryContext oldcontext;
  HeapTuple tuple, spi_tuple;
  SPITupleTable *tuptable;
  TupleDesc spi_tupdesc;
  char *pointer, buffer[40];
  char **values;
  int i;
  interval *head;

  char *relation_name = GET_STR( PG_GETARG_TEXT_P(0) );
  char *column_name = GET_STR( PG_GETARG_TEXT_P(1) );
  char *predicate = GET_STR( PG_GETARG_TEXT_P(2) );
  int32 allowed_gap_size;
  Tuplestorestate *tupstore = NULL;
  int ret, proc;
  MemoryContext context_backup;
  StringInfo sql;
  interval *p;

  if ( PG_ARGISNULL(3) )
    allowed_gap_size = 0;
  else
    allowed_gap_size = PG_GETARG_UINT32(3);  

  head = create_list();   

  /* check to see if caller supports us returning a tuplestore */
  if ( !rsinfo || !(rsinfo->allowedModes & SFRM_Materialize) )
    elog(ERROR, "coalesce: materialized mode required, but it is not allowed in this context" );


  per_query_ctx = rsinfo->econtext->ecxt_per_query_memory;
  oldcontext = MemoryContextSwitchTo(per_query_ctx);

  /* get the requested return tuple description */
  tupdesc = CreateTupleDescCopy(rsinfo->expectedDesc);

  attinmeta = TupleDescGetAttInMetadata(tupdesc);

  /* check to see if caller supports us returning a tuplestore */
  if (!rsinfo->allowedModes & SFRM_Materialize)
    elog( ERROR, "coalesce requires Materialized mode, but it is not allowed in this context");

  rsinfo->returnMode = SFRM_Materialize;

  /* real job starts here */

  /* connect to SPI manager */
  if ( (ret=SPI_connect()) < 0)
    elog(ERROR, "coalesce: SPI_connect return %d", ret);

  context_backup = MemoryContextSwitchTo(per_query_ctx);

  /* initialize our tuplestore */
  tupstore = tuplestore_begin_heap(true, false, work_mem);
  MemoryContextSwitchTo(context_backup);

  /* now go get the whole table */
  sql = makeStringInfo();
  tupdesc = attinmeta->tupdesc;

  /* Build initial sql statement */
  appendStringInfo(sql, "SELECT %s FROM %s WHERE %s ORDER BY %s", 
    column_name, relation_name, predicate, column_name);


  /* retrieve the desired rows */
  ret = SPI_exec(sql->data, 0);
  proc = SPI_processed;


  /* check for qualifying tuples */
  if ( (ret==SPI_OK_SELECT) && (proc>0) )
  {
    tuptable = SPI_tuptable;
    spi_tupdesc = tuptable->tupdesc;

    for(i=0;i<proc; i++)
    {
      bool isnull;
      spi_tuple = tuptable->vals[i];

      // the following is a demonstration of string values
      //pointer = SPI_getvalue(spi_tuple, spi_tupdesc, 1);
      pointer = ( char *) SPI_getbinval( spi_tuple, spi_tupdesc, 1, &isnull);

      if ( !isnull )
      { 
        Intseg *seg = (Intseg*) pointer;

        add_interval( head, seg->lower, seg->upper, allowed_gap_size ); 
      }
    }
  }

  p = head->next;

  while ( p != tail )
  {
    values = (char**) palloc(sizeof(char*));
    values[0] = buffer;

    sprintf( buffer, "%lld..%lld", p->lower, p->upper );
    tuple = BuildTupleFromCStrings(attinmeta, values );
    context_backup = MemoryContextSwitchTo(per_query_ctx);
    tuplestore_puttuple(tupstore, tuple);
    MemoryContextSwitchTo(context_backup);

    pfree(values);
    heap_freetuple(tuple);

    p = p->next;
  }

  free_list( head );

  SPI_finish();

  context_backup = MemoryContextSwitchTo(per_query_ctx);
  tuplestore_donestoring(tupstore);
  MemoryContextSwitchTo(context_backup);

  rsinfo->setResult = tupstore;

  /* real job ends here */

  rsinfo->setDesc = tupdesc;
  MemoryContextSwitchTo(oldcontext);

  return (Datum) 0;
}


