/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * genbank.c
 *
 *    This file contains genbank-related routines called by Postgresql.
 *
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */


#include "postgres.h"
#include "fmgr.h"
#include "funcapi.h"
#include "executor/spi.h"
#include "lib/stringinfo.h"
#include "miscadmin.h"
#include "utils/builtins.h"
#include "utils/guc.h"
#include "utils/lsyscache.h"
#include <string.h>

#include "genbank.h"
#include "genbanksvc.h"

/* some convenient macro to retrieve pointers for text */
#define GET_TEXT(cstrp) DatumGetTextP(DirectFunctionCall1(textin, CStringGetDatum(cstrp)))
#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))


/* ----------------
 *   form_genbank_tuple
 *
 *   This is the routine used by Genbank-related table function routines
 *   to form a tuple based on the genbank file returned by Genbank service.
 *   A lot of Postgresql-related types and functions are used in
 *   order to maintain memory context and manipulate tuples, please
 *   refer to Postgresql programmer manual for their definitions.
 *
 *   parameters:
 *     result - this is a character string contains a tuple of three elements.
 *              locus, accession, and genomic sequence. These three elements
 *              are separated by comma. 
 *     slot - Postgresql TableSlot object.
 *     attinmeta - Postgresql AttInMetadata (metadata for tuple attributes)
 *
 *   return:
 *     construct a tuple based on the genbank file.
 *
 *   note:
 *     this routine relies on the definition of "genbank_result" type in Postgresql and
 *     assume a particular format from BLAST hit table result. For more,
 *     see the comments in genbank_search() or genbank_retrieve();
 *
 * ----------------
 */

Datum form_genbank_tuple( GenbankRecordPointer result, TupleTableSlot *slot, AttInMetadata *attinmeta )
{
   Datum retval;
   HeapTuple tuple;
   char **values;
   int i;

   /* allocate enough memory to hold values in the tuple */
   values = (char **) palloc( MAX_GENBANK_FIELD_NO*sizeof(char *));

   for(i=0;i<MAX_GENBANK_FIELD_NO;i++)
   {
      values[i] = pstrdup(result->fields[i]);
   }

   /* build a tuple */
   tuple = BuildTupleFromCStrings(attinmeta, values);

   /* make the tuple into a datum */
   retval = TupleGetDatum(slot,tuple);

   /* free the memory */
   for(i=0;i<MAX_GENBANK_FIELD_NO;i++)  pfree(values[i]);
   pfree(values);

   return (retval);
}

/* ----------------
 *   genbank_search
 *
 *   This is the routine for retrieving genomic sequence informaton from 
 *   Genbank service. 
 *  
 *   A lot of Postgresql-related types and functions are used in
 *   order to maintain memory context and manipulate tuples, please
 *   refer to Postgresql programmer manual for their definitions.
 *
 *   parameters:
 *     seqid - specify a genomic sequence by its gene id, accession or other 
 *             unique identifier used in entrez database.
 *    
 *   return:
 *     construct a tuple based on the genbank result.
 *
 *   note:
 *     this routine relies on the definition of "genbank_result" type in Postgresql.
 *     this type must be defined in advance and its structure should be
 *     conformed to the following:
 *
 *     CREATE TYPE genbank_result AS
 *     (
 *        locus text,
 *        accession text,
 *        sequence text,
 *      );
 *
 * ----------------
 */


PG_FUNCTION_INFO_V1(genbank_search);
Datum genbank_search(PG_FUNCTION_ARGS)
{
   Datum retval;
   TupleDesc tupdesc;
   TupleTableSlot *slot;
   AttInMetadata *attinmeta;
   text *arg;
   char *term, *data;
   GenbankRecordPointer result;

   /* to ensure the input arguments are correct, we need to check if it's null */
   if (PG_ARGISNULL(0))  PG_RETURN_NULL();

   /* the sequence we are going to blast against */
   arg = PG_GETARG_TEXT_P(0);
   term = (char*) palloc(VARSIZE(arg)-VARHDRSZ);
   strcpy( term, GET_STR(arg) );                 

   if ( (data=genbank_entrez(term))==NULL ) PG_RETURN_NULL();

   if ( (result=genbank_retrieve(data,0))==NULL)
   {
      pfree(data);
      PG_RETURN_NULL();
   }

   pfree(data);

   /* build a tuple description for a BlastHitTable tuple */
   tupdesc = RelationNameGetTupleDesc("genbank_record");

   /* allocate a slot for a tuple with this tupdesc */
   slot = TupleDescGetSlot(tupdesc);

   /* generate attribute metadata needed later to produce tuples from raw c strings */
   attinmeta = TupleDescGetAttInMetadata(tupdesc);

   retval = form_genbank_tuple( result, slot, attinmeta );

   FreeGenbankRecord(result);

   return retval;
}


