/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * genbanksvc.c
 *
 *    This file contains genbank service routines used by genbank.c
 *
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */


#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include "postgres.h"
#include "network.h"
#include "genbanksvc.h"

/* default settings for BLAST connections */
//char http_buffer[DEFAULT_HTTP_BUFFER_SIZE];
char genbank_databases[MAX_GENBANK_DB][11] = { "Nucleotide", "Protein" };
char genbank_host[] = "www.ncbi.nlm.nih.gov";
char genbank_query[] = "/entrez/query.fcgi";
char genbank_retrieval[] = "/entrez/viewer.fcgi";
int genbank_server_port = 80;

char* GetNextLine( char *src );
char* GetNextLine( char *src )
{
   char *p;
   p = strchr(src,'\n');

   if (p!=NULL) return p+1;
   else         return NULL;
}

char* GetNextString( char *src );
char *GetNextString( char *src )
{
   char *p=src;

   while ( isspace(*p) ) p++;
   return p;
}

void FreeGenbankRecord( GenbankRecordPointer gb )
{
   int i;
   if (gb)
   {
      for(i=0;i<MAX_GENBANK_FIELD_NO;i++)
	 if (gb->fields[i])  pfree(gb->fields[i]);
   }
}


/* ----------------
 *   set_default_genbank_param
 *
 *   GenBank web service requires client program to set up
 *   some parameters. We set a number of parameters
 *   by their default values.
 *
 *   There might not be any official documents about Genbank
 *   search parameters. I extracted these parameters from the
 *   search web page. It is subject to change!!! 
 *
 *   Genbank search
 *   http://www.ncbi.nlm.nih.gov/entrez/
 *
 * ----------------
 */

void set_default_genbank_param( HttpContext hc )
{
  /* some default parameters. these could be overwritten by configure file settings */
  HttpContextSetParameter( hc, "DB", "Nucleotide" );
  HttpContextSetParameter( hc, "CMD", "Put" );
}


/* ----------------
 *   genbank_entrez
 *
 *   this is the routine to search "any terms" in entrez database
 *   it is powerful but it is not suitable for database environment.
 *   because database environment needs rigidly-structured, record-like
 *   data.
 *   This is the entry to search for gene (specified by accession, or gene
 *   id or even gene name). It will return a lot of relevant entries but
 *   if a particular gene is searched, the first entry will be the answer.
 *   
 *   This routine should be used to retrieve genomic sequence ONLY!
 *
 *   note:
 *    1. assumptions have been made for the format of the pages returned
 *       from entrez search result.
 *    2. this routine is not used at all by far, but I decided to keep it
 *       in case we are going to augment the functionalities of genbank
 *       search.
 *     
 *   parameters:
 *     term - the search term
 *  
 *   returns:
 *     returns the url to the genbank entry which is of genbank file format. 
 *     (this can be parsed by genbank_retrieve)
 *
 * ----------------
 */


char* genbank_entrez( char *term )
{
   HttpContext hc;
   char *result, *data, *p, *q;
   int i;

   /* initialize the parameter structure for HTTP request */
   result = NULL;
   data = NULL;
   hc = CreateHttpContext( POST, 20, NULL );
   HttpContextSetParameter( hc, "CMD", "Put" );
   HttpContextSetParameter( hc, "term", term );

   /*
    *  here is something disturbing.
    *  we don't want users to specify the query database (Necleotide, or Protein) so we have
    *  to iteratively "probe" them.
    */
   for(i=0;i<MAX_GENBANK_DB;i++)
   {
      HttpContextSetParameter( hc, "DB", genbank_databases[i] );

      if ( HTTP_OK==HttpSendRequest(hc,genbank_host,genbank_server_port,genbank_query) )
      {
	 data = HttpReceiveRequest(hc);

         if ( (p=strstr(data,GENBANK_RESULT_STARTTAG))!=NULL )
         {
            if ( (q=strstr(p,GENBANK_QUERY_URL_PREFIX))!=NULL )
            { 
	        p = q + strlen(GENBANK_QUERY_URL_PREFIX);
                if ( (*p=='\'') || (*p=='"') )
                {
                   q = p+1;
                   while ( *q!=*p ) q++;
                   *q = 0;
                   result = (char*) palloc( strlen(p+1)+1);
                   strcpy(result,p+1);
                }
            }
            else
              elog(ERROR, "can not find '%s' in genbank record, possibly due to change of genbank web interface", GENBANK_QUERY_URL_PREFIX);

            break;
         }
      }
      else
        elog( ERROR, "can not establish HTTP connection to genbank server");
   }

   FreeHttpContext(hc);
   if (data) pfree(data);
   return result;
}

GenbankRecordPointer GenbankParser(char *data)
{
   GenbankRecordPointer result;
   char buffer[3][30], *offset, *p, *q, *s;
   char definition_buffer[1024];
   int length, i;

   result = NULL;
   if ( (offset=strstr(data,"LOCUS"))!=NULL )
   {
      sscanf(offset,"LOCUS %s %d %s %s ", &(buffer[0][0]), &length, &(buffer[1][0]), &(buffer[2][0]) );

      result = (GenbankRecordPointer) palloc(sizeof(GenbankData));	
      memset(result,0,sizeof(GenbankData));
      result->fields[GB_FIELD_LOCUS] = pstrdup(buffer[0]);
      result->fields[GB_FIELD_TYPE]  = pstrdup(buffer[2]);

      // now, we need to extract the last two fields in this line  
      // it's kind of heuristic because sometimes there will be 3 fields instead of 2
      p = strstr(offset, buffer[2]) + strlen(buffer[2]);
      s = strchr(p,'\n');
      *s = 0;
      i = sscanf(p, " %s %s %s\n", &(buffer[0][0]), &(buffer[1][0]), &(buffer[2][0]));
      *s = '\n';
      result->fields[GB_FIELD_DIVISION] = pstrdup(buffer[i-2]);
      result->fields[GB_FIELD_RECORD_DATE] = pstrdup(buffer[i-1]);

      // move to next line
      offset = strchr(offset+1,'\n')+1;

      // we now want to extract the definition which could spans among lines
      if ( (p=strstr(offset,"DEFINITION"))!=NULL )
      {
         definition_buffer[0] = '\0';
         s = GetNextString(p+10)-2;        // 10 is the length of "DEFINITION"
         for(q=s+1;isspace(*q);q=s+1)
         {
            q = GetNextString(q);
            s = strchr(q,'\n');
            *s = 0;
            strcat(definition_buffer, " " );
            strcat(definition_buffer, q);
            *s = '\n';
         }
         offset = s+1;
         result->fields[GB_FIELD_DESCRIPTION] = pstrdup(definition_buffer);            
      }
      else
      {
         // it seems impossible not to have this field in genbank record, but we put this just in case
         result->fields[GB_FIELD_DESCRIPTION] = (char *) palloc(1);
         result->fields[GB_FIELD_DESCRIPTION][0] = '\0';
      }

      // now, start to look for accession
      offset = strstr(offset,"ACCESSION");
      sscanf(offset,"ACCESSION %s ", &(buffer[0][0]) );
      result->fields[GB_FIELD_ACCESSION] = pstrdup(buffer[0]);

      // now, start to look for version
      offset = strstr(offset,"VERSION");
      sscanf(offset,"VERSION %s %s", &(buffer[0][0]), &(buffer[1][0]) );
      result->fields[GB_FIELD_VERSION] = pstrdup(buffer[0]);
      result->fields[GB_FIELD_GENBANK_ID] = pstrdup(buffer[1]);

      // looking for organism
      if ( (p=strstr(offset,"\nSOURCE"))!=NULL )
      {
         s = GetNextString(p+7);      // 6 is the length of "SOURCE" with a prefix "\n"
         q = strchr(s,'\n');
         *q = 0;
         if ( (p=strchr(s,'('))!=NULL )  *(p-1)='\0';
         result->fields[GB_FIELD_SPECIES] = pstrdup(s);
         offset = q+1;

      }
      else
      {
         result->fields[GB_FIELD_SPECIES] = pstrdup("unknown");
      }

      // now, start to process sequence
      result->fields[GB_FIELD_SEQUENCE] = palloc(length+1);
      p = result->fields[GB_FIELD_SEQUENCE];

      s = strstr(offset,"ORIGIN");
      s = strchr(s+6,'\n')+1;

      while (*s!='/')
      {
        if ( isalpha(*s) )   *p++ = *s;
        s++;
      }
      *p = 0;
   }

   return result;
}



/* ----------------
 *   genbank_retrieve
 *
 *   this is the routine to retrieve genbank record for a particular genomic 
 *   sequence.
 *
 *   parameters:
 *     gi    - gene id or url to genbank entry for a particular genomic sequence
 *     is_gi - is argument is a gene id or an url?  (1: gene id, 0: url)
 *
 *   note:
 *     we assume genbank file format should not change (or at least backward-compatible)
 *     this assumption should hold true all the time.
 *
 *   returns:
 *     returns the genomic sequence of interest in the following form:
 *     (locus, accession number, genomic sequence)
 *
 * ----------------
 */

GenbankRecordPointer genbank_retrieve( char *gi, int is_gi )
{
   HttpContext hc;
   char genbank_query[512];
   char *data;
   GenbankRecordPointer result;

   result = NULL;
   hc = CreateHttpContext( GET, 10, NULL );

   /* choose different HTTP get query string */
   if ( is_gi==1 )
   {
      HttpContextSetParameter( hc, "db", "nucleotides" );
      HttpContextSetParameter( hc, "val", gi );
   }
   else
   {
      sprintf( genbank_query, "%s", gi );
   }

   /* send request to genbank server */
   if ( HTTP_OK == HttpSendRequest( hc, genbank_host, genbank_server_port, genbank_query) )
   {
      data = HttpReceiveRequest(hc);
      result = GenbankParser(data);
   }

   return result;
}
