/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * loc.c
 *
 *    This file contains routines for loc data type.
 *
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */

#include "postgres.h"
#include "fmgr.h"
#include "funcapi.h"
#include "executor/spi.h"
#include "lib/stringinfo.h"
#include "miscadmin.h"
#include <string.h>
#include <stdio.h>
#include "access/gist.h"
#include "access/rtree.h"
#include "utils/builtins.h"
#include "loc.h"
#include "range.h"
#include "segdata.h"
#include "util.h"
#include "global.h"

extern int  seg_yyparse();
extern void seg_yyerror(const char *message);
extern void seg_scanner_init(const char *str);
extern void seg_scanner_finish(void);

//#define DEBUG_LOC   TRUE



/* ----------------
 *   symbol_lookup
 *
 *   This is the routine used to lookup identifiers in the atom table (name table).
 *   this rountine check if a given identifier exists in the atom table.
 *   If not, it will return FALSE. If yes, it will return TRUE and fill in necessary 
 *   information in tuple, desc
 *
 *   The structure of the atom table can be found in
 *     loc.h and BLASTgres.sql.in
 *   
 *   notice that all manipulations of the symbol table in this function assume that callers
 *   have done appropriate SPI environment setup. (i.e. makin sure that SPI_connect is called
 *   and SPI manager is connected.
 *
 *   parameters:
 *     str - an identifier
 *
 *   return:
 *     TRUE if successful, otherwise, FALSE
 *
 * ----------------
 */
bool symbol_lookup( HeapTuple *tuple, TupleDesc *desc, char *symbol )
{
   SPITupleTable *spi_tuptable;
   char *sql;
   int ret, proc; 

   sql = (char*) palloc( strlen(symbol) + 200 );

   /* check if the given identifier exists in the atom table */
   sprintf( sql, "SELECT * FROM %s WHERE %s='%s'", ATOM_TABLE_NAME, ATOM_TABLE_ID_FIELD, symbol );


   ret = SPI_execute(sql, false, 1);
   proc = SPI_processed;

   pfree(sql);

   if ( (ret==SPI_OK_SELECT) && (proc>0) )
   {
      /* there will be only one tuple returned, which is stored in spi_tuptable->vals[0] */
      spi_tuptable = SPI_tuptable;
      *tuple = spi_tuptable->vals[0];
      *desc  = spi_tuptable->tupdesc;
      return TRUE;
   }

   return FALSE;
}

bool symbol_lookup_by_sig( HeapTuple *tuple, TupleDesc *desc, uint32 sig )
{
   SPITupleTable *spi_tuptable;
   char sql[200];
   int ret, proc;

   /* check if the given identifier exists in the atom table */
   sprintf( sql, "SELECT * FROM %s WHERE %s=%u", ATOM_TABLE_NAME, ATOM_TABLE_SIG_FIELD, sig );


   ret = SPI_execute(sql, false, 1);
   proc = SPI_processed;

   if ( (ret==SPI_OK_SELECT) && (proc>0) )
   {
      /* there will be only one tuple returned, which is stored in spi_tuptable->vals[0] */
      spi_tuptable = SPI_tuptable;
      *tuple = spi_tuptable->vals[0];
      *desc  = spi_tuptable->tupdesc;
      return TRUE;
   }

   return FALSE;
}


/* ----------------
 *   get_sig_from_identifier
 *
 *   This is the routine used to manipulate the atom table (name table).
 *   this routine will check if the given identifier exists in the name
 *   table. If not, it will insert this identifier into the name table.
 *   It also returns the signature (a unique uint32 number for identifier)
 *
 *   The mapping from identifier to its corresponding signature is more
 *   complicated than it sounds. In the name table, it not only keeps track
 *   of all the used identifiers, but also records the aliasing relationships
 *   between identifiers. So, if an identifier is an alias to another one, 
 *   we have to trace back to its root to get the signature of the original
 *   identifier since all location operations including comparison are based
 *   on the actual location value rooted in the original coordinate. Meanwhile,
 *   we have to keep doing coordinate transformation along the process.
 *
 *
 *   The structure of the atom table can be found in
 *     loc.h and BLASTgres.sql.in
 *   
 *   parameters:
 *     str - an identifier
 *
 *   return:
 *     signature of that identifier
 *
 * ----------------
 */
uint32 get_sig_from_identifier( char *str, SEG *seg )
{
   TupleDesc      tupdesc;
   HeapTuple      tuple;
   char sql[300], symbol[1024], *value;
   int base, flag;

   RANGE_int lower, upper;

   uint32 sig;

   /* we need to connect to SPI manager in order to execute SQL */
   if ( SPI_ERROR_CONNECT==SPI_connect() )   elog( ERROR, "SPI_connect failed" );

   sig = 0;
   flag = 0;
   strcpy( symbol, str );

   /* 
      start to locate the root of this identifier, if it already exists.
      if the given identifier does not exist in the name table, we simply discontinue
      this while loop and create its entry in the name table.

      signature value 0 is invalid in the system. (used internally for sentinel)
      all identifiers will have their signature values greater than 0 
   */

   while (flag==0)
   {

      if ( symbol_lookup( &tuple, &tupdesc, symbol )==FALSE )
         flag = 1;
      else
      {
         /* retrieve the base value of this identifier. if it's not 0, it is an alias to some 
            other identifier */
         value = SPI_getvalue(tuple, tupdesc, ATOM_POS_BASE);
         base = atoi(value);
         pfree(value); 

         if ( base==0 )  
         {
            value = SPI_getvalue(tuple, tupdesc, ATOM_POS_SIG);
            sig = atoi(value);
            pfree(value);
            flag = 1;
         }
         else
         {
            value = SPI_getvalue(tuple, tupdesc, ATOM_POS_LOWER);
            lower = atoi(value);
            pfree(value);

            value = SPI_getvalue(tuple, tupdesc, ATOM_POS_UPPER);
            upper = atoi(value);
            pfree(value);

            // boundary checking

            if ( (seg->lower-1+lower>upper) || (seg->upper-1+lower>upper) )
               elog( ERROR, "slice of %s[%lld..%lld] is out of bound in %s[%lld..%lld]", 
                             str, seg->lower, seg->upper, symbol, lower, upper);          

            seg->lower = seg->lower - 1 + lower;
            seg->upper = seg->upper - 1 + lower;

            if ( TRUE==symbol_lookup_by_sig( &tuple, &tupdesc, base ) )
            {
               value = SPI_getvalue(tuple, tupdesc, ATOM_POS_ID );
               strcpy( symbol, value );
               pfree(value);
            }
            else
               elog( ERROR, "fatal error: internal system inconsistency");


         }
      }
   }

   if ( sig==0 )
   {
      sprintf( sql, "INSERT INTO %s (%s) VALUES ('%s')",ATOM_TABLE_NAME, ATOM_TABLE_ID_FIELD, str);
      flag = SPI_execute(sql, false, 1);
      if ( flag!=SPI_OK_INSERT )  
               elog( ERROR, "failed to insert identifier [%s]",str);

      if ( symbol_lookup( &tuple, &tupdesc, str )==FALSE )
         elog( ERROR, "fatal error: internal data inconsistency occurs");

      value = SPI_getvalue(tuple, tupdesc,1);
      sig = atoi(value);
      pfree(value);
   }       

   SPI_finish();
   return sig;      
}

/* ----------------
 *   get_id_from_sig
 *
 *   This is the routine used to manipulate the atom table (name table).
 *   this routine will return the identifier given its signature.
 *   The big assumption here is that the signature must exist!!!
 *
 *   The structure of the atom table can be found in
 *     loc.h and BLASTgres.sql.in
 *
 *   parameters:
 *     sig - signature
 *
 *   return:
 *     identifier corresponding to the signature
 *
 * ----------------
 */
char* get_id_from_sig( uint32 sig )
{
   SPITupleTable *spi_tuptable;
   TupleDesc	 spi_tupdesc;
   HeapTuple	 spi_tuple;
   int ret, proc;
   char sql[200], *result;

   if ( SPI_ERROR_CONNECT==SPI_connect() )
      elog( ERROR, "SPI_connect failed" );

   result = NULL;
   sprintf( sql, "SELECT %s FROM %s WHERE %s=%u", ATOM_TABLE_ID_FIELD, ATOM_TABLE_NAME, ATOM_TABLE_SIG_FIELD, sig );


   ret = SPI_execute(sql, false, 1);
   proc = SPI_processed;

   if ((ret==SPI_OK_SELECT) && (proc>0))
   {
       spi_tuptable = SPI_tuptable;
       spi_tupdesc = spi_tuptable->tupdesc;
       spi_tuple = SPI_tuptable->vals[0];
       result = SPI_getvalue(spi_tuple, spi_tupdesc, 1);
   }
   else
      elog(LOG, "failed to execute SQL");

   SPI_finish();
   return result;
}

/* ----------------
 *   loc_in
 *
 *   This is the routine used to create internal representation of loc
 *   data type in Postgresql. Any new value of user data types in Postgresql
 *   is provided in string and Postgresql server calls this routine to form
 *   the internal representation of this user data type.
 *
 *   It is our responsibility to parse the string (according to the format we
 *   specify) and form the internal representation of loc object.
 *
 *   the supported syntax for location is defined in segscan.l segparse.y
 *
 *   typical examples are:
 *   IDENTIFIER[INTEGER..INTEGER]      -- perl slice syntax
 *   IDENTIFIER[INTEGER:INTEGER]       -- python slice syntax
 *   IDENTIFIER[INTEGER:INTEGER]~      -- reverse strand of IDENTIFIER[INTEGER:INTEGER]
 *   IDENTIFIER[INTEGER..INTEGER]~[INTEGER:INTEGER]
 *   IDENTIFIER[INTEGER..INTEGER]~([INTEGER..INTEGER][INTEGER..INTEGER]~)~
 *
 *   parameters:
 *     str - a C string from user input
 *
 *   return:
 *     construct the internal representation of the data type.
 *
 *   note:
 *
 *   if the given string does not follow the syntax, error must be returned by
 *   elog( ERROR, "error message" ). This routine will return immediately.
 *
 *   Numbers can be prefix of an identifier in this syntax. The reason why I enable this is because
 *   BLAST server will return a random number sequence as sequence identifier if query is specified
 *   as a sequence of DNA( protein, or mrna) letters.
 *
 * ---------------
 */
LOC *loc_in(char *str)
{
  SEG *seg = palloc(sizeof(SEG));
  LOC *result = palloc(sizeof(LOC));

  seg_scanner_init(str);

  if (seg_yyparse(seg) != 0)
      seg_yyerror("bogus input");

  seg_scanner_finish();
  result->sig = get_sig_from_identifier(seg->id, seg);
  result->interval.lower = seg->lower;
  result->interval.upper = seg->upper;
  result->strand = seg->strand;

  return (result);
}

/* ----------------
 *   loc_to_string
 *
 *   This is the routine used to transform internal representation of loc
 *   data type to a string form in Postgresql. Whenever Postgresql needs to
 *   display the loc data, it will call this routine to get its string form.
 *
 *   Even though we support two forms of range specification, the returned
 *   format will be as follows:
 *
 *   INTEGER    <- [1-9]*[0-9]
 *   RANGE      <- INTEGER .. INTEGER
 *   LOC        <- IDENTIFIER "[" RANGE "]"
 *   LOC        <- "~" IDENTIFIER "[" RANGE "]"
 *
 *   in other words, it looks like
 *   IDENTIFIER[INTEGER..INTEGER]      -- perl slice syntax
 *
 *   parameters:
 *     loc - the internal representation of loc data
 *
 *   return:
 *     construct a string representation of the loc data type.
 *
 * ----------------
 */
char* loc_to_string( LOC* loc )
{
   char *result, *identifier;

   if (loc==NULL) return NULL;

   identifier = get_id_from_sig(loc->sig);
   result = (char*) palloc(60+strlen(identifier));

   if ( loc->strand==0 )
      sprintf( result, "%s[%s]", identifier, range_out(&(loc->interval)));
   else
      sprintf( result, "%s[%s]~", identifier, range_out(&(loc->interval)));


   pfree(identifier);
   return result;
}

bool alias( char *str )
{
  RANGE_int lower, upper;

  uint32 signature;
  char *p, *q, *sql;
  HeapTuple tuple;
  TupleDesc desc;
  int ret;
  SEG *seg;

  seg = (SEG*) palloc(sizeof(SEG));

  while ( isspace(*str) )  str++;
  p = strchr( str, '=' );
  if ( p==NULL )  return FALSE;

  // strip the alias name
  *p = 0;
  q = p-1;
  while ( isspace(*q) ) *q-- = '\0';

  seg_scanner_init(p+1);

  if (seg_yyparse(seg) != 0)
      seg_yyerror("invalid location input");

  seg_scanner_finish();


  // start to record this relationship in system atom table
  // 1. look up to see if the alias_id exists, and the boundary is fine
  // 2. record this relationship back to table

   /* we need to connect to SPI manager in order to execute SQL */
   if ( SPI_ERROR_CONNECT==SPI_connect() )
       elog( ERROR, "SPI_connect failed" );

   if ( TRUE==symbol_lookup(&tuple, &desc, seg->id) )
   {
      // now, tuple has the value
      // string = SPI_getvalue(tuple, desc, ATOM_POS_SIG );   1<- position
      p = SPI_getvalue(tuple,desc,ATOM_POS_SIG);
      signature = (uint32) atoi(p);
      pfree(p);

      p = SPI_getvalue(tuple,desc, ATOM_POS_LOWER);
      lower = atoi(p);
      pfree(p);

      p = SPI_getvalue(tuple,desc,ATOM_POS_UPPER);
      upper = atoi(p);
      pfree(p);

      if ( (seg->lower-1+lower>upper) || (seg->upper-1+lower)>upper )
        elog( ERROR, "slice out of boundary" );

      sql = (char*) palloc(300+strlen(str));

      sprintf( sql, "INSERT INTO %s (%s,%s,%s,%s) VALUES ('%s',%u, %lld, %lld)", 
                    ATOM_TABLE_NAME, ATOM_TABLE_ID_FIELD, ATOM_TABLE_BASE_FIELD,
                    ATOM_TABLE_LOWER_FIELD, ATOM_TABLE_UPPER_FIELD,
                    str, signature, seg->lower, seg->upper ); 

#ifdef DEBUG_LOC
      elog( LOG, "sql=[%s]", sql);
#endif       

      ret = SPI_execute(sql, false, 1);
      pfree(sql);

      if ( ret!=SPI_OK_INSERT )
          elog( ERROR, "failed to insert identifier [%s]",str);

   }
   else
     elog( ERROR, "can not create alias to undefined identifier - %s", seg->id );

  SPI_finish();
  return (TRUE);    
}

/* ----------------
 *   loc_seqid
 *
 *   This is the one of the membership functions, which returns
 *   the identifier of the location.
 *   notice that it does not return the "~" character if it
 *   represents the negative strand sequence. if you need to
 *   test which strand it is, you can use loc_positive_strand()
 *   or loc_negative_strand() to do that.
 *
 *   parameters:
 *     loc - the internal representation of loc data
 *
 *   return:
 *     return the identifier of the location.
 *
 * ----------------
 */
text* loc_seqid( LOC* loc )
{
   text *new_t;
   char *identifier;

   if ( loc==NULL ) return NULL;
   identifier = get_id_from_sig( loc->sig );

   new_t = cstring2text(identifier);
   pfree(identifier);

   return new_t;
}

/* ----------------
 *   loc_positive_strand
 *
 *   This is a routine which tests if the location is in the
 *   positive strand or not.
 *
 *   parameters:
 *     loc - the internal representation of loc data
 *
 *   return:
 *     return a boolean value. TRUE if it's in positive strand.
 *     otherwise, return FALSE
 *
 * ----------------
 */
bool loc_positive_strand( LOC *loc )
{
  if ( loc==NULL) return FALSE;
  if ( loc->strand==0 ) return TRUE;
  else return FALSE;
}

/* ----------------
 *   loc_negative_strand
 *
 *   This is a routine which tests if the location is in the
 *   negative strand or not.
 *
 *   parameters:
 *     loc - the internal representation of loc data
 *
 *   return:
 *     return a boolean value. TRUE if it's in positive strand.
 *     otherwise, return FALSE
 *
 * ----------------
 */
bool loc_negative_strand( LOC *loc )
{
  return !loc_positive_strand(loc);
}

/* ----------------
 *   loc_same_strand
 *
 *   This is a routine which tests if two locations are in the
 *   same strand.
 *
 *   parameters:
 *     a - the internal representation of loc data
 *     b - the internal representation of loc data
 *
 *   return:
 *     return a boolean value. TRUE if both locations are in the same
 *     strand. otherwise, return FALSE
 *
 * ----------------
 */
bool loc_same_strand( LOC *a, LOC *b )
{

  if ( (a==NULL) || (b==NULL) )  return FALSE;

  if ( a->strand == b->strand )  return TRUE;
  else return FALSE;
}

/* ----------------
 *   loc_negate
 *
 *   This is a routine which returns the negative strand of location
 *   of the given location.
 *
 *   parameters:
 *     a - the internal representation of loc data
 *
 *   return:
 *     return a location object which is the negative strand of the
 *     given loc data.
 *
 * ----------------
 */

LOC* loc_negate( LOC *a)
{
   LOC *n;

   if (a==NULL) return NULL;

   n = (LOC*) palloc(sizeof(LOC));
   n->sig = a->sig;
   n->interval.lower = a->interval.lower;
   n->interval.upper = a->interval.upper;
   n->strand = 1 - a->strand;

   return n;
}

/* ----------------
 *   loc_range
 *
 *   This is the one of the membership functions, which returns
 *   the range of the genomic sequence.
 *
 *   parameters:
 *     loc - the internal representation of loc data
 *
 *   return:
 *     return the range object of a location.
 *
 * ----------------
 */

RANGE *loc_range( LOC* loc )
{
   RANGE *s;
   if ( loc==NULL ) return NULL;

   s = (RANGE *) palloc(sizeof(RANGE));
   s->lower = loc->interval.lower;
   s->upper = loc->interval.upper;

   return s;
}

/* ----------------
 *   loc_lower
 *
 *   This is the one of the membership functions, which returns
 *   the lower bound of the genomic sequence.
 *
 *   parameters:
 *     loc - the internal representation of loc data
 *
 *   return:
 *     return the lower bound of a location.
 *
 * ----------------
 */

RANGE_int *loc_lower( LOC* loc)
{
  RANGE_int *result;

  result = (RANGE_int*) palloc(sizeof(RANGE_int));

  *result = loc->interval.lower;
  return (result);
}

/* ----------------
 *   loc_upper
 *
 *   This is the one of the membership functions, which returns
 *   the upper bound of the genomic sequence.
 *
 *   parameters:
 *     loc - the internal representation of loc data
 *
 *   return:
 *     return the upper bound of a location.
 *
 * ----------------
 */

RANGE_int *loc_upper( LOC* loc)
{
  RANGE_int* result;

  result = (RANGE_int*) palloc(sizeof(RANGE_int));

  *result = loc->interval.upper;
  return (result);
}



/* ----------------
 *   boolean testing functions
 *
 *   The following set of routines are used to test the
 *   relationship between two location objects, say a and b.
 *
 *   they will return a boolean value. TRUE means the relationship
 *   exists between these a and b.
 *
 *   they are intutive and the meaning of them can be easily seen
 *   from the implementation.
 *
 *   parameters:
 *     a - the first loc object
 *     b - the second loc object
 *
 *   return:
 *     return TRUE if the relationship exists between a and b; otherwise,
 *     return FALSE
 *
 * ----------------
 */

bool loc_eq(LOC *a, LOC *b)
{
   return (
             (a->sig==b->sig) && 
             (a->strand==b->strand) && 
             range_cmp( &(a->interval), &(b->interval) )==0 
          ); 
}

bool loc_eq_range(LOC*a, RANGE* r)
{
   return range_cmp( &(a->interval), r);
}

bool loc_ne(LOC *a, LOC *b)
{
   return !loc_eq(a,b);
}

bool loc_ne_range(LOC* a, RANGE *b)
{
   return !loc_eq_range(a,b);
}

bool loc_contains_int( LOC *a, RANGE_int* b)
{
   return range_contains_int( &(a->interval), b);
}

bool loc_contains( LOC *a, LOC *b)
{  
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) )  return FALSE;
   return range_contains( &(a->interval), &(b->interval) );
}

bool loc_contains_range( LOC *a, RANGE *r)
{
   return range_contains( &(a->interval), r );
}

bool loc_contained( LOC *a, LOC *b)
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) )  return FALSE;

   return range_contained( &(a->interval), &(b->interval) );
}

bool loc_range_contained(RANGE*r, LOC *b)
{
   return range_contained( r, &(b->interval) );
}

bool loc_overlaps( LOC *a, LOC *b )
{
  if ( (a->sig!=b->sig) || (a->strand!=b->strand) )  return FALSE;
  return range_overlaps( &(a->interval), &(b->interval) );
}

bool loc_overlaps_range( LOC*a, RANGE *r)
{
   return range_overlaps( &(a->interval), r);
}

bool loc_left( LOC *a, LOC *b )
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE;
   return range_left( &(a->interval), &(b->interval) );
}

bool loc_left_range(LOC* a, RANGE *r)
{
   return range_left( &(a->interval), r);
}

bool loc_over_left( LOC *a, LOC *b )
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE;
   return range_over_left( &(a->interval), &(b->interval) );
}

bool loc_over_left_range(LOC *a, RANGE* b)
{
   return range_over_left( &(a->interval), b);
}

bool loc_right(LOC * a, LOC * b)
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE;
   return range_right( &(a->interval), &(b->interval) );
}

bool loc_right_range(LOC *a, RANGE *b)
{
   return range_right( &(a->interval), b);
}

bool loc_over_right(LOC * a, LOC * b)
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE;
   return range_over_right( &(a->interval), &(b->interval) );
}

bool loc_over_right_range(LOC* a, RANGE *b)
{
   return range_over_right( &(a->interval), b);
}

RANGE_int* loc_range_size(LOC *a)
{
   return range_size( &(a->interval) );
}

bool loc_lt(LOC * a, LOC * b)
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE;
   return range_lt( &(a->interval), &(b->interval) );
}

bool loc_lt_range(LOC* a, RANGE *r)
{
   return range_lt( &(a->interval), r);
}

bool loc_le(LOC * a, LOC * b)
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE;
   return range_le( &(a->interval), &(b->interval) );
}

bool loc_le_range( LOC* a, RANGE *r )
{
   return range_le( &(a->interval), r );
}

bool loc_gt(LOC * a, LOC * b)
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE; 
   return range_gt( &(a->interval), &(b->interval) );
}

bool loc_gt_range( LOC* a, RANGE *r )
{
   return range_gt( &(a->interval), r );
}

bool loc_ge(LOC * a, LOC * b)
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE;
   return range_ge( &(a->interval), &(b->interval) );
}

bool loc_ge_range( LOC* a, RANGE* b)
{
   return range_ge( &(a->interval), b);
}

bool loc_meets(LOC *a, LOC *b)
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE;
   return range_meets( &(a->interval), &(b->interval) );
}

bool loc_meets_range(LOC*a, RANGE *r)
{
   return range_meets( &(a->interval), r);
}

bool loc_met_by(LOC *a, LOC *b)
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE;   
   return range_met_by( &(a->interval), &(b->interval) );
}

bool loc_met_by_range( LOC*a, RANGE *b)
{
   return range_met_by( &(a->interval), b);
}

bool loc_starts(LOC *a, LOC *b)
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE;
   return range_starts( &(a->interval), &(b->interval) );
}

bool loc_starts_range(LOC*a, RANGE *r)
{
   return range_starts( &(a->interval), r);
}

bool loc_started_by(LOC *a, LOC *b)
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE;
   return range_started_by( &(a->interval), &(b->interval) );
}

bool loc_started_by_range(LOC* a, RANGE *r)
{
   return range_started_by( &(a->interval), r );
}

bool loc_finishes(LOC *a, LOC *b)
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE;
   return range_finishes( &(a->interval), &(b->interval) );
}

bool loc_finishes_range( LOC* a, RANGE *r)
{
   return range_finishes( &(a->interval), r );
}

bool loc_finished_by(LOC *a, LOC *b)
{
   if ( (a->sig!=b->sig) || (a->strand!=b->strand) ) return FALSE;
   return range_finished_by( &(a->interval), &(b->interval) );
}

bool loc_finished_by_range(LOC*a, RANGE *r)
{
   return range_finished_by( &(a->interval), r );
}

/* ----------------
 *   LOC generator
 *
 *   The following set of routines are called LOC generator because
 *   they take necessary parameters (typically two loc objects) and
 *   generate a new loc object.
 *
 *   they are intuitive and the meanings of them can be easily seen from
 *   the name and the the implementation.
 *
 *   return:
 *     return a new LOC object
 *
 * ----------------
 */
LOC *loc_union(LOC * a, LOC * b)
{
  LOC *v;
  RANGE *s;

  if ( ( a->sig != b->sig) || (a->strand!=b->strand) )
     elog( ERROR, "function is not allowed for this loc pair" );

  v = (LOC*) palloc( sizeof(LOC) );
  v->sig = a->sig;
  v->strand = a->strand;
  s = range_union( &(a->interval), &(b->interval) );
  v->interval.lower = s->lower;
  v->interval.upper = s->upper;

  pfree(s);

  return v;
}

LOC *loc_inter(LOC * a, LOC * b)
{
  LOC *v;
  RANGE *s;

  if  ( (a->sig!=b->sig)  || (a->strand!=b->strand) )
    elog( ERROR, "function is not allowed for this loc pair" );

  v = (LOC*) palloc( sizeof(LOC) );
  v->sig = a->sig;
  v->strand = a->strand;
  s = range_inter( &(a->interval), &(b->interval) );
  v->interval.lower = s->lower;
  v->interval.upper = s->upper;

  pfree(s);

  return v;
}

LOC* loc_plus(LOC *a, LOC *b)
{
  LOC *v;
  RANGE *s;

  if ( (a->sig!=b->sig)  || (a->strand!=b->strand) )
     elog( ERROR, "function is not allowed for this loc pair" );

  v = (LOC*) palloc( sizeof(LOC));
  v->sig = a->sig;
  v->strand = a->strand;
  s = range_plus( &(a->interval), &(b->interval) );
  v->interval.lower = s->lower;
  v->interval.upper = s->upper;

  pfree(s);                                                                               
  return v;
}

LOC* loc_minus(LOC *a, LOC *b)
{
  LOC *v;
  RANGE *s;

  if ( (a->sig!=b->sig)  || (a->strand!=b->strand) )
    elog( ERROR, "function is not allowed for this loc pair" );

  v = (LOC*) palloc( sizeof(LOC) );
  v->sig = a->sig;
  v->strand = a->strand;
  s = range_minus( &(a->interval), &(b->interval) );
  v->interval.lower = s->lower;
  v->interval.upper = s->upper;                                                                                                                              
  pfree(s);                                                                                                   
  return v;
}

LOC* loc_minus_const(LOC*a, int32 b)
{
  LOC *v;

  v = (LOC*) palloc( sizeof(LOC) );
  v->sig = a->sig;
  v->strand = a->strand;
  v->interval.lower = a->interval.lower-b;
  v->interval.upper = a->interval.upper-b; 
  return v;
}

LOC* loc_minus_range(LOC*a, RANGE *r)
{
  LOC *v;

  v = (LOC*) palloc(sizeof(LOC) );
  v->sig = a->sig;
  v->strand = a->strand;
  v->interval.lower = a->interval.lower - r->lower;
  v->interval.upper = a->interval.upper - r->upper;
  return v;
}


LOC* loc_plus_const(LOC*a, int32 b)
{
  LOC *v;

  v = (LOC*) palloc( sizeof(LOC) );
  v->sig= a->sig;
  v->strand = a->strand;
  v->interval.lower = a->interval.lower+b;
  v->interval.upper = a->interval.upper+b;
  return v;
}

LOC* loc_const_plus(int32 b, LOC* a)
{
  return loc_plus_const(a,b);
}

LOC* loc_plus_range(LOC*a, RANGE *r)
{
  LOC *v;

  v = (LOC*) palloc( sizeof(LOC) );
  v->sig = a->sig;
  v->strand = a->strand;
  v->interval.lower = a->interval.lower + r->lower;
  v->interval.upper = a->interval.upper + r->upper;
  return v;
}

LOC* loc_times(LOC *a, LOC *b)
{
  LOC *v;
  RANGE *s;

  if ( (a->sig!=b->sig) || (a->strand!=b->strand) )
     elog( ERROR, "function is not allowed for this loc pair" );

  v = (LOC*) palloc( sizeof(LOC) );
  v->sig = a->sig;               
  v->strand = a->strand;                                                                     
  s = range_times( &(a->interval), &(b->interval) );
  v->interval.lower = s->lower;
  v->interval.upper = s->upper;

  pfree(s);                                                      
  return v;
}

LOC* loc_times_const(LOC *a, int32 b)
{
  LOC *v;

  v = (LOC*) palloc( sizeof(LOC));
  v->sig = a->sig;
  v->strand = a->strand;
  v->interval.lower = a->interval.lower*b;
  v->interval.upper = a->interval.upper*b;

  return v;
}

LOC* loc_const_times(int32 b, LOC* a)
{
  return loc_times_const(a,b);
}

LOC* loc_times_range(LOC *a, RANGE* b)
{
  LOC *v;

  v = (LOC*) palloc( sizeof(LOC) );
  v->sig = a->sig;
  v->strand = a->strand;
  v->interval.lower = a->interval.lower*b->lower;
  v->interval.upper = a->interval.upper*b->upper;

  return v;
}


LOC* loc_maxmin(LOC *a, LOC *b)
{
  LOC *v;
  RANGE *s;

  if ( (a->sig!=b->sig) || (a->strand!=b->strand) )
     elog( ERROR, "function is not allowed for this loc pair" );

  v = (LOC*) palloc( sizeof(LOC) );
  v->sig = a->sig;
  v->strand = a->strand;
  s = range_maxmin( &(a->interval), &(b->interval) );
  v->interval.lower = s->lower;
  v->interval.upper = s->upper;

  pfree(s);

  return v;
}

LOC* loc_maxmin_range(LOC* a, RANGE *b)
{
  LOC *v;

  v = (LOC*) palloc(sizeof(LOC));
  v->sig = a->sig;
  v->strand = a->strand;
  v->interval.lower = Max( a->interval.lower, b->lower );
  v->interval.upper = Min( a->interval.upper, b->upper );

  return v;
}

LOC* loc_minmax(LOC *a, LOC *b)
{
  LOC *v;
  RANGE *s;

  if ( (a->sig!=b->sig) || (a->strand!=b->strand) )
     elog( ERROR, "function is not allowed for this loc pair" );

  v = (LOC*) palloc( sizeof(LOC) );
  v->sig = a->sig;               
  v->strand = a->strand;                                                                                  
  s = range_minmax( &(a->interval), &(b->interval) );
  v->interval.lower = s->lower;
  v->interval.upper = s->upper;

  pfree(s);       
  return v;
}

LOC* loc_minmax_range(LOC* a, RANGE *b)
{
  LOC *v;

  v = (LOC*) palloc(sizeof(LOC));
  v->sig = a->sig;
  v->strand = a->strand;
  v->interval.lower = Min( a->interval.lower, b->lower );
  v->interval.upper = Max( a->interval.upper, b->upper );

  return v;
}

LOC* loc_extend(LOC* a, int32 b)
{
  LOC *v;

  v = (LOC*) palloc(sizeof(LOC));
  v->sig = a->sig;
  v->strand = a->strand;
  v->interval.lower = Max(a->interval.lower-b,0);
  v->interval.upper = a->interval.upper+b;

  return v;   

}

/* ----------------
 *   toloc
 *
 *   This is the routine used to transform  internal representation of loc
 *   data type from a string in Postgresql.
 *
 *   parameters:
 *     id  - c string for identifier
 *     a   - integer
 *     b   - integer
 *
 *   return:
 *     construct the internal representation of the loc data type.
 *
 * ----------------
 */

LOC* toloc(char *id, RANGE_int* a, RANGE_int* b)
{
   LOC *v;
   char *buffer;

   buffer = (char*) palloc(200+strlen(id));
   sprintf( buffer,"%s[%lld..%lld", id, *a, *b);
   v = loc_in(buffer);
   pfree(buffer);

   return v;
}

/* ----------------
 *   rtoloc
 *
 *   This is the routine used to transform  internal representation of loc
 *   data type from a string in Postgresql.
 *
 *   parameters:
 *     id    - c string for identifier
 *     range - range object
 *
 *   return:
 *     construct the internal representation of the loc data type.
 *
 * ----------------
 */
LOC* rtoloc(char *id, RANGE *r)
{
   LOC *v;
   char *buffer;

   buffer = (char*) palloc( 200 + strlen(id) );
   sprintf( buffer, "%s[%s]", id, range_out(r) );
   v = loc_in(buffer);
   pfree(buffer);

   return v;
}

/* ----------------
 *   stoloc
 *
 *   This is the routine used to transform  internal representation of loc
 *   data type from a string in Postgresql.
 *
 *   parameters:
 *     id    - c string for identifier
 *     range - c string for range
 *
 *   return:
 *     construct the internal representation of the loc data type.
 *
 * ----------------
 */
LOC* stoloc(char *id, char *range)
{
   LOC *v;
   char *buffer;

   buffer = (char*) palloc( 200 + strlen(id) );
   sprintf( buffer,  "%s[%s]", id, range);
   v = loc_in(buffer);
   pfree(buffer);
   return v;
}

LOC* text_toloc(text *t, int a, int b) 
{
   LOC *v;
   char *p = text2cstring(t);
   buffer = (char*) palloc(20+strlen(p));
   sprintf( buffer, "%s[%d..%d]", p, a, b);

   v = loc_in(buffer);
   pfree(buffer);

   return v;
}

LOC* text_rtoloc(text *t, RANGE *r)
{
   LOC *v;
   char *p = text2cstring(t);
   buffer = (char*) palloc(20+strlen(p));
   sprintf( buffer, "%s[%d..%d]", p, r->lower, r->upper);

   v = loc_in(buffer);
   pfree(buffer);

   return v;
}

