/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * loc.h
 *
 *    Definitions of functions for loc datatype.
 *    the loc datatype is defined as an identifier for a particular
 *    genomic sequence and a pair of integers specifying an
 *    interval. end points (lower, upper) are included.
 *
 * This file is included in loc.c and all other files which require
 * range datatype in Postgresql.
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */

#ifndef LOC_H
#define LOC_H

#include "postgres.h"
#include "range.h"
#include "segdata.h"
#include "global.h"

typedef struct LOC
{
  uint32 sig;
  RANGE interval;
  char strand;
} LOC;

/*
 *   system-wide symbol table
 *   ------------------------
 *  
 *   this symbol table is used to keep track of every
 *   identifiers used in ths system as well as their
 *   aliases.
 *   aliases could be specified for the following 
 *   purposes:
 * 
 *   1. simple aliasing, identifiers are used in different
 *      namespaces but they are essentially referring to
 *      the same genomic sequence.
 *
 *      eg. AL135903 is an alias for GI:12830388
 *   
 *      AL135903 is the locus and accession for a pseudogene
 *      (importin alpha 6) (KPNA5) . its genbank id is
 *      GI:12830388
 * 
 *   2. aliasing a slice of a sequence
 *      
 *      eg. DUT-exon3 = AF018430[560..651]
 *    
 *      AF018430 is part of the Homo sapiens dUTP gene.
 *      (the 2nd segment out of 4, which contains the
 *      3rd exon in the gene). 
 *      
 *      Sometimes we want to manipulate the inner structure
 *      inside a particular slice. This is why we need this
 *      kind of aliasing mechanism.
 *
 *      This symbol table has to record the information about
 *      identifiers and keep track of the aliasing relationship
 *
 *      table schema
 *      ============
 *
 *      CREATE TABLE sys_atom_table (
 *         atom     bigserial,     -- unique number (>=1)
 *         name     text,          -- identifier
 *         base	    uint8,         -- the aliased identifier, or 0
 *         lower    uint8,         -- lower bound of the slice, or 0
 *         upper    uint8          -- upper bound of the slice, or infinity
 *      );
 *     
 *       
 */

#define ATOM_TABLE_NAME          "sys_atom_table"
#define ATOM_TABLE_SIG_FIELD     "signature"
#define ATOM_TABLE_ID_FIELD      "identifier"
#define ATOM_TABLE_BASE_FIELD    "base"
#define ATOM_TABLE_LOWER_FIELD   "lower"
#define ATOM_TABLE_UPPER_FIELD   "upper"

#define ATOM_POS_SIG	1
#define ATOM_POS_ID	2
#define ATOM_POS_BASE	3
#define ATOM_POS_LOWER	4
#define ATOM_POS_UPPER	5

bool symbol_lookup( HeapTuple *tuple, TupleDesc *desc, char *symbol  );
bool symbol_lookup_by_sig( HeapTuple *tuple, TupleDesc *desc, uint32 sig);
LOC*  loc_in( char* );
char* loc_to_string( LOC* );
bool alias( char *str);
text* loc_seqid( LOC* );
bool loc_positive_strand(LOC*);
bool loc_negative_strand(LOC*);
bool loc_same_strand(LOC*,LOC*);
LOC* loc_negate(LOC*);
RANGE* loc_range( LOC* );

RANGE_int *loc_range_size( LOC* );
RANGE_int *loc_lower(LOC*);
RANGE_int *loc_upper(LOC*);
bool loc_contains_int( LOC*, RANGE_int* );

bool loc_contains( LOC*, LOC* );
bool loc_contains_range(LOC*, RANGE*);
bool loc_contained( LOC* , LOC* );
bool loc_range_contained( RANGE*, LOC*);
bool loc_overlaps( LOC*, LOC* );
bool loc_overlaps_range( LOC*, RANGE* );
bool loc_left( LOC*, LOC*);
bool loc_left_range( LOC*, RANGE *);
bool loc_range_left( RANGE*, LOC*);
bool loc_over_left( LOC*, LOC *);
bool loc_over_left_range( LOC*, RANGE *);
bool loc_range_over_left(RANGE*, LOC*);
bool loc_right(LOC *, LOC *);
bool loc_right_range(LOC*, RANGE*);
bool loc_over_right(LOC* , LOC* );
bool loc_over_right_range(LOC*, RANGE*);
bool loc_eq(LOC*, LOC* );
bool loc_eq_range(LOC*, RANGE*);
bool loc_ne(LOC*, LOC* );
bool loc_ne_range(LOC*, RANGE*);
bool loc_lt(LOC*, LOC* );
bool loc_lt_range(LOC*, RANGE*);
bool loc_le(LOC*, LOC* );
bool loc_le_range(LOC*, RANGE*);
bool loc_gt(LOC*, LOC* );
bool loc_gt_range(LOC*, RANGE*);
bool loc_ge(LOC*, LOC* );
bool loc_ge_range(LOC*, RANGE*);
bool loc_meets(LOC*, LOC*);
bool loc_meets_range(LOC*, RANGE*);
bool loc_met_by(LOC*, LOC*);
bool loc_met_by_range(LOC*, RANGE*);
bool loc_starts(LOC*, LOC*);
bool loc_starts_range(LOC*, RANGE*);
bool loc_started_by(LOC*, LOC*);
bool loc_started_by_range(LOC*, RANGE*);
bool loc_finishes(LOC*, LOC*);
bool loc_finishes_range(LOC*, RANGE*);
bool loc_finished_by(LOC*, LOC*);
bool loc_finished_by_range(LOC*,RANGE*);
LOC *loc_union(LOC*, LOC* );
LOC* loc_inter(LOC*, LOC* );
LOC* loc_plus(LOC*, LOC* );
LOC* loc_plus_const(LOC*, int32);
LOC* loc_const_plus(int32, LOC*);
LOC* loc_plus_range(LOC*, RANGE*);
LOC* loc_minus(LOC* , LOC* );
LOC* loc_minus_const(LOC*, int32);
LOC* loc_minus_range(LOC*, RANGE*);
LOC* loc_times(LOC*, LOC* );
LOC* loc_times_const(LOC*, int32);
LOC* loc_const_times(int32, LOC*);
LOC* loc_times_range(LOC*, RANGE*);
LOC* loc_maxmin(LOC*, LOC* );
LOC* loc_maxmin_range(LOC*, RANGE*);
LOC* loc_minmax(LOC*, LOC* );
LOC* loc_minmax_range(LOC*, RANGE*);
LOC* loc_extend(LOC*, int32); 

LOC* toloc(char *, RANGE_int*, RANGE_int* );

LOC* rtoloc(char *, RANGE *);
LOC* stoloc(char *, char *);
LOC* text_toloc(text *, int a, int b);
LOC* text_rtoloc( text *, RANGE *);
uint32 get_sig_from_identifier(char *str, SEG *seg);
char* get_id_from_sig(uint32);

#endif  /* LOC_H */
