/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * loc_aggregate.c
 *
 *    This file contains routines for loc array aggregator and 
 *    enumerator
 *
 * This is part of the BLASTgres project.
 * 
 *-------------------------------------------------------------------------
 */

#include "postgres.h"

#include <string.h>
#include <ctype.h>
#include <sys/types.h>

#include "access/heapam.h"
#include "catalog/indexing.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_type.h"
#include "executor/executor.h"
#include "utils/syscache.h"
#include "access/tupmacs.h"
#include "access/xact.h"
#include "fmgr.h"
#include "funcapi.h"
#include "executor/spi.h"
#include "miscadmin.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "utils/memutils.h"
#include "utils/lsyscache.h"

#include "loc.h"


/* 
 * this is a standard varlena array type in postgresql
 * ArrayType is defined in postgresql src/includes/utils/array.h
 * A typical array in Postgresql consists the following fields:
 *
 *   int32 size              - total array size  
 *   int ndim                - dimension of the array
 *   int flags               - implementation flags
 *   Oid elemtype            - element object id
 *   int dim[ndim]           - number of elements in each dimension
 *   int dim_lower[ndim]     - the lower bound for each dimension
 *   char data[1]            - real data is stored here.
 *
 *  the following PGARRAY structure fits a 1-D array
 */

typedef struct
{
  ArrayType  a;
  int  items;
  int  lower;
  LOC  array[1];
} PGARRAY;

/* 
 * this structure is used between different call context and to
 * save the state of our aggregate function
 */
typedef struct callContext
{
  PGARRAY  *p;
  int  num;
  int  flags;
} CTX;

#define TOASTED		1
#define START_NUM	8	/* initial size of arrays */
#define PGARRAY_SIZE(n) (sizeof(PGARRAY) + (((n)-1)*sizeof(LOC)))

static PGARRAY *GetPGArray(PGARRAY *p, int fAdd);
static PGARRAY *ShrinkPGArray(PGARRAY *p);

Datum		loc_agg_state(PG_FUNCTION_ARGS);
Datum		loc_agg_final_array(PG_FUNCTION_ARGS);
Datum		loc_enum(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(loc_agg_state);
PG_FUNCTION_INFO_V1(loc_agg_final_array);
PG_FUNCTION_INFO_V1(loc_enum);


/* ----------------
 *   GetPGArray
 *
 *   This is the routine used to generate a PGArray or enlarge a PGArray
 *   when necessary. This PGARRAY is used to manage the aggregation
 *   state of the array
 *
 *   parameters:
 *     p - PGARRAY
 *     fAdd - 1 means we are going to add a new array element
 *
 *   return:
 *     PGARRAY
 */

static PGARRAY *GetPGArray(PGARRAY *p, int fAdd)
{
   int ret, proc, cb;

   if (!p)
   {
      cb = PGARRAY_SIZE(START_NUM);

      p = (PGARRAY *) MemoryContextAlloc(PortalContext, cb);
      p->a.size = cb;
      p->a.ndim = 0;
      p->a.flags = 0;

      // lookup the loc type's oid
      if ((ret=SPI_connect())<0)
          elog( ERROR, "GetPGArray: SPI_connect returns %d",ret);

      ret = SPI_execute("select oid from pg_type where typname='loc'", true,0);
      proc = SPI_processed;
      if ((ret==SPI_OK_SELECT) &&(proc>0))
      {
         SPITupleTable *spi_tuptable = SPI_tuptable;
         TupleDesc spi_tupdesc = spi_tuptable->tupdesc; 
         HeapTuple spi_tuple = spi_tuptable->vals[0];
         char *val = SPI_getvalue(spi_tuple, spi_tupdesc,1);
         p->a.elemtype = atoi(val);
      }

      SPI_finish();      

      p->items = 0;
      p->lower = START_NUM;
   }
   else if (fAdd)
   { 
       // ensure array has space 
      if (p->items >= p->lower)
      {
         PGARRAY *pn;
         int n = p->lower + p->lower;  // we always double the space
         int cbNew = PGARRAY_SIZE(n);
         pn = (PGARRAY *) repalloc(p, cbNew);
	 pn->a.size = cbNew;
	 pn->lower = n;
	 return pn;
      }
   }

   return p;
}


/* ----------------
 *   ShrinkPGArray
 *
 *   This is the routine used to shrink the space we pre-allocated to array, freeing up
 *   the working memory. The amount of space allocated is doubled when there's no space 
 *   (starting from 8) so there might be a lot of space left when the aggregation function 
 *   finishes.
 *   
 *   parameters:
 *     p - PGARRAY
 *
 *   return:
 *     shrinked PGARRAY
 *
 * ----------------
 */

static PGARRAY *ShrinkPGArray(PGARRAY * p)
{
   int cb;
   PGARRAY *pnew = NULL;

   if (p)
   {
      // this is the actual array size
      cb = PGARRAY_SIZE(p->items);
      // use current transaction context
      pnew = palloc(cb);

      // Fix up the fields in the new structure, so Postgres understands
      memcpy(pnew, p, cb);
      pnew->a.size = cb;
      pnew->a.ndim = 1;
      pnew->a.flags = 0;
      pnew->a.elemtype = p->a.elemtype;
      pnew->lower = 1;
      pfree(p);
   }
   return pnew;
}

/* ----------------
 *   loc_agg_state
 *
 *   This is the routine called during the calculating of the aggregate.
 *   It maintains the aggregated version of the PGARRAY
 *
 *   parameters:
 *     1st argument - LOC data
 *
 *   return:
 *     aggregated PGARRAY
 *
 * ----------------
 */

Datum loc_agg_state(PG_FUNCTION_ARGS)
{
   PGARRAY *state;
   PGARRAY *p;

   if (PG_ARGISNULL(0))
      state = NULL;
   else
      state = (PGARRAY *) PG_GETARG_POINTER(0);
      p = GetPGArray(state, 1);

   if (!PG_ARGISNULL(1))
   {
     LOC *value = (LOC*) PG_GETARG_POINTER(1);

     if (!p)		/* internal error */
        elog(ERROR, "no aggregate storage");
     else if (p->items >= p->lower)		/* internal error */
        elog(ERROR, "aggregate storage too small");
     else
        memcpy( &(p->array[p->items++]), value, sizeof(LOC));
   }

   PG_RETURN_POINTER(p);
}

/* ----------------
 *   loc_agg_final_array
 *
 *   This is the routine called to finalize the calculating of the aggregate.
 *   It gives us the chance to repack (such as shrink the array) the final
 *   aggregate result.
 *
 *   parameters:
 *     1st argument - LOC data
 *
 *   return:
 *     aggregated PGARRAY
 *
 * ----------------
 */

Datum loc_agg_final_array(PG_FUNCTION_ARGS)
{
   PGARRAY *state = (PGARRAY *) PG_GETARG_POINTER(0);
   PGARRAY *pnew = ShrinkPGArray(GetPGArray(state, 0));

   if (pnew)
     PG_RETURN_POINTER(pnew);
   else
    PG_RETURN_NULL();
}

/* This function accepts an array, and returns one item for each entry in the array */

/* ----------------
 *   loc_enum
 *
 *   This is the routine used to enumerate each element in the array.
 *   It accepts an array and returns one item for each entry in the array.
 *
 *   parameters:
 *     1st argument - LOC[]
 *
 *   return:
 *     loc data
 *
 * ----------------
 */

Datum loc_enum(PG_FUNCTION_ARGS)
{
   PGARRAY *p = (PGARRAY *) PG_GETARG_POINTER(0);
   CTX     *pc;
   ReturnSetInfo *rsi = (ReturnSetInfo *) fcinfo->resultinfo;

   if (!rsi || !IsA(rsi, ReturnSetInfo))
     ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
	      errmsg("range_enum called in context that cannot accept a set")));

   if (!p)
   {
     elog(WARNING, "no data sent");
     PG_RETURN_NULL();
   }

   if (!fcinfo->context)
   {
      // Allocate a working context 
      MemoryContext oldcontext;

      oldcontext = MemoryContextSwitchTo(fcinfo->flinfo->fn_mcxt);
      pc = (CTX *) palloc(sizeof(CTX));

      // Don't copy attribute if you don't need to 
      if (VARATT_IS_EXTENDED(p))
      {
         // Toasted!!! 
	 pc->p = (PGARRAY *) PG_DETOAST_DATUM_COPY(p);
	 pc->flags = TOASTED;
      }
      else
      {
         // Untoasted 
	 pc->p = p;
	 pc->flags = 0;
      }
      pc->num = 0;
      fcinfo->context = (Node *) pc;
      MemoryContextSwitchTo(oldcontext);
   }
   else	/* use an existing one */
      pc = (CTX *) fcinfo->context;

   // Are we done yet? 
   if (pc->num >= pc->p->items)
   {
      // We are done 
      if (pc->flags & TOASTED)  pfree(pc->p);
      pfree(fcinfo->context);
      fcinfo->context = NULL;
      rsi->isDone = ExprEndResult;
   }
   else
   {
      // nope, return the next value 
      LOC *r = (LOC*) palloc(sizeof(LOC));
      memcpy(r, &(pc->p->array[pc->num++]), sizeof(LOC));
      rsi->isDone = ExprMultipleResult;
      PG_RETURN_POINTER(r);
   }
   PG_RETURN_NULL();
}
