/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * loc_gist.c
 *
 *    This file contains routines for loc gist access method
 *
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */

#include "postgres.h"
#include "fmgr.h"
#include "funcapi.h"
#include "executor/spi.h"
#include "lib/stringinfo.h"
#include "miscadmin.h"
#include "access/gist.h"
#include "access/rtree.h"
#include "utils/builtins.h"
#include <stdio.h>
#include <float.h>
#include <string.h>
#include "range.h"
#include "loc.h"
#include "loc_gist.h"

//#define DEBUG_LOC_GIST   TRUE

/* ----------------
 *   lockey_in
 *
 *   This is the routine used to create internal representation of lockey
 *   data type. LOCKEY is the internal structure for nodes in loc gist
 *   access method.
 *   As the implementation shows, users are not supposed to call this
 *   function. Postgresql requires every datatype has its own
 *   input/output function, which is why this function is needed.
 *
 * ----------------
 */
Datum lockey_in(PG_FUNCTION_ARGS)
{
  ereport(ERROR,
                (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
                 errmsg("<datatype>key_in() not implemented")));

  PG_RETURN_POINTER(NULL);
}

/* ----------------
 *   lockey_out
 *
 *   This is the routine used to create internal representation of lockey
 *   data type. LOCKEY is the internal structure for nodes in loc gist
 *   access method.
 *   As the implementation shows, users are not supposed to call this
 *   function. Postgresql requires every datatype has its own
 *   input/output function, which is why this function is needed.
 *
 * ----------------
 */
Datum lockey_out(PG_FUNCTION_ARGS)
{
  ereport(ERROR,
                (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
                 errmsg("<datatype>key_out() not implemented")));

  PG_RETURN_POINTER(NULL);
}

/*
   The following routines are for GiST index implementation
   GiST requires us to provide six methods: (according to the GiST paper)

   1. Consistent(E,q): determine if datum in the sub tree of the given entry E can be
      guaranteed un-satisfiable to query q or not.
   2. Union(P): given a set P of entries, form a predicate that holds true for all entries.
   3. Compress(E): return the compressed representation of the given entry E.
   4. Decompress(E): return the decompressed representation of the given entry E.
   5. Penalty(E1,E2):  return a domain-specific penalty for inserting E2 into the subtree of E1.
      GiST inserts the entry with least penalty into a sub tree.
   6. PickSplit(P): split a set P of entries into two sets of entries.

   In Postgresql, query type is specified by StrategyNumber so we are able to support
   as many predicate as we could as long as the index structure supports it efficiently.

*/

/* ----------------
 *   LOC_leaf_consistent
 *
 *   This is the routine called by LOC_consistent when the given GiST node is a leaf
 *   node.
 *   for detailed explanation of GiST consistent method, please refer to the comment
 *   for grange_consistent and GiST paper.
 *
 *   parameters:
 *     value - a given GiST leaf node
 *     query - user-specified query (a range object)
 *     strategy - the strategy number (defined in range.h)
 *
 *   return:
 *     return true if this query is consistent with the given GiST entry; otherwise,
 *     return false.
 *
 * ----------------
 */
Datum LOC_leaf_consistent( LOCKEY *value, LOC *query, StrategyNumber strategy )
{
  bool retval;


  if ( (value->s_lower!=query->sig) || (value->strand!=query->strand) )   
     PG_RETURN_BOOL(FALSE);

  switch (strategy)
  {
    case LOCEqualStrategyNumber:  
	retval = (bool) range_eq( &(value->interval), &(query->interval) );
        break;
    case LOCContainsStrategyNumber:
	retval = (bool) range_contains( &(value->interval), &(query->interval) );
        break;
    case LOCLeftStrategyNumber:
	retval = (bool) range_left( &(value->interval), &(query->interval) );
	break;
    case LOCOverLeftStrategyNumber:
        retval = (bool) range_over_left( &(value->interval), &(query->interval) );
	break;
    case LOCRightStrategyNumber:
	retval = (bool) range_right( &(value->interval), &(query->interval) );
	break;
    case LOCOverRightStrategyNumber:
	retval = (bool) range_over_right( &(value->interval), &(query->interval) );
   	break;
    case LOCContainedStrategyNumber:
	retval = (bool) range_contained( &(value->interval), &(query->interval) );
    default:
	retval = FALSE;
  }

#ifdef DEBUG_LOC_GIST
  elog( LOG, "retval=%d", retval);
#endif


  PG_RETURN_BOOL(retval);  
}

/* ----------------
 *   LOC_internal_consistent
 *
 *   This is the routine called by LOC_consistent when the given GiST node is a leaf
 *   node.
 *   for detailed explanation of GiST consistent method, please refer to the comment
 *   for grange_consistent and GiST paper.
 *
 *   parameters:
 *     value - a given GiST leaf node
 *     query - user-specified query (a range object)
 *     strategy - the strategy number (defined in range.h)
 *
 *   return:
 *     return true if this query is consistent with the given GiST entry; otherwise,
 *     return false.
 *
 *   note:
 *     If you're very sure that the entries under this given GiST node is not consistent
 *     with the query, you should always return FALSE. Returning TRUE is not harmful if
 *     subnodes are not consistent with the query, but the performance degradation is the
 *     penalty.
 *
 * ----------------
 */
Datum LOC_internal_consistent( LOCKEY *value, LOC *query, StrategyNumber strategy )
{
  bool retval;
  int signature_in_range; 

  signature_in_range = ((query->sig>=value->s_lower) && (query->sig<=value->s_upper));


  /* comparison of range value only happens when the sub tree contains entries of this signature value */
  retval = FALSE;
  if ( !signature_in_range )  PG_RETURN_BOOL(retval);

  switch (strategy)
  {
    case LOCLeftStrategyNumber:
    case LOCOverLeftStrategyNumber:
           retval = (bool) range_over_left( &(value->interval), &(query->interval) );
	   break;
    case LOCContainsStrategyNumber:
    case LOCEqualStrategyNumber:
           retval = range_contains( &(value->interval), &(query->interval) );
           break;
    case LOCOverRightStrategyNumber:
    case LOCRightStrategyNumber:
	   retval = (bool) range_right( &(value->interval), &(query->interval) );
	   break;
    case LOCContainedStrategyNumber:
	   retval = (bool) range_overlaps( &(value->interval), &(query->interval) );
           break;
    default:
	   retval = FALSE;
  }

#ifdef DEBUG_LOC_GIST
  elog( LOG, "retval=%d",  retval);
#endif

  PG_RETURN_BOOL(retval);
}

/* ----------------
 *   gloc_consistent
 *
 *   This is one of the 6 GiST routines. Given a GiST entry and a query (in this case,
 *   RANGE), this routine should determine whether this particular entry is consistent
 *   with the query with respect to the given strategy. StrategyNumber is specified in
 *   the header file and should be provided when defining the GiST access method
 *   in Postgresql.
 *
 *   Since the determination of consistency differs between leaf nodes and internal
 *   nodes, so this routine is implemented as a wrapper to two other routines. (for
 *   leaf node, internal node respectively)
 *
 *   parameters:
 *     entry - a given GiST node
 *     query - user-specified query (a range object)
 *     strategy - the strategy number (defined in range.h)
 *
 *   return:
 *     return true if this query is consistent with the given GiST entry; otherwise,
 *     return false.
 *
 *   note:
 *     If you're very sure that the entries under this given GiST node is not consistent
 *     with the query, you should always return FALSE. Returning TRUE is not harmful if
 *     subnodes are not consistent with the query, but the performance degradation is the
 *     penalty.
 *
 * ----------------
 */
Datum gloc_consistent(PG_FUNCTION_ARGS)
{
   GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
   LOC *query = (LOC*) PG_GETARG_POINTER(1);
   StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);

   // I don't know why this if is needed, just test!!! (saw it in a code)
   if (!(DatumGetPointer(entry->key) !=NULL && query ))
      PG_RETURN_BOOL(FALSE);

   if (GIST_LEAF(entry))
      return LOC_leaf_consistent( (LOCKEY*) DatumGetPointer(entry->key), query, strategy );
   else
      return LOC_internal_consistent( (LOCKEY*) DatumGetPointer(entry->key), query, strategy );
}

/* ----------------
 *   gloc_compress
 *  *   This is one of the six required GiST methods.
 *   We have to convert our LOC representation to LOCKEY representation.
 *
 *   parameters:
 *     entry - GiST entries
 *
 *   return:
 *     return the compressed form of the given GiST entry (which is itself)
 *
 * ----------------
 */
Datum gloc_compress(PG_FUNCTION_ARGS)
{
   GISTENTRY  *retval, *entry;
   LOCKEY *r;
   LOC *leaf;
   int length;

   entry = (GISTENTRY *) PG_GETARG_POINTER(0);

   if (entry->leafkey)
   {
     retval = (GISTENTRY*) palloc(sizeof(GISTENTRY));
     if (DatumGetPointer(entry->key)!=NULL)
     {
        leaf = (LOC*) DatumGetPointer(entry->key);
        length = sizeof(LOCKEY);
        r = (LOCKEY*) palloc(length);
        r->s_lower = leaf->sig;
        r->s_upper = leaf->sig;
        r->strand = leaf->strand;
        r->interval.lower = leaf->interval.lower;
        r->interval.upper = leaf->interval.upper;
        gistentryinit(*retval, PointerGetDatum(r), entry->rel, entry->page, entry->offset, length, FALSE);
     }
     else
     {
        gistentryinit(*retval, (Datum) 0, entry->rel, entry->page, entry->offset, 0, FALSE);
     }
   }
   else
     retval = entry;

   PG_RETURN_POINTER(retval);
}

/* ----------------
 *   gloc_decompress
 *
 *   This is one of the six required GiST methods.
 *   Since we do not compress any GiST nodes, so it does nothing.
 *
 *   parameters:
 *     entry - GiST entries
 *
 *   return:
 *     return the decompressed form of the given GiST entry (which is itself)
 *
 * ----------------
 */
Datum gloc_decompress(PG_FUNCTION_ARGS)
{
   PG_RETURN_POINTER(PG_GETARG_POINTER(0));
}

/* ----------------
 *   gloc_union
 *
 *   This is one of the six required GiST methods.
 *   It will be invoked with a number of GiST entries and should return a GiST
 *   entry which holds true for any of these entries.
 *
 *   In our implementation, we returns the minimal bounding range that encloses
 *   all the entries in the entryvec.
 *
 *   parameters:
 *     entryvec - GiST entries
 *     sizep - the size of a single range object.
 *
 *   return:
 *     return a range data object which holds true for all the given entries.
 *
 * ----------------
 */
Datum gloc_union(PG_FUNCTION_ARGS)
{
  GistEntryVector *entryvec = (GistEntryVector *) PG_GETARG_POINTER(0);
  int i, numranges, length;
  LOCKEY *cur, *out = NULL;
  int64 lower, upper;
  uint32 slower, supper;

  numranges = entryvec->n;

#ifdef DEBUG_LOC_GIST
  elog( LOG, "number of entries = %d", numranges);
#endif

  length = sizeof(LOCKEY);
  *(int *) PG_GETARG_POINTER(1) = length;

  cur = (LOCKEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[0].key));

  lower = cur->interval.lower;
  upper = cur->interval.upper;
  slower = cur->s_lower;
  supper = cur->s_upper;


  for (i = 1; i < numranges; i++)
  {
    cur = (LOCKEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[i].key));
    if (lower > cur->interval.lower) lower = cur->interval.lower;
    if (upper < cur->interval.upper) upper = cur->interval.upper;
    if (slower > cur->s_lower ) slower = cur->s_lower;
    if (supper < cur->s_upper ) supper = cur->s_upper;

  }

  out = (LOCKEY*) palloc( length ); 
  out->interval.lower = lower;
  out->interval.upper = upper;
  out->s_lower = slower;
  out->s_upper = supper;
  out->strand = '+';   // default value

  PG_RETURN_POINTER(out);
}

/* ----------------
 *   gloc_penalty
 *
 *   This is one of the six required GiST methods.
 *   GiST kernel requires us to give it domain penalty for insertion an
 *   entry into a subtree. This routine tells GiST what penalty value should
 *   it receive if GiST change the original entry into a new entry.
 *
 *   For now, I implemented the penalty as:
 *     the cost of enlarging the interval formed by signature +
 *   WEIGHT_FACTOR * the cost of enlarging the interval formed by range data.
 *
 *   parameters:
 *     origentry - original GiST entry
 *     newentry  - new GiST entry
 *
 *   return:
 *     return the penalty value
 *
 * ----------------
 */
Datum gloc_penalty(PG_FUNCTION_ARGS)
{
  LOCKEY *origentry = (LOCKEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(0))->key);
  LOCKEY *newentry = (LOCKEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(1))->key);
  float penalty, * result = (float *) PG_GETARG_POINTER(2);

  penalty = Max(newentry->s_upper-origentry->s_upper,0) + Max(origentry->s_lower-newentry->s_lower,0) +
            ( Max(newentry->interval.upper -origentry->interval.upper, 0) + 
              Max(origentry->interval.lower - newentry->interval.lower, 0) ) * 0.00000000001;

  *result = penalty;

  PG_RETURN_POINTER(result);
}


uint32 get_union_size( uint32 a1, uint32 a2, uint32 b1, uint32 b2)
{
   return (uint32) ( Max(a2,b2) - Min(a1,b1) );
}

uint32 get_inter_size( uint32 a1, uint32 a2, uint32 b1, uint32 b2)
{
   return (uint32) Abs( Min(a2,b2) - Max(a1,b1)  );
}


typedef struct
{
   LOCKEY *key;
   int    pos;
} KBsort;

int compare_KB(const void *a, const void *b);
int compare_KB(const void *a, const void *b)
{   
   LOCKEY *v1 = (LOCKEY*) (((KBsort*) a)->key);
   LOCKEY *v2 = (LOCKEY*) (((KBsort*) b)->key);

   // first, compare the signature range
   if (v1->s_lower<v2->s_lower)  return -1;
   if (v1->s_lower>v2->s_lower)  return 1;
   if (v1->s_upper<v2->s_upper)  return -1;
   if (v1->s_upper>v2->s_upper)  return 1;

   if (v1->interval.lower<v2->interval.lower) return -1;
   if (v1->interval.lower>v2->interval.lower) return 1;
   if (v1->interval.upper<v2->interval.upper) return -1;
   if (v1->interval.upper>v2->interval.upper) return -1;

   return 0;
}

/* ----------------
 *   gloc_picksplit
 *
 *   This is one of the six required GiST methods.
 *   When a node in the GiST search tree is full, we need to split this node into
 *   two. GiST requires us to dictate the split process.
 *
 *   In this implementation, we use Guttman's poly time split algorithm.
 *
 *   parameters:
 *     entryvec - entries in the old GiST node
 *     v        - the split vector. It has a rather complex structure, but the part
 *                that we need to know is
 *
 *     v->spl_ldatum, v->spl_rdatum  : pointers to left and right of the split nodes.
 *     v->spl_nleft,  v->spl_nright  : number of entries are there in each of the left, right node 
 *     v->spl_left,  v->spl_right   : the entry numbers which belong to the left, right node respectively
 *
 *   return:
 *     the GiST split vector structure.
 *
 * ----------------
 */

/* the following implementation is preserved, since it provides another way of implementation */
/*
Datum gloc_picksplit(PG_FUNCTION_ARGS)
{
  GistEntryVector *entryvec = (GistEntryVector *) PG_GETARG_POINTER(0);
  GIST_SPLITVEC *v = (GIST_SPLITVEC *) PG_GETARG_POINTER(1);
  OffsetNumber maxoff, i, j;

  LOCKEY *datum_l, *datum_r, *alpha, *beta;
  int  length, seed_1=0, seed_2=0;
  int64 ai_lower, ai_upper, bi_lower, bi_upper;
  uint32 alower, aupper, blower, bupper, size_l, size_r, size_alpha, size_beta;
  int size_waste, waste;
  bool firsttime;

  maxoff = entryvec->n - 1;
  length = (maxoff + 2) *sizeof(OffsetNumber);

  v->spl_left = (OffsetNumber *) palloc(length);
  v->spl_right = (OffsetNumber *) palloc(length);
  v->spl_nleft = 0;
  v->spl_nright = 0;

  firsttime = TRUE;
  waste = 0;

  for ( i=FirstOffsetNumber; i <maxoff; i = OffsetNumberNext(i) )
  {
    alpha = (LOCKEY*) DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[i].key);
    for( j=OffsetNumberNext(i); j<=maxoff; j=OffsetNumberNext(j))
    {
       beta = (LOCKEY*) DatumGetPointer(((GISTENTRY*) VARDATA(entryvec))[j].key);

       // compute the wasted space by union these guys
       // size_waste = size_union - size_inter

       size_waste = Abs( get_union_size(alpha->s_lower,alpha->s_upper,beta->s_lower,beta->s_upper) -
                         get_inter_size(alpha->s_lower,alpha->s_upper,beta->s_lower, beta->s_upper) );
       if ( size_waste >  waste || firsttime )
       {
          waste = size_waste;
          seed_1 = i;
          seed_2 = j;
          firsttime = FALSE;
       }
    }
  }

  // we split the node evenly into two nodes
  alpha = (LOCKEY*) DatumGetPointer(((GISTENTRY*) VARDATA(entryvec))[seed_1].key);
  beta  = (LOCKEY*) DatumGetPointer(((GISTENTRY*) VARDATA(entryvec))[seed_2].key);

  alower = alpha->s_lower;
  aupper = alpha->s_upper;
  blower = beta->s_lower;
  bupper = beta->s_upper;
  ai_lower = alpha->interval.lower;
  ai_upper = alpha->interval.upper;
  bi_lower = beta->interval.lower;
  bi_upper = beta->interval.upper;

  size_l = aupper-alower;
  size_r = bupper-blower;

  //maxoff = OffsetNumberNext(maxoff);
  for (i=FirstOffsetNumber; i<=maxoff; i=OffsetNumberNext(i) )
  {
     if (i==seed_1)
     {
       v->spl_left[v->spl_nleft++] = i;
       continue;
     }
     else if (i==seed_2)
     {
        v->spl_right[v->spl_nright++] = i;
        continue;
     }

     // now, consider which page needs least enlargement?
     alpha = (LOCKEY*) DatumGetPointer(((GISTENTRY*) VARDATA(entryvec))[i].key);
     size_alpha = get_union_size( alower, aupper, alpha->s_lower, alpha->s_upper );
     size_beta  = get_union_size( blower, bupper, alpha->s_lower, alpha->s_upper );

     if ( Abs(size_alpha-size_l) < Abs(size_beta-size_r) )
     {
        v->spl_left[v->spl_nleft++] = i;
        alower = Min( alower, alpha->s_lower );
        aupper = Max( aupper, alpha->s_upper );
        if ( ai_lower > alpha->interval.lower ) ai_lower = alpha->interval.lower;
        if ( ai_upper < alpha->interval.upper ) ai_upper = alpha->interval.upper;
        size_l = aupper - alower;
     }
     else
     {
        v->spl_right[v->spl_nright++] = i;
        blower = Min( blower, alpha->s_lower );
        bupper = Max( bupper, alpha->s_upper );
        if ( bi_lower > alpha->interval.lower ) bi_lower = alpha->interval.lower;
        if ( bi_upper > alpha->interval.upper ) bi_upper = alpha->interval.upper;
        size_r = bupper - blower;
     }
  }
     // sentinel value
     v->spl_left[v->spl_nleft] = v->spl_right[v->spl_nright] = FirstOffsetNumber;

  length = sizeof(LOCKEY);
  datum_r = (LOCKEY*) palloc(length);
  datum_l = (LOCKEY*) palloc(length);
  datum_l->s_lower = alower;
  datum_l->s_upper = aupper;
  datum_l->strand = '+';
  datum_r->s_lower = blower;
  datum_r->s_upper = bupper;
  datum_r->strand = '+';
  datum_l->interval.lower = ai_lower;
  datum_l->interval.upper = ai_upper;
  datum_r->interval.lower = bi_lower;
  datum_r->interval.upper = bi_upper;

  v->spl_ldatum = PointerGetDatum(datum_l);
  v->spl_rdatum = PointerGetDatum(datum_r);

  PG_RETURN_POINTER(v);
}
*/

Datum gloc_picksplit(PG_FUNCTION_ARGS)
{
  GistEntryVector *entryvec;
  GIST_SPLITVEC *v;
  OffsetNumber maxoff, i;
  int length;
  KBsort *keys;
  uint32 alower, aupper, blower, bupper;
  int64  ailower, aiupper, bilower, biupper;
  LOCKEY *datum_l, *datum_r, *cur;

#ifdef DEBUG_LOC_GIST
  elog(LOG, "picksplit called");
#endif

  entryvec = (GistEntryVector *) PG_GETARG_POINTER(0);
  v = (GIST_SPLITVEC *) PG_GETARG_POINTER(1);

  maxoff = entryvec->n - 1;
  length = (maxoff + 2) *sizeof(OffsetNumber);
  v->spl_left = (OffsetNumber *) palloc(length);
  v->spl_right = (OffsetNumber *) palloc(length);
  v->spl_nleft = 0;
  v->spl_nright = 0;

  keys = (KBsort*) palloc( sizeof(KBsort) * (maxoff+1) );
  for(i=FirstOffsetNumber; i<=maxoff; i = OffsetNumberNext(i))
  {
     cur = (LOCKEY*) DatumGetPointer(((GISTENTRY*) VARDATA(entryvec))[i].key);
     keys[i].key = cur;
     keys[i].pos = i;
  }
  qsort((void*) &keys[FirstOffsetNumber], maxoff-FirstOffsetNumber+1, sizeof(KBsort), compare_KB);

  cur = keys[FirstOffsetNumber].key;
  alower = cur->s_lower;
  aupper = cur->s_upper;
  ailower = cur->interval.lower;
  aiupper = cur->interval.upper;
  cur = keys[(maxoff-FirstOffsetNumber+1)/2+1].key;
  blower = cur->s_lower;
  bupper = cur->s_upper;
  bilower = cur->interval.lower;
  biupper = cur->interval.upper;

  for(i=FirstOffsetNumber; i<=maxoff; i=OffsetNumberNext(i) )
  {
     cur = keys[i].key;
     if ( i<=(maxoff -FirstOffsetNumber+1)/2 )
     {
        v->spl_left[v->spl_nleft] = keys[i].pos;
        v->spl_nleft++;
        if ( alower>cur->s_lower ) alower = cur->s_lower;
        if ( aupper<cur->s_upper ) aupper = cur->s_upper;
        if ( ailower>cur->interval.lower ) ailower = cur->interval.lower;
        if ( aiupper<cur->interval.upper ) aiupper = cur->interval.upper;
     }
     else
     {
        v->spl_right[v->spl_nright] = keys[i].pos;
        v->spl_nright++;
        if ( blower>cur->s_lower ) blower = cur->s_lower;
        if ( bupper<cur->s_upper ) aupper = cur->s_upper;
        if ( bilower>cur->interval.lower ) bilower = cur->interval.lower;
        if ( biupper<cur->interval.upper ) biupper = cur->interval.upper;
     }
  }

  length = sizeof(LOCKEY);
  datum_r = (LOCKEY*) palloc(length);
  datum_l = (LOCKEY*) palloc(length);
  datum_l->s_lower = alower;
  datum_l->s_upper = aupper;
  datum_l->strand = '+';
  datum_r->s_lower = blower;
  datum_r->s_upper = bupper;
  datum_r->strand = '+';
  datum_l->interval.lower = ailower;
  datum_l->interval.upper = aiupper;
  datum_r->interval.lower = bilower;
  datum_r->interval.upper = biupper;

  v->spl_ldatum = PointerGetDatum(datum_l);
  v->spl_rdatum = PointerGetDatum(datum_r);
  pfree(keys);

  PG_RETURN_POINTER(v);

}

/* ----------------
 *   gloc_same
 *
 *   This is one of the six required GiST methods.
 *   GiST kernel requires us to tell it whether two objects are the same.
 *   (since GiST does not know the internal representation of the user datatype)
 *
 *   In this implementation, if the interval defined by these two objects are
 *   equal, then they are the same
 *
 *   parameters:
 *     b1 - loc datatype
 *     b2 - loc datatype
 *     result - the result (pass by pointer )
 *
 *   return:
 *     return the boolean value (pass by pointer)
 *
 * ----------------
 */
Datum gloc_same(PG_FUNCTION_ARGS)
{
  LOCKEY *b1 = (LOCKEY *) PG_GETARG_POINTER(0);
  LOCKEY *b2 = (LOCKEY *) PG_GETARG_POINTER(1);
  bool *result = (bool *) PG_GETARG_POINTER(2);
  bool v;

  v = TRUE;

  if (b1->s_lower!=b2->s_lower) v=FALSE;
  else if (b1->s_upper!=b2->s_upper) v=FALSE;
  else if (b1->strand!=b2->strand) v=FALSE;
  else if (b1->interval.lower!=b2->interval.lower) v=FALSE;
  else if (b1->interval.upper!=b2->interval.upper) v=FALSE;

  *result = v;

  PG_RETURN_POINTER(result);
}

/* ----------------
 *   bloc_ops_cmp
 *
 *   This is for the btree op class.
 *
 *   In order to enable the use of ORDER BY, GROUP BY, DISTINCT in SQL
 *   Postgresql needs datatype implementation to have this.
 *
 *   parameters:
 *     a - loc datatype
 *     b - loc datatype
 *
 *   return:
 *     return integer. -1 means smaller than, 0 means equal, 1 means larger than
 *
 * ----------------
 */
Datum bloc_ops_cmp(PG_FUNCTION_ARGS)
{
  LOC *a = (LOC*) PG_GETARG_POINTER(0);
  LOC *b = (LOC*) PG_GETARG_POINTER(1);
  int retval;

  if ( (retval=(a->sig==b->sig)) ) 
    if ( (retval=(a->strand==b->strand)) )
      retval = range_cmp( &(a->interval), &(b->interval) );

  PG_RETURN_INT32(retval);
}
