/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * network.c
 *
 *    This file contains network-related and http protocol-related routines 
 * called by Postgresql.
 *
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */

#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include "postgres.h"
#include "network.h"

#define iMAX(a,b)	((a>b) ? (a) : (b))

char http_special_char[] = "$&+,/;:=?@ <>";

/* ----------------
 *   get_inet_addr
 *
 *   This is the routine used to translate string host name into 4-byte
 *   inet_addr. This kind of translation is needed when host name is specified
 *   in the string format. Since DNS lookup could be slow, a DNS cache is also
 *   implemented. The DNS lookup is archieved through recording (host->ip) mapping.
 *   
 *   A regular cache should also have a replacement policy. Our DNS cache,however, 
 *   does not have it. The reason for that is because we expect our package only
 *   connects to NCBI server and it is not a general network client (like a browser),
 *   we can safely skip this part.
 *
 *   parameters:
 *     host - this is a string containing the server address in string format or in
 *            number format. If it's in string format, we need to do the DNS lookup.
 *   return:
 *     returns the ip4_address data for internal use.
 *
 * ----------------
 */


ip4_address get_inet_addr ( char *host )
{
  ip4_address inaddr;
  struct hostent *hp;

  //elog( LOG, "[G] get_inet_addr invoked with host=[%s]", host );
  //elog( LOG, "original map[%s]", host_ip_map[0].host );

  if ( (inaddr=inet_addr(host)) == INADDR_NONE )
  {
    /* dns cache hit */
    if ( ( host_ip_map[0].host[0] !=0 ) && !strcasecmp( host_ip_map[0].host, host ) )
    {
       //elog( LOG, "DNS hit map=[%s]", host_ip_map[0].host );
       return host_ip_map[0].ip;
    }    

    //elog( LOG, "DNS miss map=[%s]", host_ip_map[0].host );

    if ( (hp = gethostbyname(host)) == NULL )
      return INADDR_NONE;

    bcopy( hp->h_addr, (char*) &inaddr, sizeof(ip4_address) );

    //elog( LOG, "DNS miss: original host[%s]", host_ip_map[0].host);

    strcpy( host_ip_map[0].host, host);
    host_ip_map[0].ip = inaddr;
  }

  return inaddr;
}

/* ----------------
 *   read_bytes
 *
 *   This is the routine used to receive data from network.
 *   The number of bytes to read is specified and a buffer for the data should
 *   be allocated before calling this function.
 *
 *   parameters:
 *     fd - this should be a valid socket file descriptor.
 *     pivot - buffer to store the data from network
 *     nbyte - maximum number of bytes to read from the network
 *
 *   return:
 *     returns the number of bytes read from network
 *
 * ----------------
 */

int read_bytes ( int fd, char *pivot, int nbytes )
{
  int nleft, nread;

  nleft = nbytes;
  while ( nleft > 0 )
  {
    nread = read( fd, pivot, nleft );
    if ( nread < 0 )    /* error */
      return nread;
    else if ( nread == 0 ) /* end of file */
      break;

    nleft -= nread;
    pivot += nread;
  }

  return (nbytes-nleft);
}

/* ----------------
 *   write_bytes
 *
 *   This is the routine used to send data to network.
 *   The number of bytes to write is specified and a buffer for the data should
 *   be provided and filled with data before calling this function.
 *
 *   parameters:
 *     fd - this should be a valid socket file descriptor.
 *     byte_array - data buffer
 *     nbyte - maximum number of bytes to write to network
 *
 *   return:
 *     returns the number of bytes written to network
 *
 * ----------------
 */

int write_bytes ( int fd, char *byte_array, int nbytes )
{
  int nleft, nwritten;

  nleft = nbytes;
  while ( nleft > 0 )
  {
    nwritten = write( fd, byte_array, nleft );
    /* error */
    if ( nwritten <= 0 )   return nwritten;  

    nleft -= nwritten;
    byte_array += nwritten;
  }

  return (nbytes-nleft);
}

/* ----------------
 *   write_line
 *
 *   This is the routine used to send data (appened with a new line) to
 *   the network. In HTTP protocol, a new line is defined as a combination
 *   of a carriage return and a line feed. 
 *
 *   parameters:
 *     fd - this should be a valid socket file descriptor.
 *     string - data buffer.
 *
 *   note:
 *     this routine does not check if the sending goes wrong. It simply assumes
 *     the connection is established and safe to send data.
 * ----------------
 */

void write_line ( int fd, char *string )
{
  write_bytes( fd, string, strlen(string) );
  write_bytes( fd, CRLF, 2 );
}

/* ----------------
 *   url_encode
 *
 *   This is the routine used to encode url string sent to http server.
 *   In HTTP protocol, url string is not allowed to contain certain special
 *   characters ( defined in http_special_char ). If they are included in
 *   the query string, we need to encode them by using a very simple scheme -
 *   change them into their hexidecimal represetations.
 *
 *   parameters:
 *     src  - original string
 *     dest - encoded string
 *
 *   note:
 *     dest should be allocated with enough space to store the encoded string
 *     before this routine is invoked.
 * ----------------
 */

char* UrlEncode(char *src, char *dest)
{
   char *p;
   int ascii, value, pos = 0;

   while ( *src )
   {
      /* check if special character is encountered */
      if ( (p=strchr(http_special_char, *src))!=NULL )
      {
         /*
          *   encoding scheme:
          *   suppose the target character's hexdecimal
          *   represetation is 0x9E, then we simply output '%9E'.   */

          ascii = *p;
          dest[pos++] = '%';

          value = ascii / 16;
          if ( value > 9 )  value = 'A' + value-9;
          else  value += '0';

          dest[pos++] = value;
          value = ascii % 16;

          if ( value > 9 ) value = 'A' + value-9;
          else  value += '0';

          dest[pos++] = value;
      }
      else
         dest[pos++] = *src;

      src++;
   }

   dest[pos] = '\0';
   return dest;
}


/* ----------------
 *   connect_alarm
 *
 *   This is the routine used to implement the timeout mechanism for connect().
 *   TCP/IP connection establishment has a long default timeout value which is
 *   untolerable. We use the timer signal to implement a shorter timeout for
 *   socket connection.
 *
 *   parameters:
 *     signo - timeout signal number
 * ----------------
 */


static void connect_alarm( int signo )
{
}


// changes starting here
static char http_methods[][10] = { "GET", "POST" };

char *HttpContextGetMethod(HttpContext context)
{
   return http_methods[context->method];
}

HttpContext CreateHttpContext(HttpMethod method,  int num_cgi_params , char *agent)
{
   HttpContext hc;

   hc = (HttpContext) palloc(sizeof(HttpContextData));
   hc->socket = InvalidSocket;
   hc->method = method;
   if ( num_cgi_params < 1 )  num_cgi_params = 1;
   hc->parameters = CreateCgiParameters( num_cgi_params );

   if ( agent )
      hc->user_agent = pstrdup(agent);
   else
      hc->user_agent = pstrdup(DefaultUserAgent);

   return hc;
}

void FreeHttpContext(HttpContext c)
{
   if ( HttpContextIsValid(c) )
   {
      if ( c->user_agent )  pfree(c->user_agent);
      FreeCgiParameters(c->parameters);
      pfree(c);
   }
}

void SetCgiParameters(HttpContext context, CgiParameters params)
{
   if ( HttpContextNotValid(context) )  return;

   if ( CgiParametersIsValid(context->parameters) )
      FreeCgiParameters(context->parameters);

   context->parameters = params;
}

void HttpContextSetParameter( HttpContext context, char *name, char *value )
{
   if ( HttpContextNotValid(context) ) return;
   context->parameters = CgiParametersAdd( context->parameters, name, value );
}

CgiParameters CreateCgiParameters(int capacity)
{
   int size;
   CgiParameters cp;

   capacity = iMAX(capacity, 1);
   size = CalculateCgiParameterSize(capacity);
   cp = (CgiParameters) palloc(size);

   cp->capacity = capacity;
   cp->size = 0;

   return cp;
}

void FreeCgiParameters(CgiParameters cp)
{
   int i;

   if ( CgiParametersNotValid(cp) )  return;

   for(i=0;i<cp->size;i++)
   {
      pfree(cp->items[i].name);
      pfree(cp->items[i].value);
   }
   pfree(cp);
}

char *CgiParametersLookup(CgiParameters cp, char *name, int *offset )
{
   int i;

   if ( CgiParametersNotValid(cp) )
      return NULL;

   for(i=0;i<cp->size;i++)
      if ( strcasecmp(cp->items[i].name,name)==0 )
      {
         *offset = i;
         return cp->items[i].value;
      }

   return (char *) NULL;
}

CgiParameters CgiParametersAdd(CgiParameters cp, char *name, char *value)
{
   int offset, size;

   if ( CgiParametersNotValid(cp) )   return (CgiParameters) NULL;

   if ( CgiParametersLookup(cp,name,&offset)!=NULL )
   {
      pfree(cp->items[offset].value);
      cp->items[offset].value = pstrdup(value);
   }
   else
   {
      if ( cp->capacity==cp->size )
      {
         size = CalculateCgiParameterSize(2*cp->capacity);

         cp = (CgiParameters) repalloc(cp,size);
         cp->capacity = 2*cp->capacity;
      }

      cp->items[cp->size].name  = pstrdup(name);
      cp->items[cp->size].value = pstrdup(value);
      cp->size++;

   }

   return cp;
}

void CgiParametersList(CgiParameters cp)
{
   int i;

   if ( CgiParametersNotValid(cp) )   return;
   for(i=0;i<cp->size;i++)
     printf("%s => %s\n", cp->items[i].name, cp->items[i].value );

}

#define BUFFER_SIZE   1025
#define INIT_CAPACITY 128
#define LIMIT_SIZE    256

char *CgiParametersRepr( CgiParameters cp )
{
   char buffer[BUFFER_SIZE];
   char *ret, *data;
   int i, capacity, length, extra_storage;
   int size;

   /* we start from a smaller size */
   if ( CgiParametersNotValid(cp) )  return (char *) NULL;
   capacity = INIT_CAPACITY;
   length = 0;
   ret = (char *) palloc(capacity);

   for(i=0;i<cp->size;i++)
   {
      extra_storage = 0;

      /* we are assuming that the size of header is not too big */
      UrlEncode( cp->items[i].name, buffer );
      size = strlen( buffer );

      /* detect if we are out of memory here !!! */
      if ( (capacity-length) < (size+2) )
      {
         capacity = iMAX(2 * capacity, capacity-length+size+2);
         ret = repalloc(ret, capacity);
      }

      memcpy( (char *) (ret+length), buffer, size);
      length += size;
      ret[length++] = '=';

      /* we are being extremely cautious here */
      if ( (size=strlen(cp->items[i].value)) > LIMIT_SIZE )
      {
         extra_storage = 1;
         data = (char *) palloc( size*4+1 );
      }
      else
         data = buffer;

      UrlEncode( cp->items[i].value, data );
      size = strlen(data);

      /* detect if we are out of memory here !!! */
      if ( (capacity-length) < (size+2) )
      {
         capacity = iMAX(2 * capacity,capacity-length+size+2);
         ret = repalloc(ret, capacity);
      }

      memcpy( (char *) (ret+length), data, size);
      length += size;
      ret[length++] = '&';

      if ( extra_storage )  pfree(data);
   }

   if (length>0)
      ret[length-1] = 0;
   else
   {
      pfree(ret);
      ret = NULL;
   }

   return ret;
}

char* HttpReceiveRequest( HttpContext context )
{
   char *ret;
   char buffer[1024];
   int capacity, length, size;

   capacity = 1025;
   length = 0;

   ret = (char *) palloc( 1025 );
   ret[0] = 0;

   while (1)
   {
      size = read_bytes( context->socket, buffer, 1024 );

      if ( size<0 ) return ret;

      if ( length+size+1>capacity )
      {
         capacity = iMAX( 2*capacity, length+size+1 );
         ret = repalloc( ret, capacity );
      }

      memcpy( ret+length, buffer, size );
      length += size;
      ret[length] = 0;

      if ( size<1024 ) break;
   }

   close(context->socket);
   context->socket = InvalidSocket;

   return ret;
}

/* ----------------
 *   http_get_request
 *
 *   This is the routine for requesting a HTTP get command for remote server.
 *   It sets up the TCP/IP connection and issues HTTP GET command according to
 *   HTTP protocol specification. Default connection parameters are used.
 *
 *   parameters:
 *     host - server host name
 *     port - server port number
 *     file - server cgi program name + parameter strings
 *
 *   return:
 *     it returns the socket file descriptor if connection is established.
 *
 *   note:
 *     After calling this routine, you have to check if a connection has already
 *     established by ( CONNECTION_ESTABLISHED() macro ). If so, a HTTP connection
 *     has been made to the server, you can start to retrieve the data from the
 *     server.
 *
 * ----------------
 */


HttpResult HttpSendRequest( HttpContext context, char *host, int port, char *file )
{
   sig_t default_handler;
   struct sockaddr_in serv_addr;
   char *parameters, buffer[256];
   int fdescriptor, length;

   //elog(LOG, "[HS] HttpSendRequest invoked" );

   if ( HttpContextNotValid(context) )   return SOCKET_ERROR;

   memset( &serv_addr, 0, sizeof(char) );
   serv_addr.sin_family = AF_INET;
   serv_addr.sin_addr.s_addr = get_inet_addr(host);
   serv_addr.sin_port = htons( port );

   if ( (fdescriptor=socket( AF_INET, SOCK_STREAM, 0)) < 0 )  return SOCKET_ERROR;

   default_handler = signal( SIGALRM, connect_alarm );
   alarm( DEFAULT_CONNECT_TIMEOUT );

   if ( connect(fdescriptor, (struct sockaddr *) &serv_addr, sizeof(serv_addr) ) < 0 )
   {
      close(fdescriptor);
      if  ( errno==EINTR )  errno = ETIMEDOUT;
      return CONNECTION_ERROR;
   }

   alarm(0);
   signal( SIGALRM, default_handler );

   context->socket = fdescriptor;

   parameters = CgiParametersRepr(context->parameters);

   if ( (context->method==GET) && (parameters!=NULL) )
   {
      sprintf( buffer, "%s %s?%s HTTP/1.0", HttpContextGetMethod(context), file, parameters );
   }
   else
   {
      sprintf( buffer, "%s %s HTTP/1.0", HttpContextGetMethod(context), file );
   }

   write_line( fdescriptor, buffer );
   sprintf( buffer, "User-Agent: %s", context->user_agent );
   write_line( fdescriptor, buffer );
   write_line( fdescriptor, "Connection: close" );

   if ( context->method==POST )
   {
      length = strlen(parameters);
      sprintf( buffer, "Content-length: %d", length );
      write_line( fdescriptor, buffer );
      write_line( fdescriptor, "Content-type: application/x-www-form-urlencoded" );
      write_line( fdescriptor, "");
      write_bytes( fdescriptor, parameters, length );
   }
   else
      write_line( fdescriptor, "" );

   return HTTP_OK;
}


// changes ending here
