/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * network.h
 *
 *    Definitions of functions for network connectivities.
 *
 * This file is included in network.c and all other files which require
 * network connections.
 *
 * I've been debating whether to use libwww program or to keep using this
 * implementation. libwww is powerful but huge. Using it also increases the
 * complexity of installation because I am not sure every linux system comes
 * with it. On the other hand, this network connectivity implementation is
 * not thread-safe and obviously less stable than libwww program.
 *
 * status: still trying to find the balance.
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */

#ifndef NETWORK_H
#define NETWORK_H

#ifndef INADDR_NONE
#define INADDR_NONE   0xFFFFFFFF
#endif

#define   InvalidSocket   -1
#define   SocketIsValid(s)   ((s)!=InvalidSocket)

/* useful type and definition used in socket programming */
typedef unsigned long ip4_address;

/* http data buffer for connection. 
 * this size should be enough for a typical web page.
 * special care should be taken care of if the target server could return a page 
 * larger than this. 
 */
#define DEFAULT_HTTP_BUFFER_SIZE   4*1024*1024
#define DEFAULT_CONNECT_TIMEOUT    10
#define CRLF                       "\r\n"

typedef enum HttpMethod
{
   DEFAULT = 0, GET = 0,
   POST,
   HEAD
}  HttpMethod;

typedef struct CgiParameterData
{
   int capacity,        /* total number of buckets */
       size;            /* total number of items */
   struct
   {
      char *name,
           *value;
   } items[1];
} CgiParameterData;

#define CalculateCgiParameterSize(size)  (sizeof(CgiParameterData)+2*((size)-1)*sizeof(char*))

typedef CgiParameterData *CgiParameters;

#define DefaultUserAgent    "Mozilla/1.0.4"

typedef struct HttpContextData
{
   int socket;          /* socket used in this connection */
   HttpMethod method;
   CgiParameters parameters;
   char *user_agent;
} HttpContextData;

typedef HttpContextData *HttpContext;

#define InvalidHttpContext      NULL
#define HttpContextIsValid(h)   ((h)!=NULL)
#define HttpContextNotValid(h)  ((h)==NULL)
#define CgiParametersIsValid(cp)        ((cp)!=NULL)
#define CgiParametersNotValid(cp)       ((cp)==NULL)

typedef enum HttpResult
{
  SOCKET_ERROR     = -10,
  CONNECTION_ERROR = -9,
  HTTP_OK          = 1
} HttpResult;

HttpContext CreateHttpContext(HttpMethod method, int num_cgi_params, char *agent);
void FreeHttpContext(HttpContext);
void SetCgiParameters(HttpContext context, CgiParameters params);
CgiParameters CreateCgiParameters(int size);
void FreeCgiParameters(CgiParameters cp);
CgiParameters CgiParametersAdd(CgiParameters cp, char *name, char *value);
char *CgiParametersLookup(CgiParameters cp, char *name, int *offset );
char *CgiParametersRepr(CgiParameters cp);
char* UrlEncode(char *src, char *dest);
void HttpContextSetParameter(HttpContext hc, char *name, char *value);

char *HttpReceiveRequest(HttpContext context);
HttpResult HttpSendRequest(HttpContext context, char *host, int port, char *file);
char *HttpContextGetMethod(HttpContext context);
void CgiParametersList(CgiParameters cp);

// changes ending here


/*  http clients are supposed to maintain a DNS cache to avoid slow DNS lookup
 *  everytime, since a single logical webpage can contain several files (such as
 * .html, .gif...) in the same host. */

typedef struct _host_ip_pair {
  char host[100];
  ip4_address ip;
} host_ip_pair;

/*  since we expect that this package only interacts with NCBI website, 
 *  we only need to cache the only one host ip */
host_ip_pair host_ip_map[1];

/* functions used internally */
ip4_address get_inet_addr ( char * );
int read_bytes ( int, char*, int );
int write_bytes ( int, char*, int );
void write_line ( int, char* );
//char* url_encode ( char *, char * );
//int http_post_request( char*, int, char*, char* );
//int http_get_request( char*, int, char* );
//int search_query_param( query_param*, char* );
//void put_query_param( query_param*, char*, char* );
//void free_query_param( query_param* );
//char* generate_query_string( query_param* );

#endif  /* NETWORK.H */
