/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

#include "postgres.h"

#include <math.h>
#include <string.h>

#include "fmgr.h"
#include "funcapi.h"
#include "executor/spi.h"
#include "miscadmin.h"
#include "utils/builtins.h"
#include "utils/guc.h"
#include "utils/lsyscache.h"
#include "lib/stringinfo.h"

#include "partition.h"
#include "global.h"
#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))

typedef struct
{
  SPITupleTable *spi_tuptable;
  int nItems;
  int iqueryid, isubjectid;
  RANGE_int iseq1, iseq2;
} partition_fctx;

typedef struct _Intseg
{
  int64 lower, upper;
} Intseg;


/* helper functions listed here
 * note: they must be declared as static */

typedef struct _node
{
  int n;
  struct _node *next;
} node;

static node tail = { 0, &tail };

static node* create(void)
{
  node *tmp;

  tmp = (node*) malloc(sizeof(node));
  tmp->n = -1;
  tmp->next = &tail;
  return tmp;
}

static node* insert( node* h, int v )
{
  node *p = h, *q = h->next;
  node *tmp;
  int done = 0;

  while ( (q->next!=q) && !done )
  {
    if (q->n==v) done = 1;
    else if (q->n>v) break;
    else
    {
      p = q;
      q = q->next;
    }
  }
  if (!done)
  {
    tmp = (node*) malloc(sizeof(node));
    tmp->n = v;
    p->next = tmp;
    tmp->next = q;
  }

  return p;
}

static void destroy( node *p )
{
  node* q;

  while ( p!=&tail )
  {
     q=p;
     p=p->next;
     free(q);
  }
}

static void set_interval( HeapTuple t, TupleDesc d, int id, int lower, int upper )
{
  Intseg tmp;
  bool isnull;
  unsigned char *pointer;

  tmp.lower = lower;
  tmp.upper = upper;

  pointer = (unsigned char*) SPI_getbinval( t,d, id, &isnull );
  memcpy(pointer, &tmp, sizeof(Intseg));
}

static void get_interval( HeapTuple t, TupleDesc d, int id, RANGE_int* lower, RANGE_int* upper )
{
  bool isnull;
  unsigned char *pointer;

  pointer = (unsigned char*) SPI_getbinval( t, d, id, &isnull );
  if ( !isnull )
  {
     Intseg *seg = (Intseg*) pointer;
     *lower = seg->lower;
     *upper = seg->upper;
  } 
  else
  {
     *lower = 0;
     *upper = 0;
  }
}

static void partition_and_output( MemoryContext mc, Tuplestorestate *tupstore,
                SPITupleTable *spi_tuptable, TupleDesc spi_tupdesc,
                node *list, int iseq1, int iseq2, int start, int end)
{
  MemoryContext context_backup;
  HeapTuple new_tuple;
  RANGE_int query_lower, query_upper, sub_lower, sub_upper, lower, upper;
  int i;
  RANGE_int projected_lower, projected_upper;
  float ratio;
  node *cut_points;

  for (i=start;i<=end;i++)
  {
     cut_points = list->next;
     get_interval( spi_tuptable->vals[i], spi_tupdesc, iseq1, &query_lower, &query_upper );
     get_interval( spi_tuptable->vals[i], spi_tupdesc, iseq2, &sub_lower, &sub_upper );

     ratio = (float) (sub_upper-sub_lower) / (query_upper-query_lower);
     lower = query_lower;
     upper = query_upper;

     while ( cut_points->next!=cut_points )
     {
       if ( cut_points->n<=lower) cut_points=cut_points->next;
       else if ( cut_points->n>=upper) break;
       else
       {
          new_tuple = SPI_copytuple(spi_tuptable->vals[i]);
          set_interval(new_tuple, spi_tupdesc, iseq1, lower, cut_points->n);

          projected_lower = ratio * (lower-query_lower) + sub_lower;
          projected_upper = ratio * (cut_points->n-query_lower) + sub_lower;

          set_interval(new_tuple, spi_tupdesc, iseq2, projected_lower, projected_upper);

          context_backup = MemoryContextSwitchTo(mc);
          tuplestore_puttuple(tupstore, new_tuple);
          MemoryContextSwitchTo(context_backup);
          heap_freetuple(new_tuple);

          lower = cut_points->n;
          cut_points=cut_points->next;
       }
     }
     if ( lower<upper )
     {
       new_tuple = SPI_copytuple(spi_tuptable->vals[i]);
       set_interval(new_tuple, spi_tupdesc, iseq1, lower, upper);
       projected_lower = ratio*(lower-query_lower) + sub_lower;
       projected_upper = sub_upper;
       set_interval(new_tuple, spi_tupdesc, iseq2, projected_lower, projected_upper);

       context_backup = MemoryContextSwitchTo(mc);
       tuplestore_puttuple(tupstore, new_tuple);
       MemoryContextSwitchTo(context_backup);
       heap_freetuple(new_tuple);
     }
  }
}


PG_FUNCTION_INFO_V1(partition);
Datum partition(PG_FUNCTION_ARGS)
{
  FuncCallContext *funcctx;
  partition_fctx *fctx;
  ReturnSetInfo *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
  MemoryContext per_query_ctx;
  TupleTableSlot *slot;
  AttInMetadata *attinmeta;
  MemoryContext oldcontext;
  HeapTuple spi_tuple;
  SPITupleTable *spi_tuptable=NULL;
  TupleDesc spi_tupdesc, ret_tupdesc;
  node *partition_points, *tmp_partition_points;
  char *pointer;
  int i, number_of_items;
  int iqueryid, iseq1, iseq2;
  int batch_start;
  char *current_query;
  RANGE_int lower, upper;
  Tuplestorestate *tupstore = NULL;
  MemoryContext context_backup;


  /* check to see if caller supports us returning a tuplestore */
  if (!rsinfo || !(rsinfo->allowedModes & SFRM_Materialize))
    ereport( ERROR, (errcode(ERRCODE_SYNTAX_ERROR),
                     errmsg("materizlize mode required, but it is not allowed in this context")));
  per_query_ctx = rsinfo->econtext->ecxt_per_query_memory;
  oldcontext = MemoryContextSwitchTo(per_query_ctx);
  ret_tupdesc = CreateTupleDescCopy(rsinfo->expectedDesc);

  /* stuff done only on the first call of the function */
  if (SRF_IS_FIRSTCALL())
  {
    TupleDesc tupdesc = NULL; 
    int ret;
    int proc;
    StringInfo sql = makeStringInfo();
    char *relname = GET_STR( PG_GETARG_TEXT_P(0) );
    char *queryid = GET_STR( PG_GETARG_TEXT_P(1) );
    char *seq1 = GET_STR( PG_GETARG_TEXT_P(2) );
    char *subjectid = GET_STR( PG_GETARG_TEXT_P(3) );
    char *seq2 = GET_STR( PG_GETARG_TEXT_P(4) );


    /* create a function context for cross-call persistence */
    funcctx = SRF_FIRSTCALL_INIT();

    /* switch to memory context appropriate for multiple function calls */
    oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

    /* connect to SPI mamager */
    if ( (ret=SPI_connect()) < 0 )
      /* internal error */
      elog( ERROR, "partition: SPI_connect returned %d", ret );

    appendStringInfo(sql, "SELECT * FROM %s ORDER BY %s,%s", relname, queryid, seq1);

    ret = SPI_exec( sql->data, 0 );
    proc = SPI_processed;


    /* check for qualifying tuples */
    if ( (ret==SPI_OK_SELECT) && (proc>0) )
    {
      spi_tuptable = SPI_tuptable;
      spi_tupdesc = spi_tuptable->tupdesc;
    }
    else
    {
      /* no qualifying tuples */
      SPI_finish();
      SRF_RETURN_DONE(funcctx);
    }

    /* SPI switches context on us, so reset it */
    MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

    /* build the output tuple desction */
    tupdesc = CreateTupleDescCopy(spi_tupdesc);

    /* allocate a slot for a tuple with this tupdesc */
    slot = TupleDescGetSlot(tupdesc);

    /* assign slot to function context */
    funcctx->slot = slot;

    /* generate attribute metadata needed later to produce tuples from raw C strings */
    attinmeta = TupleDescGetAttInMetadata(tupdesc);
    funcctx->attinmeta = attinmeta;

    fctx = (partition_fctx *) palloc(sizeof(partition_fctx));

    funcctx->user_fctx = fctx;
    fctx->spi_tuptable = spi_tuptable;
    fctx->nItems = proc;
    fctx->iqueryid = SPI_fnumber( spi_tupdesc, queryid );
    fctx->isubjectid = SPI_fnumber( spi_tupdesc, subjectid );
    fctx->iseq1 = SPI_fnumber( spi_tupdesc, seq1 );
    fctx->iseq2 = SPI_fnumber( spi_tupdesc, seq2 );
    funcctx->max_calls = 1;

    SPI_finish();   
    MemoryContextSwitchTo(oldcontext);
  }

  /* stuff done on every call of the function */
  funcctx = SRF_PERCALL_SETUP();

  /* initialize per-call variables */
  /* return slot for our tuple */
  slot = funcctx->slot;
  fctx = (partition_fctx*) funcctx->user_fctx;
  spi_tuptable = fctx->spi_tuptable;
  /* the sql tuple */
  spi_tupdesc = spi_tuptable->tupdesc;
  attinmeta = funcctx->attinmeta;
  ret_tupdesc = attinmeta->tupdesc;

  /* since we set the max_calls to 1, this will only be done once */
  number_of_items = fctx->nItems;
  iqueryid = fctx->iqueryid;
  iseq1 = fctx->iseq1;
  iseq2 = fctx->iseq2;
  batch_start = 0;

  tupstore = NULL;
  context_backup = MemoryContextSwitchTo(per_query_ctx);
  tupstore = tuplestore_begin_heap(true,false,work_mem);
  MemoryContextSwitchTo(context_backup);

  /* retrieve the field number of "queryid", "subjectid", "seq1", "seq2" */
  partition_points = create();  
  get_interval( spi_tuptable->vals[0], spi_tupdesc, iseq1, &lower, &upper );
  tmp_partition_points = insert( partition_points, lower );
  insert( tmp_partition_points, upper );
  current_query = SPI_getvalue(spi_tuptable->vals[0], spi_tupdesc, iqueryid );

  /* create the break points of the intervals and reorganize all the tuples */
  for(i=1; i<number_of_items; i++)
  {
     spi_tuple = spi_tuptable->vals[i];
     pointer = SPI_getvalue(spi_tuple, spi_tupdesc, iqueryid );

     /* check if we encounter the same query id */
     if ( strcmp(current_query,pointer) != 0 )
     {
       partition_and_output( per_query_ctx, tupstore, spi_tuptable, spi_tupdesc,
                             partition_points, iseq1, iseq2, batch_start, i-1 );
       destroy(partition_points);
       partition_points = create();
       batch_start = i;
       current_query = SPI_getvalue(spi_tuptable->vals[i], spi_tupdesc, iqueryid );
     }
     else
     {
        get_interval( spi_tuple, spi_tupdesc, iseq1, &lower, &upper );
        tmp_partition_points = insert( partition_points, lower );
        insert( tmp_partition_points, upper );
     }
  }

  partition_and_output( per_query_ctx, tupstore, spi_tuptable, spi_tupdesc,
                        partition_points, iseq1, iseq2, batch_start, i-1 );
  destroy(partition_points);

  SPI_finish();

  context_backup = MemoryContextSwitchTo(per_query_ctx);
  tuplestore_donestoring(tupstore);
  MemoryContextSwitchTo(context_backup);

  rsinfo->setResult = tupstore;
  rsinfo->setDesc = ret_tupdesc;
  SRF_RETURN_DONE(funcctx);
  return (Datum) 0;
}


