/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * range.c
 *
 *    This file contains routines for range  data type.
 *
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */

#include "postgres.h"
#include "access/gist.h"
#include "access/rtree.h"
#include "utils/builtins.h"
#include "range.h"

#define max(a,b)		((a) >	(b) ? (a) : (b))
#define min(a,b)		((a) <= (b) ? (a) : (b))
#define abs(a)			((a) <	(0) ? (-a) : (a))

//#define DEBUG_RANGE	TRUE


/* ----------------
 *   range_in
 *
 *   This is the routine used to create internal representation of range 
 *   data type in Postgresql. Any new value of user data types in Postgresql
 *   is provided in string and Postgresql server calls this routine to form
 *   the internal representation of this user data type.
 *
 *   It is our responsibility to parse the string (according to the format we 
 *   specify) and form the internal representation of range object.
 *
 *   the supported syntax for range data type is as follows:
 * 
 *   INTEGER    <- [1-9]*[0-9]
 *   RANGE      <- INTEGER .. INTEGER
 *   RANGE      <- INTEGER : INTEGER
 *
 *   in other words, it looks like
 *   INTEGER..INTEGER      -- perl slice syntax
 *   INTEGER:INTEGER       -- python slice syntax
 *
 *
 *   parameters:
 *     str - a C string from user input
 *
 *   return:
 *     construct the internal representation of the data type.
 *
 *   note:
 *     
 *   if the given string does not follow the syntax, error must be returned by
 *   elog( ERROR, "error message" ). This routine will return immediately.
 *
 * ---------------
 */


RANGE *range_in(char *str)
{
  RANGE  *result = NULL;
  char buffer[100]; 
  RANGE_int lower, upper;

  int error;

  error = 0;

  if ( strchr(str,':')!=NULL )
  {
    if ( 2!=sscanf(str, " %lld : %lld %s", &lower, &upper, buffer) )
      error = 1;
  }
  else if ( 2!=sscanf( str, " %lld .. %lld %s", &lower, &upper, buffer ) )  error = 1;

  if ( !error )
  {
    result = palloc(sizeof(RANGE));
    result->lower = lower;
    result->upper = upper;
  } 
  else elog( ERROR, "invalid range format:  [INTEGER..INTEGER] or [INTEGER:INTEGER]");


  return (result);
}


/* ----------------
 *   range_out
 *
 *   This is the routine used to transform  internal representation of range
 *   data type to a string form in Postgresql. Whenever Postgresql needs to
 *   display the range data, it will call this routine to get its string form.
 *   
 *   Even though we support two forms of range specification, the returned 
 *   format will be as follows:
 *
 *   INTEGER    <- [1-9]*[0-9]
 *   RANGE      <- INTEGER .. INTEGER
 *
 *   in other words, it looks like
 *   INTEGER..INTEGER      -- perl slice syntax
 *
 *   parameters:
 *     range - the internal representation of range data
 *
 *   return:
 *     construct a string representation of the range data type.
 *
 * ----------------
 */
char *range_out(RANGE* range)
{
  char buffer[60];
  char *result;

  if (range == NULL)  return (NULL);

  sprintf( buffer, "%lld..%lld", range->lower, range->upper );

  result = (char *) palloc( strlen(buffer)+1 );
  strcpy(result, buffer);  

  return (result);
}

/* ----------------
 *   range_center
 *
 *   This is the routine used to retrieve the center point of the range.
 *   If the central point is not an integer, it will be rounded to the
 *   closet integer.
 *
 *   parameters:
 *     range - the internal representation of range data
 *
 *   return:
 *     return an int64 value.
 *
 * ----------------
 */

RANGE_int* range_center(RANGE *r)
{
  RANGE_int* result;

  if (!r) return (RANGE_int*) NULL;

  result = (RANGE_int*) palloc(sizeof(RANGE_int));
  *result = (r->lower+r->upper) / 2;
  return (result);
}

/* ----------------
 *   range_lower
 *
 *   This is the routine used to retrieve the lower point of the range.
 *
 *   parameters:
 *     range - the internal representation of range data
 *
 *   return:
 *     return the lower point of the range (int64).
 *
 * ----------------
 */
RANGE_int* range_lower(RANGE *r)
{
  RANGE_int *result;
  if (!r) return (RANGE_int*) NULL;

  result = (RANGE_int*) palloc(sizeof(RANGE_int)); 
  *result = r->lower;
  return (result);
}

/* ----------------
 *   range_upper
 *
 *   This is the routine used to retrieve the upper point of the range.
 *
 *   parameters:
 *     range - the internal representation of range data
 *
 *   return:
 *     return the upper point of the range (int64).
 *
 * ----------------
 */
RANGE_int* range_upper(RANGE *r)
{
  RANGE_int* result;

  if (!r) return (RANGE_int*) NULL;

  result = (RANGE_int *) palloc(sizeof(RANGE_int));
  *result = r->upper;
  return (result);
}


/* ----------------
 *   boolean testing functions
 *
 *   The following set of routines are used to test the 
 *   relationship between two range objects, say a and b.
 *   
 *   they will return a boolean value. TRUE means the relationship
 *   exists between these a and b.
 *
 *   they are intutive and the meaning of them can be easily seen
 *   from the implementation.
 *
 *   parameters:
 *     a - the first range object
 *     b - the second range object
 *     
 *   return:
 *     return TRUE if the relationship exists between a and b; otherwise,
 *     return FALSE
 *
 * ----------------
 */

bool range_contains(RANGE * a, RANGE * b)
{
  return ((a->lower <= b->lower) && (a->upper >= b->upper));
}

bool range_inside(RANGE* a, RANGE *b)
{
  return ((a->lower>b->lower) && (a->upper<b->upper));
}

bool range_contained(RANGE * a, RANGE * b)
{
  return (range_contains(b, a));
}

bool range_eq(RANGE * a, RANGE * b)
{
  return range_cmp(a, b) == 0;
}

/*	range_overlaps -- does a overlap b? */
bool range_overlaps(RANGE * a, RANGE * b)
{
  return (
 	   ((a->upper >= b->upper) && (a->lower <= b->upper))
	   ||
	   ((b->upper >= a->upper) && (b->lower <= a->upper))
	 );
}

/* range_overleft -- is the right edge of (a) located to the left of the 
                   right edge of (b)? */

bool range_over_left(RANGE * a, RANGE * b)
{
  return (a->upper <= b->upper && !range_left(a, b) && !range_right(a, b));
}

/*	range_left -- is (a) entirely on the left of (b)?  */
bool range_left(RANGE * a,RANGE * b)
{
  return (a->upper < b->lower);
}

/*	ramge_right -- is (a) entirely on the right of (b)?  */
bool range_right(RANGE * a, RANGE * b)
{
  return (a->lower > b->upper);
}

/*range_overright -- is the left edge of (a) located to the right of the left edge of (b)?  */
bool range_over_right(RANGE* a, RANGE * b)
{
  return (a->lower >= b->lower && !range_left(a, b) && !range_right(a, b));
}

bool range_lt(RANGE * a, RANGE * b)
{
  return range_cmp(a, b) < 0;
}

bool range_le(RANGE * a, RANGE * b)
{
  return range_cmp(a, b) <= 0;
}

bool range_gt(RANGE * a, RANGE * b)
{
  return range_cmp(a, b) > 0;
}

bool range_ge(RANGE * a, RANGE * b)
{
  return range_cmp(a, b) >= 0;
}

bool range_ne(RANGE * a, RANGE * b)
{
  return range_cmp(a, b) != 0;
}

bool range_meets(RANGE *a, RANGE *b)
{
  return (a->upper==b->lower);
}

bool range_met_by(RANGE *a, RANGE *b)
{
  return (b->upper==a->lower);
}

bool range_starts(RANGE *a, RANGE *b)
{
  return ( (a->lower==b->lower) && (a->upper<=b->upper) );
}

bool range_started_by(RANGE *a, RANGE *b)
{
  return ((b->lower==a->lower) && (a->upper>=b->upper) );
}

bool range_finishes(RANGE *a, RANGE *b)
{
  return ( (a->upper==b->upper) && (a->lower>=b->lower) );
}

bool range_finished_by(RANGE *a, RANGE *b)
{
  return ( (a->upper==b->upper) && (b->lower>=a->lower) );
}


bool range_contains_int( RANGE *a, RANGE_int* b)
{
  return ((a->lower <= *b) && (a->upper >= *b));
}

bool range_same_lower(RANGE *a, RANGE *b)
{
  return (a->lower==b->lower);
}

bool range_same_upper(RANGE *a, RANGE *b)
{
  return (a->upper==b->upper);
}


/* ----------------
 *   RANGE generator
 *
 *   The following set of routines are called RANGE generator because
 *   they take necessary parameters (typically two range objects) and
 *   generate a new range object.
 *
 *   they are intuitive and the meanings of them can be easily seen from
 *   the name and the the implementation.
 *
 *   return:
 *     return a new RANGE object
 *
 * ----------------
 */



RANGE *range_union(RANGE * a, RANGE * b)
{
  RANGE *n;

  n = (RANGE *) palloc(sizeof(*n));

  /* take max of upper endpoints */
  if (a->upper > b->upper)
    n->upper = a->upper;
  else
    n->upper = b->upper;

  /* take min of lower endpoints */
  if (a->lower < b->lower)
    n->lower = a->lower;
  else
    n->lower = b->lower;

  return (n);
}


RANGE *range_inter(RANGE * a, RANGE * b)
{
  RANGE *n;

  n = (RANGE *) palloc(sizeof(*n));

  /* take min of upper endpoints */
  if (a->upper < b->upper)
    n->upper = a->upper;
  else
    n->upper = b->upper;

  /* take max of lower endpoints */
  if (a->lower > b->lower)
    n->lower = a->lower;
  else
    n->lower = b->lower;

  return (n);
}

RANGE* range_torange( RANGE_int* a, RANGE_int* b)
{
  RANGE *n = (RANGE*) palloc(sizeof(RANGE));

  n->lower = *a;
  n->upper = *b;

  return n;
}

RANGE* range_duplicate( RANGE *r )
{
  RANGE *n = (RANGE*) palloc(sizeof(RANGE));
  n->lower = r->lower;
  n->upper = r->upper;
  return n;
}

void range_copy( RANGE *dest, RANGE *src )
{
  dest->lower = src->lower;
  dest->upper = src->upper;
}

RANGE* range_maxmin( RANGE *a,RANGE *b)
{
  RANGE *n = (RANGE*) palloc(sizeof(RANGE));
  RANGE_int tmp;

  n->lower = max(a->lower, b->lower);
  n->upper = min(a->upper, b->upper);

  if ( n->lower>n->upper )
  {
    tmp = n->lower;
    n->lower = n->upper;
    n->upper = tmp;
  }

  return n;
}

RANGE* range_minmax( RANGE *a, RANGE *b)
{
  RANGE *n = (RANGE*) palloc(sizeof(RANGE));

  n->lower = min(a->lower, b->lower);
  n->upper = max(a->upper, b->upper);

  return n;
}

RANGE* range_times( RANGE *a, RANGE *b)
{
  RANGE *n = (RANGE*) palloc(sizeof(RANGE));

  n->lower = a->lower * b->lower;
  n->upper = a->upper * b->upper;

  return n;
}

RANGE* range_timesconst( RANGE *a, int32 b)
{
  RANGE *n = (RANGE*) palloc(sizeof(RANGE));

  n->lower = a->lower * b;
  n->upper = a->upper * b;

  return n;
}

RANGE* range_consttimes( int32 b, RANGE *a)
{
  return range_timesconst( a, b);
}


RANGE* range_plusconst( RANGE *a, RANGE_int* b)
{
  RANGE *n = (RANGE*) palloc(sizeof(RANGE));

  n->lower = a->lower + (*b);
  n->upper = a->upper + (*b);

  return n;
}

RANGE* range_constplus(RANGE_int* b, RANGE *a)
{
  return range_plusconst(a,b);
}

RANGE* range_minusconst( RANGE *a, RANGE_int* b)
{
  RANGE *n = (RANGE*) palloc(sizeof(RANGE));

  n->lower = a->lower - (*b);
  n->upper = a->upper - (*b);

  return n;
}

RANGE* range_constminus(RANGE_int* b, RANGE *a)
{
  return range_minusconst(a,b);
}

RANGE* range_plus( RANGE *a, RANGE *b)
{
  RANGE *n = (RANGE*) palloc(sizeof(RANGE));

  n->lower = a->lower + b->lower;
  n->upper = a->upper + b->upper;
  return n;
}

RANGE* range_minus( RANGE *a, RANGE *b)
{
  RANGE *n = (RANGE*) palloc(sizeof(RANGE));

  n->lower = a->lower - b->lower;
  n->upper = a->upper - b->upper;
  return n;
}

RANGE* range_extend( RANGE *a, int32 b)
{
  RANGE *n = (RANGE*) palloc(sizeof(RANGE));

  n->lower = max(a->lower - b,0);
  n->upper = a->upper + b;
  return n;
}

/* Miscellaneous operators */

void rt_range_size(RANGE *a, RANGE_int* size)
{
  if (a == (RANGE *) NULL || a->upper <= a->lower)
    *size = 0;
  else
    *size = abs(a->upper - a->lower)+1;

  return;
}

RANGE_int* range_size(RANGE *a)
{
  RANGE_int* result;

  result = (RANGE_int*) palloc(sizeof(RANGE_int));

  *result = abs(a->upper - a->lower)+1;
  return (result);
}

int32 range_cmp(RANGE * a, RANGE * b)
{
  /* First compare on lower boundary position  */
  if (a->lower < b->lower) 
    return -1;
  if (a->lower > b->lower)
    return 1;

  /* at this point, the lower boundaries are identical */

  /* First compare on upper boundary position */
  if (a->upper < b->upper)
    return -1;
  if (a->upper > b->upper)
    return 1;

  return 0;
}

