/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * range_gist.c
 *
 *    This file contains GiST method implementation for range
 *    data type.
 *
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */

#include "postgres.h"
#include "access/gist.h"
#include "access/rtree.h"
#include "utils/builtins.h"
#include "range.h"
#include "range_gist.h"

#define max(a,b)		((a) >	(b) ? (a) : (b))
#define min(a,b)		((a) <= (b) ? (a) : (b))
#define abs(a)			((a) <	(0) ? (-a) : (a))


/*
   The following routines are for GiST index implementation
   GiST requires us to provide six methods: (according to the GiST paper)

   1. Consistent(E,q): determine if datum in the sub tree of the given entry E can be
      guaranteed un-satisfiable to query q or not.
   2. Union(P): given a set P of entries, form a predicate that holds true for all entries.
   3. Compress(E): return the compressed representation of the given entry E.
   4. Decompress(E): return the decompressed representation of the given entry E.
   5. Penalty(E1,E2):  return a domain-specific penalty for inserting E2 into the subtree of E1.
      GiST inserts the entry with least penalty into a sub tree.
   6. PickSplit(P): split a set P of entries into two sets of entries.

   In Postgresql, query type is specified by StrategyNumber so we are able to support
   as many predicate as we could as long as the index structure supports it efficiently.

*/



/* ----------------
 *   grange_consistent
 *
 *   This is one of the 6 GiST routines. Given a GiST entry and a query (in this case, 
 *   RANGE), this routine should determine whether this particular entry is consistent
 *   with the query with respect to the given strategy. StrategyNumber is specified in
 *   the header file and should be provided when defining the GiST access method
 *   in Postgresql.
 *
 *   Since the determination of consistency differs between leaf nodes and internal 
 *   nodes, so this routine is implemented as a wrapper to two other routines. (for
 *   leaf node, internal node respectively)
 *
 *   parameters:
 *     entry - a given GiST node
 *     query - user-specified query (a range object)
 *     strategy - the strategy number (defined in range.h)
 *
 *   return:
 *     return true if this query is consistent with the given GiST entry; otherwise, 
 *     return false.
 *
 *   note:
 *     If you're very sure that the entries under this given GiST node is not consistent
 *     with the query, you should always return FALSE. Returning TRUE is not harmful if
 *     subnodes are not consistent with the query, but the performance degradation is the
 *     penalty.
 *
 * ----------------
 */
bool grange_consistent(GISTENTRY *entry, RANGE * query, StrategyNumber strategy)
{

   /* if entry is not leaf, use grange_internal_consistent,  
      else use grange_leaf_consistent  */  
  if (GIST_LEAF(entry))
     return (grange_leaf_consistent((RANGE *) DatumGetPointer(entry->key), query, strategy));
  else
     return (grange_internal_consistent((RANGE *) DatumGetPointer(entry->key), query, strategy));
}


/* ----------------
 *   grange_leaf_consistent
 *
 *   This is the routine called by grange_consistent when the given GiST node is a leaf
 *   node. 
 *   for detailed explanation of GiST consistent method, please refer to the comment
 *   for grange_consistent and GiST paper.
 *
 *   parameters:
 *     key - a given GiST leaf node
 *     query - user-specified query (a range object)
 *     strategy - the strategy number (defined in range.h)
 *
 *   return:
 *     return true if this query is consistent with the given GiST entry; otherwise,
 *     return false.
 *
 * ----------------
 */
bool grange_leaf_consistent(RANGE* key, RANGE* query, StrategyNumber strategy)
{
  bool retval;


  switch (strategy)
  {
     case GiSTRangeSameLowerStrategyNumber:
             retval = (bool) (key->lower==query->lower);
             break;
     case GiSTRangeSameUpperStrategyNumber:
             retval = (bool) (key->upper==query->upper);
             break;
     case RTLeftStrategyNumber:
             retval = (bool) range_left(key, query);
             break;
     case RTOverLeftStrategyNumber:
             retval = (bool) range_over_left(key, query);
             break;
     case RTOverlapStrategyNumber:
             retval = (bool) range_overlaps(key, query);
             break;
     case RTOverRightStrategyNumber:
             retval = (bool) range_over_right(key, query);
             break;
     case RTRightStrategyNumber:
             retval = (bool) range_right(key, query);
             break;
     case RTSameStrategyNumber:
             retval = (bool) range_eq(key, query);
             break;
     case RTContainsStrategyNumber:
             retval = (bool) range_contains(key, query);
             break;
     case RTContainedByStrategyNumber:
             retval = (bool) range_contained(key, query);
             break;
     default:
             retval = FALSE;
  }
  return (retval);
}

/* ----------------
 *   grange_internal_consistent
 *
 *   This is the routine called by grange_consistent when the given GiST node is a leaf
 *   node.
 *   for detailed explanation of GiST consistent method, please refer to the comment
 *   for grange_consistent and GiST paper.
 *
 *   parameters:
 *     key - a given GiST leaf node
 *     query - user-specified query (a range object)
 *     strategy - the strategy number (defined in range.h)
 *
 *   return:
 *     return true if this query is consistent with the given GiST entry; otherwise,
 *     return false.
 *
 *   note:
 *     If you're very sure that the entries under this given GiST node is not consistent
 *     with the query, you should always return FALSE. Returning TRUE is not harmful if
 *     subnodes are not consistent with the query, but the performance degradation is the
 *     penalty.
 *
 * ----------------
 */

bool grange_internal_consistent(RANGE* key,RANGE * query, StrategyNumber strategy)
{
  bool retval;


  switch (strategy)
  {
    case RTLeftStrategyNumber:
    case RTOverLeftStrategyNumber:
    case GiSTRangeSameUpperStrategyNumber:
            retval = (bool) range_over_left(key, query);
            break;
    case RTOverlapStrategyNumber:
            retval = (bool) range_overlaps(key, query);
            break;
    case RTOverRightStrategyNumber:
    case RTRightStrategyNumber:
    case GiSTRangeSameLowerStrategyNumber:
            retval = (bool) range_right(key, query);
            break;
    case RTSameStrategyNumber:
    case RTContainsStrategyNumber:
            retval = (bool) range_contains(key, query);
            break;
    case RTContainedByStrategyNumber:
            retval = (bool) range_overlaps(key, query);
            break;
    default:
            retval = FALSE;
  }
  return (retval);
}



/* ----------------
 *   grange_union
 *
 *   This is one of the six required GiST methods. 
 *   It will be invoked with a number of GiST entries and should return a GiST
 *   entry which holds true for any of these entries.
 *
 *   In our implementation, we returns the minimal bounding range that encloses
 *   all the entries in the entryvec.
 *
 *   parameters:
 *     entryvec - GiST entries
 *     sizep - the size of a single range object.
 *
 *   return:
 *     return a range data object which holds true for all the given entries.
 *
 * ----------------
 */

Datum grange_union(PG_FUNCTION_ARGS)
{
  GistEntryVector *entryvec = (GistEntryVector *) PG_GETARG_POINTER(0);
  int numranges, i;
  RANGE *out = (RANGE *) NULL;
  RANGE *tmp;
  int length;

  numranges = entryvec->n;

  tmp = (RANGE *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[0].key));
  length = sizeof(RANGE);
  *(int *) PG_GETARG_POINTER(1) = sizeof(RANGE);

  for (i = 1; i < numranges; i++)
  {
    out = grange_binary_union(tmp, (RANGE*) DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[i].key), &length);
    if (i > 1)
      pfree(tmp);
    tmp = out;
  }

  PG_RETURN_POINTER(out);
}


/* ----------------
 *   grange_compress
 *
 *   This is one of the six required GiST methods.
 *   Since we do not compress any GiST nodes, so it does nothing.
 *
 *   parameters:
 *     entry - GiST entries
 *
 *   return:
 *     return the compressed form of the given GiST entry (which is itself)
 *
 * ----------------
 */

GISTENTRY *grange_compress(GISTENTRY *entry)
{
  return (entry);
}

/* ----------------
 *   grange_decompress
 *
 *   This is one of the six required GiST methods.
 *   Since we do not compress any GiST nodes, so it does nothing.
 *
 *   parameters:
 *     entry - GiST entries
 *
 *   return:
 *     return the decompressed form of the given GiST entry (which is itself)
 *
 * ----------------
 */

GISTENTRY *grange_decompress(GISTENTRY *entry)
{
  return (entry);
}

/* ----------------
 *   grange_penalty
 *
 *   This is one of the six required GiST methods.
 *   GiST kernel requires us to give it domain penalty for insertion an
 *   entry into a subtree. This routine tells GiST what penalty value should
 *   it receive if GiST change the original entry into a new entry.
 *
 *   In this implemenation, this penalty is determined by the change in area
 *   as described in the R-tree paper.
 *
 *   parameters:
 *     origentry - original GiST entry
 *     newentry  - new GiST entry
 *
 *   return:
 *     return the penalty value
 *
 * ----------------
 */

RANGE_int* grange_penalty(GISTENTRY *origentry, GISTENTRY *newentry, RANGE_int* result)
{
  RANGE *ud;
  RANGE_int tmp1, tmp2;

  ud = range_union((RANGE *) DatumGetPointer(origentry->key), (RANGE *) DatumGetPointer(newentry->key));
  rt_range_size(ud, &tmp1);
  rt_range_size((RANGE*) DatumGetPointer(origentry->key), &tmp2);  

  *result = tmp1 - tmp2;
  pfree(ud);

  return (result);
}

/* ----------------
 *   grange_picksplit
 *
 *   This is one of the six required GiST methods.
 *   When a node in the GiST search tree is full, we need to split this node into
 *   two. GiST requires us to dictate the split process.
 *   
 *   In this implementation, we use Guttman's poly time split algorithm.
 *
 *   parameters:
 *     entryvec - entries in the old GiST node
 *     v        - the split vector. It has a rather complex structure, but the part
 *                that we need to know is
 *
 *     v->spl_ldatum, v->spl_rdatum  : pointers to left and right of the split nodes.
 *     v->spl_nleft,  v->spl_nright  : number of entries are there in each of the left, right node
 *     v->spl_left,   v->spl_right   : the entry numbers which belong to the left, right node respectively
 *
 *   return:
 *     the GiST split vector structure.
 *
 * ----------------
 */


//GIST_SPLITVEC *grange_picksplit(bytea *entryvec, GIST_SPLITVEC *v)
Datum grange_picksplit(PG_FUNCTION_ARGS)
{
  GistEntryVector *entryvec = (GistEntryVector *) PG_GETARG_POINTER(0);
  GIST_SPLITVEC *v = (GIST_SPLITVEC *) PG_GETARG_POINTER(1);
  OffsetNumber i, j;
  RANGE *datum_alpha, *datum_beta;
  RANGE *datum_l, *datum_r;
  RANGE *union_d, *union_dl, *union_dr;
  RANGE *inter_d;
  bool firsttime;
  RANGE_int size_alpha, size_beta, size_union, size_inter;
  RANGE_int size_waste, waste;
  RANGE_int size_l, size_r;

  int nbytes;
  OffsetNumber seed_1 = 0, seed_2 = 0;
  OffsetNumber *left, *right;
  OffsetNumber maxoff;

  maxoff = entryvec->n - 2;
  nbytes = (maxoff + 2) * sizeof(OffsetNumber);
  v->spl_left = (OffsetNumber *) palloc(nbytes);
  v->spl_right = (OffsetNumber *) palloc(nbytes);

  firsttime = true;
  waste = 0;

  for (i = FirstOffsetNumber; i < maxoff; i = OffsetNumberNext(i))
  {
    datum_alpha = (RANGE *) DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[i].key);
    for (j = OffsetNumberNext(i); j <= maxoff; j = OffsetNumberNext(j))
    {
      datum_beta = (RANGE *) DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[j].key);

      /* compute the wasted space by unioning these guys */
      /* size_waste = size_union - size_inter; */
      union_d = range_union(datum_alpha, datum_beta);
      rt_range_size(union_d, &size_union);
      inter_d = range_inter(datum_alpha, datum_beta);
      rt_range_size(inter_d, &size_inter);
      size_waste = size_union - size_inter;

      pfree(union_d);

      if (inter_d != (RANGE *) NULL)	pfree(inter_d);

      /* are these a more promising split that what we've already seen?  */

      if (size_waste > waste || firsttime)
      {
 	waste = size_waste;
	seed_1 = i;
	seed_2 = j;
 	firsttime = false;
      }
    }
  }

  left = v->spl_left;
  v->spl_nleft = 0;
  right = v->spl_right;
  v->spl_nright = 0;

  datum_alpha = (RANGE *) DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[seed_1].key);
  datum_l = range_union(datum_alpha, datum_alpha);
  rt_range_size(datum_l, &size_l);
  datum_beta = (RANGE *) DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[seed_2].key);
  datum_r = range_union(datum_beta, datum_beta);
  rt_range_size(datum_r, &size_r);

  /*
   * Now split up the regions between the two seeds.	An important
   * property of this split algorithm is that the split vector v has the
   * indices of items to be split in order in its left and right
   * vectors.  We exploit this property by doing a merge in the code
   * that actually splits the page.
   *
   * For efficiency, we also place the new index tuple in this loop. This
   * is handled at the very end, when we have placed all the existing
   * tuples and i == maxoff + 1.
   */

  maxoff = OffsetNumberNext(maxoff);
  for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i))
  {
    /*
     * If we've already decided where to place this item, just put it
     * on the right list.  Otherwise, we need to figure out which page
     * needs the least enlargement in order to store the item.
     */

    if (i == seed_1)
    {
      *left++ = i;
      v->spl_nleft++;
      continue;
    }
    else if (i == seed_2)
    {
      *right++ = i;
      v->spl_nright++;
      continue;
    }

    /* okay, which page needs least enlargement? */
    datum_alpha = (RANGE *) DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[i].key);
    union_dl = range_union(datum_l, datum_alpha);
    union_dr = range_union(datum_r, datum_alpha);
    rt_range_size(union_dl, &size_alpha);
    rt_range_size(union_dr, &size_beta);

    /* pick which page to add it to */
    if (size_alpha - size_l < size_beta - size_r)
    {
      pfree(datum_l);
      pfree(union_dr);
      datum_l = union_dl;
      size_l = size_alpha;
      *left++ = i;
      v->spl_nleft++;
    }
    else
    {
      pfree(datum_r);
      pfree(union_dl);
      datum_r = union_dr;
      size_r = size_alpha;
      *right++ = i;
      v->spl_nright++;
    }
  }
  *left = *right = FirstOffsetNumber; /* sentinel value, see dosplit() */
  v->spl_ldatum = PointerGetDatum(datum_l);
  v->spl_rdatum = PointerGetDatum(datum_r);

  PG_RETURN_POINTER( v);
}

/* ----------------
 *   grang_same
 *
 *   This is one of the six required GiST methods.
 *   GiST kernel requires us to tell it whether two objects are the same.
 *   (since GiST does not know the internal representation of the user datatype)
 *
 *   In this implementation, if the interval defined by these two objects are
 *   equal, then they are the same
 *
 *   parameters:
 *     b1 - range datatype
 *     b2 - range datatype
 *     result - the result (pass by pointer )
 *
 *   return:
 *     return the boolean value (pass by pointer)
 *
 * ----------------
 */

bool *grange_same(RANGE * b1, RANGE * b2, bool *result)
{
  if (range_eq(b1, b2))
    *result = TRUE;
  else
    *result = FALSE;

  return (result);
}


RANGE *grange_binary_union(RANGE * r1, RANGE * r2, int *sizep)
{
  RANGE *retval;

  retval = range_union(r1, r2);
  *sizep = sizeof(RANGE);

  return (retval);
}

