/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * seq.c
 *
 *    This file contains functions for manipulating genomic sequences
 *    such as translate, transcribe, revcom. Additional functions such
 *    as substring, length are provided in Postgresql.
 *
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */


#include "postgres.h"
#include "funcapi.h"
#include "miscadmin.h"
#include "utils/builtins.h"
#include "utils/guc.h"
#include "utils/lsyscache.h"
#include <string.h>
#include <ctype.h>
#include "seq.h"

/* some convenient macro to retrieve pointers for text */
#define GET_TEXT(cstrp) DatumGetTextP(DirectFunctionCall1(textin, CStringGetDatum(cstrp)))
#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))

/* ----------------
 *   genetic code table
 *
 * this genetic code table is indexed by the order of the codon, starting 
 * from AAA to TTT. the order is defined by the translate() function.
 * Basically, a codon can be represented as an integer from 0 - 63. A is encoded 
 * as 0, C as 1, G as 2, T as 3, then suppose we have a code X1X2X3 (which is mapped 
 * into integers Y1Y2Y3), its corresponding index will be
 *
 * (Y1 << 4) + (Y2 << 2) + Y3
 *
 * ----------------
 */


char genetic_code[64] = {
     'K', 'N', 'T', 'N',
     'T', 'T', 'T', 'N',
     'R', 'S', 'R', 'S',
     'I', 'I', 'M', 'I',
     'Q', 'H', 'Q', 'H',
     'P', 'P', 'P', 'P', 
     'R', 'R', 'R', 'R',
     'L', 'L', 'L', 'L',
     'E', 'D', 'E', 'D',
     'A', 'A', 'A', 'A',
     'G', 'G', 'G', 'G',
     'V', 'V', 'V', 'V',
     '*', 'Y', '*', 'Y',
     'S', 'S', 'S', 'S',
     '*', 'C', 'W', 'C',
     'L', 'F', 'L', 'F'
};


/* ----------------
 *   revcom
 *
 *   This is the routine used to get the reverse complementary genomic sequence
 *   of the given one. I preserve the case of the given sequence in case that
 *   users make a distinction for the cases. 
 *   In all the genomic sequence functions, I don't make any assumption about the
 *   cases of the letters.
 *
 *   parameters:
 *     seq - genomic sequence
 *
 *   return:
 *     returns the reverse complementary genomic sequence.
 *
 *   note:
 *     this routine should ONLY be called with DNA sequences. If called
 *     with other types of sequences (such as mRNA and protein), only 
 *     A,T,C,G would be changed and the rest of the character set will remain
 *     unchanged.
 *
 * ----------------
 */

Datum revcom(PG_FUNCTION_ARGS)
{
  text *t = PG_GETARG_TEXT_P(0);
  text *new_t = (text *) palloc( VARSIZE(t));
  char *p, *q;
  int length;

  /* since the reverse complementary sequence has the same length, so it is
     safe to allocate memory with the same size as the given sequence */
  VARATT_SIZEP(new_t) = VARSIZE(t);
  p = GET_STR(t);
  q = (char*) VARDATA(new_t);
  length = VARSIZE(t)-VARHDRSZ;

  while (length-->0)
  {
     switch (*p )
     {
        case 'A':    *q = 'T'; break;
	case 'T':    *q = 'A'; break;
	case 'C':    *q = 'G'; break;
	case 'G':    *q = 'C'; break;
	case 'a':    *q = 't'; break;
	case 't':    *q = 'a'; break;
	case 'c':    *q = 'g'; break;
	case 'g':    *q = 'c'; break;	
	default:     *q = *p;  
     }
     p++;
     q++;
  }

  PG_RETURN_TEXT_P(new_t);
}

/* ----------------
 *   transcribe
 *
 *   This is the routine used to get the transcribed version of a DNA sequence.
 *   I preserve the case of the given sequence in case that users make a distinction 
 *   for the cases.
 *   In all the genomic sequence functions, I don't make any assumption about the
 *   cases of the letters.
 *
 *   parameters:
 *     seq - genomic sequence
 *
 *   return:
 *     returns the transcribed version of the given genomic sequence.
 *
 *   note:
 *     this routine should ONLY be called with DNA sequences. If called
 *     with other types of sequences (such as mRNA and protein), only
 *     T would be changed and the rest of the character set will remain
 *     unchanged.
 *
 * ----------------
 */

Datum transcription(PG_FUNCTION_ARGS)
{
  text *t = PG_GETARG_TEXT_P(0);
  text *new_t = (text *) palloc( VARSIZE(t));
  char *p, *q;
  int length;

  /* since the reverse complementary sequence has the same length, so it is
     safe to allocate memory with the same size as the given sequence */
  VARATT_SIZEP(new_t) = VARSIZE(t);
  p = GET_STR(t);
  q = (char*) VARDATA(new_t);
  length = VARSIZE(t)-VARHDRSZ;

  /* we only need to change T(t) to U(u) */
  while (length-->0)
  {
     switch (*p )
     {
        case 'T':    *q = 'U'; break;
        case 't':    *q = 'u'; break;
        default:     *q = *p;
     }
     p++;
     q++;
  }

  PG_RETURN_TEXT_P(new_t);
}

/* ----------------
 *   translation
 *
 *   This is the routine used to map codon into amino acid letter.
 *   All amino acids will be in UPPER case and stop codon is represented as '*'

 *   parameters:
 *     p - genomic sequence pointer (codon is the next three characters)
 *
 *   return:
 *     returns the translated letter for the current codon.
 *
 *   note:
 *     1. this routine should ONLY be called with DNA sequences. If called
 *     with other types of sequences (such as mRNA and protein), the result
 *     will not be correct at all.
 *     2. this routine DOES NOT take care of alternative splicing mechanism.
 *
 * ----------------
 */

char translation_table( char *p )
{
   char ch;
   int i, code[3];

   for(i=0;i<3;i++,p++)
   {
      ch = tolower(*p);
      switch (ch)
      {
         case 'a':  code[i] = 0;   break;
         case 'c':  code[i] = 1;   break;
         case 'g':  code[i] = 2;   break;
         case 't':  code[i] = 3;   break;
         default:   code[i] = 0;
      }
   }

   i = (code[0]<< 4)+ (code[1]<<2) + code[2];

   return genetic_code[i]; 
}

/* ----------------
 *   translate
 *
 *   This is the routine used to get the translated version of a DNA sequence.
 *   I preserve the case of the given sequence in case that users make a distinction
 *   for the cases. 
 *   In all the genomic sequence functions, I don't make any assumption about the
 *   cases of the letters.
 *
 *   parameters:
 *     seq - genomic sequence
 *     offset - (optional), translation starting offset (1-based).
 *
 *   return:
 *     returns the translated version of the given genomic sequence.
 *
 *   note:
 *     1. this routine should ONLY be called with DNA sequences. If called
 *     with other types of sequences (such as mRNA and protein), the result
 *     will not be correct at all.
 *     2. this routine DOES NOT take care of alternative splicing mechanism.
 *
 * ----------------
 */

Datum translation(PG_FUNCTION_ARGS)
{
  text *new_t, *t = PG_GETARG_TEXT_P(0);
  int32 i, length, new_length, offset;
  char *p, *q;

  /* check if offset is provided, if not, 1 is assumed */
  if (PG_ARGISNULL(1)==TRUE) offset = 1;
  else offset = PG_GETARG_INT32(1);

  if (offset<1) offset = 1;


  length = VARSIZE(t)-VARHDRSZ;

  /* calculate the length of the translated string and allocate memory for it*/
  new_length = length - offset + 1;
  new_length = new_length - new_length % 3;

  new_t = (text*) palloc(  VARHDRSZ + new_length+1 );
  VARATT_SIZEP(new_t) = VARHDRSZ + new_length + 1;

  p = GET_STR(t);
  q = (char *) VARDATA(new_t);

  /* the actually translation work */
  for(i=offset-1;i<length-2;i+=3, q++)
     *q = translation_table( p+i );

  *q = 0;

  PG_RETURN_TEXT_P(new_t);
}

