/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * util.c
 *
 *    This file contains common utility functions 
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */


#include "postgres.h"
#include "funcapi.h"
#include "miscadmin.h"
#include "utils/builtins.h"
#include "utils/guc.h"
#include "utils/lsyscache.h"
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "util.h"

/* some convenient macro to retrieve pointers for text */
#define GET_TEXT(cstrp) DatumGetTextP(DirectFunctionCall1(textin, CStringGetDatum(cstrp)))
#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))

int cstring_cmp( char* s1, char* s2 )
{
   return strcmp(s1,s2);
}

bool cstring_eq( char *s1, char *s2 )
{
   return ( cstring_cmp(s1,s2)==0 );
}

bool cstring_ne( char* s1, char* s2 )
{
   return !cstring_eq(s1,s2);
}

/* ----------------
 *   hash
 *
 *   This is a string-based hash function used to hash an identifier
 *   to a uint32 integer. The actual representation of Location datatype
 *   is a uint32 (representing the identifier) and its associated interval
 *   (range datatype).
 *
 *   The purpose of using a uint32 is to accelerate the comparisons between
 *   location datatypes. So if two location values with different hash values,
 *   they must be different. HOWEVER, two locations with the same hash value
 *   are not necessary the same. In this case, we need to compare their interval
 *   values and the actual identifiers.
 *
 *   parameters:
 *     str - identifier string
 *
 *   return:
 *     a uint32 integer. (the hash value of the input string)
 */
uint32 hash(unsigned char *str)
{
  uint32 g, h = 0;

  while ( *str )
  {
    h = (h<<4) + *str++;
    g = h & 0XF0000000L;
    if (g) h ^= g >> 24;
    h &= ~g;
  }

  return h;
}

/* ----------------
 *   cstring2text
 *
 *   This routine is used internally to convert cstring representation
 *   to text type in PostgreSQL.
 *
 *   parameters:
 *     str - cstring
 *
 *   return:
 *     text representation
 *
 * ----------------
 */
text* cstring2text( char *str )
{
   text *new_t;
   int length;

   if ( str==NULL ) return NULL;

   length = strlen(str);
   new_t = (text *) palloc(VARHDRSZ+length+1);
   VARATT_SIZEP(new_t) = VARHDRSZ+length;
   memcpy((void*) VARDATA(new_t), str, length+1);

   return new_t;
}


int CheckFileExistence( char *file )
{
   char buffer[1024];
   char *p;

   if ( access(file,F_OK|R_OK)==0 )
      return 1;

   p = getenv("BLASTDB");
   if ( p )
   {
      sprintf(buffer,"%s%s",p,file);
      return (( access(buffer,F_OK|R_OK)==0 ) ? 1: 0);
   }

   return 0;
}

char *__exec_program( const char *cmd )
{
   FILE *fp;
   char buffer[1024];
   char *command, *ret;
   int size, length, bytes;

   length = strlen(cmd);
   command = (char *) palloc(length+5);
   sprintf( command, "%s 2>&1", cmd );

   ret = NULL;
   fp = popen( command, "r" );
   pfree(command);

   if ( fp )
   {
      ret = (char *) palloc(1024);
      size = 1024;
      length = 0;
      while ( (bytes=fread(buffer,sizeof(char),1024,fp))>0 )
      {
         if ( length+bytes+1>size )
         {
            size = 2*size;
            ret = (char *) repalloc(ret, size);
         }
         memcpy( ret+length, buffer, bytes );
         length += bytes;
         ret[length] = '\0';
      }

      if ( strstr( ret, "command not found") )
      {
         pfree(ret);
         ret = NULL;
      }

      pclose(fp);
   }

   return ret;
}

