/*
**    BLASTgres -- Biosequence data management in PostgreSQL
**    http://www.blastgres.org/
**
**    Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**    (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**    UCLA Center for Computational Biology
**    635 Charles Young Drive South, Suite 225
**    Los Angeles, CA 90095-7332  USA
**
**    UCLA Computer Science Dept.
**    3532 Boelter Hall
**    Los Angeles, CA 90095-1596  USA
**
**    This program is free software; you can redistribute it and/or
**    modify it under the terms of the GNU General Public License
**    as published by the Free Software Foundation; either version 2 of
**    the License, or (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**    See the GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public
**    License along with this program; if not, write to the
**    Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**    Boston, MA  02110-1301  USA
*/

/*-------------------------------------------------------------------------
 *
 * blast.c
 *
 *    This file contains blast-related routines called by Postgresql. Most
 * of the BLAST query handlings are in blastsvc.c 
 *
 * This is part of the BLASTgres project.
 *
 *-------------------------------------------------------------------------
 */

#include <string.h>
#include <stdio.h>

#include "postgres.h"
#include "fmgr.h"
#include "funcapi.h"
#include "executor/spi.h"
#include "lib/stringinfo.h"
#include "miscadmin.h"
#include "utils/builtins.h"
#include "utils/guc.h"
#include "utils/lsyscache.h"
#include "global.h"
#include "blast.h"
#include "blastsvc.h"

/* some convenient macro to retrieve pointers for text */
#define GET_TEXT(cstrp) DatumGetTextP(DirectFunctionCall1(textin, CStringGetDatum(cstrp)))
#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))

/* ----------------
 *   isSequenceFile
 *
 *   This is a routine used by BLAST-related functions to detect if user
 *   specify a fasta file instead of sequence letters.
 *
 *   parameters:
 *     str - user specified file name or sequence letters
 *
 *   return:
 *     1 if the argument is actually a sequence file, 0 otherwise
 *
 * ----------------
 */

int isSequenceFile( char *str )
{
   char *file;
   int ret, i, length;

   /* just convert this to lower case and facilitate the comparison */
   length = strlen(str);
   file = (char *) palloc(length+1);

   for(i=0;i<length;i++)
     file[i] = tolower(str[i]);
   file[length] = '\0';

   ret = 0;
   if ( (strcmp(file+length-4,".fas")==0) || (strcmp(file+length-6,".fasta")==0) )  ret = 1;
   pfree(file);

   return ret;
}


/*
char* read_seq_from_file( char *filename )
{
   FILE *fp;
   long length;
   char *seq, *p, buffer[200];

   fp = fopen(filename, "rt");
   if ( fp==NULL )   return NULL;

   fseek( fp, 0, SEEK_END );
   length = ftell(fp);
   fseek( fp, 0, SEEK_SET );

   seq = (char*) palloc( length );
   fgets(buffer, 199, fp);
   *seq = 0;   

   while ( fgets(buffer, 199, fp)!=NULL )
   {
      p = strrchr( buffer, '\n');
      if ( p!=NULL)  *p = '\0';

      strcat( seq, buffer );          
   }

   fclose(fp);
   return seq;
}
*/

/* ----------------
 *   HitTableFormTuple
 *
 *   This is the routine used by BLAST-related table function routines
 *   to form a tuple based on the hit table records.
 *   A lot of Postgresql-related types and functions are used in
 *   order to maintain memory context and manipulate tuples, please
 *   refer to Postgresql programmer manual for their definitions.
 *
 *   parameters:
 *     fctx - this is our blast_fctx data which contains the hit table and
 *            a character pointer pointing at the current record we are
 *            processing.
 *     slot - Postgresql TableSlot object.
 *     attinmeta - Postgresql AttInMetadata (metadata for tuple attributes)
 *
 *   return:
 *     construct a tuple based on the hit table records.
 *
 *   note:
 *     this routine relies on the definition of "hit" type in Postgresql and
 *     assume a particular format from BLAST hit table result. For more, 
 *     see the comments in local_blast_hit() or local_blast_hit_v();
 *
 * ----------------
 */

Datum HitTableFormTuple( HitTableEntryPointer entry, TupleTableSlot *slot, AttInMetadata *attinmeta )
{
   Datum result;
   HeapTuple tuple;

   int length1, length2, i;    
   char qbuffer[512], sbuffer[512];
   char **values;

   /* create a tuple from this result */
   if ( HitTableEntryQueryStrand(entry)=='-' )
      sprintf( qbuffer, "%s[%d..%d]~", 
               HitTableEntryQueryId(entry), HitTableEntryQueryLower(entry), HitTableEntryQueryUpper(entry) );
   else
      sprintf( qbuffer, "%s[%d..%d]", 
               HitTableEntryQueryId(entry), HitTableEntryQueryLower(entry), HitTableEntryQueryUpper(entry) );

   if ( HitTableEntrySbjctStrand(entry)=='-' )
      sprintf( sbuffer, "%s[%d..%d]~",
               HitTableEntrySbjctId(entry), HitTableEntrySbjctLower(entry), HitTableEntrySbjctUpper(entry) );
   else
      sprintf( sbuffer, "%s[%d..%d]",
               HitTableEntrySbjctId(entry), HitTableEntrySbjctLower(entry), HitTableEntrySbjctUpper(entry) );

   length1 = strlen( qbuffer )+1;
   length2 = strlen( sbuffer )+1;

   /* allocate enough memory for the tuple structure and fill in the data */
   values = (char**) palloc( 8*sizeof(char*));
   values[0] = (char *) palloc( length1 * sizeof(char) );
   values[1] = (char *) palloc( length2 * sizeof(char) );
   values[2] = (char *) palloc( 10 * sizeof(char) );
   values[3] = (char *) palloc( 10 * sizeof(char) );
   values[4] = (char *) palloc( 10 * sizeof(char) );
   values[5] = (char *) palloc( 10 * sizeof(char) );
   values[6] = (char *) palloc( 10 * sizeof(char) );
   values[7] = (char *) palloc( 10 * sizeof(char) );

   snprintf( values[0], length1, "%s", qbuffer );
   snprintf( values[1], length2, "%s", sbuffer );
   snprintf( values[2], 10, "%f", HitTableEntryIdentity(entry) );
   snprintf( values[3], 10, "%d", HitTableEntryLength(entry) );
   snprintf( values[4], 10, "%d", HitTableEntryMismatches(entry) );
   snprintf( values[5], 10, "%d", HitTableEntryGap(entry) );
   snprintf( values[6], 10, "%f", HitTableEntryExpect(entry) );
   snprintf( values[7], 10, "%f", HitTableEntryScore(entry) );

   /* build a tuple */
   tuple = BuildTupleFromCStrings(attinmeta, values);

   /* make the tuple into a datum */
   result = TupleGetDatum(slot,tuple);

   /* memory deallocation for tuple structure */
   for(i=0;i<8;i++)  pfree(values[i]);
   pfree(values);

   return (result);
}

/* ----------------
 *   MegaBlastFormTuple
 *
 *   This is the routine used by BLAST-related table function routines
 *   to form a tuple based on the megablast results.
 *   A lot of Postgresql-related types and functions are used in
 *   order to maintain memory context and manipulate tuples, please
 *   refer to Postgresql programmer manual for their definitions.
 *
 *   parameters:
 *     fctx - this is our blast_fctx data which contains the megablast results and
 *            a character pointer pointing at the current record we are
 *            processing.
 *     slot - Postgresql TableSlot object.
 *     attinmeta - Postgresql AttInMetadata (metadata for tuple attributes)
 *
 *   return:
 *     construct a tuple based on the megablast records.
 *
 *   note:
 *     this routine relies on the definition of "megablast" type in Postgresql and
 *     assume a particular format from BLAST hit table result. For more,
 *     see the comments in local_megablast_hit() or local_megablast_hit_v();
 *
 * ----------------
 */

Datum MegaBlastFormTuple( MegaBlastEntryPointer entry, TupleTableSlot *slot, AttInMetadata *attinmeta )
{
   Datum result;
   HeapTuple tuple;

   int length1, length2, i;
   char qbuffer[512], sbuffer[512];
   char **values;

   /* create a tuple from this result */
   if ( MegaBlastEntryQueryStrand(entry)=='-' )
      sprintf( qbuffer, "%s[%d..%d]~",
               MegaBlastEntryQueryId(entry), MegaBlastEntryQueryLower(entry), MegaBlastEntryQueryUpper(entry) );
   else
      sprintf( qbuffer, "%s[%d..%d]",
               MegaBlastEntryQueryId(entry), MegaBlastEntryQueryLower(entry), MegaBlastEntryQueryUpper(entry) );

   if ( MegaBlastEntrySbjctStrand(entry)=='-' )
      sprintf( sbuffer, "%s[%d..%d]~",
               MegaBlastEntrySbjctId(entry), MegaBlastEntrySbjctLower(entry), MegaBlastEntrySbjctUpper(entry) );
   else
      sprintf( sbuffer, "%s[%d..%d]",
               MegaBlastEntrySbjctId(entry), MegaBlastEntrySbjctLower(entry), MegaBlastEntrySbjctUpper(entry) );

   length1 = strlen( qbuffer )+1;
   length2 = strlen( sbuffer )+1;


   /* allocate enough memory for the tuple structure and fill in the data */
   values = (char**) palloc( 3*sizeof(char*));
   values[0] = (char *) palloc( length1 * sizeof(char) );
   values[1] = (char *) palloc( length2 * sizeof(char) );
   values[2] = (char *) palloc( 10 * sizeof(char) );

   snprintf( values[0], length1, "%s", qbuffer );
   snprintf( values[1], length2, "%s", sbuffer );
   snprintf( values[2], 10, "%f", MegaBlastEntryScore(entry) );

   /* build a tuple */
   tuple = BuildTupleFromCStrings(attinmeta, values);

   /* make the tuple into a datum */
   result = TupleGetDatum(slot,tuple);
   /* memory deallocation for tuple structure */
   for(i=0;i<3;i++)  pfree(values[i]);
   pfree(values);

   return (result);
}


/* ----------------
 *   HspFormTuple
 *
 *   This is the routine used by BLAST-related table function routines
 *   to form a tuple based on the hsp (high scoring pairs) records.
 *   A lot of Postgresql-related types and functions are used in
 *   order to maintain memory context and manipulate tuples, please
 *   refer to Postgresql programmer manual for their definitions.
 *
 *   parameters:
 *     fctx - this is our blast_fctx data which contains the hit table and
 *            a character pointer pointing at the current record we are
 *            processing.
 *     slot - Postgresql TableSlot object.
 *     attinmeta - Postgresql AttInMetadata (metadata for tuple attributes)
 *
 *   return:
 *     construct a tuple based on the hit table records.
 *
 *   note:
 *     this routine relies on the definition of "hit" type in Postgresql and
 *     assume a particular format from BLAST hit table result. For more,
 *     see the comments in local_blast_hit() or local_blast_hit_v();
 *
 * ----------------
 */


Datum HspFormTuple( HspEntryPointer entry, TupleTableSlot *slot, AttInMetadata *attinmeta )
{
   Datum result;
   HeapTuple tuple;  
   int length1, length2, i;
   char qbuffer[128], sbuffer[128], **values;


   /* form the text representation of query location and subject location */
   if ( HspEntryQueryStrand(entry)=='-' )
      sprintf( qbuffer, "%s[%d..%d]~",  HspEntryQueryId(entry), 
                                        HspEntryQueryLower(entry), HspEntryQueryUpper(entry) );
   else
      sprintf( qbuffer, "%s[%d..%d]",   HspEntryQueryId(entry),
                                        HspEntryQueryLower(entry), HspEntryQueryUpper(entry) ); 

   if ( HspEntrySbjctStrand(entry)=='-' )
      sprintf( sbuffer, "%s[%d..%d]~",  HspEntrySbjctId(entry), 
                                        HspEntrySbjctLower(entry), HspEntrySbjctUpper(entry) );
   else
      sprintf( sbuffer, "%s[%d..%d]",   HspEntrySbjctId(entry),
                                        HspEntrySbjctLower(entry), HspEntrySbjctUpper(entry) );

   length1 = strlen( qbuffer )+1;
   length2 = strlen( sbuffer )+1;

   /* allocate enough memory for the tuple structure and fill in the data */
   values = (char**) palloc( 5*sizeof(char*));
   values[0] = (char *) palloc( length1 * sizeof(char) );
   values[1] = (char *) palloc( length2 * sizeof(char) );
   values[2] = (char *) palloc( 10 * sizeof(char) );
   values[3] = (char *) palloc( 10 * sizeof(char) );
   values[4] = (char *) palloc( 10 * sizeof(char) );

   snprintf( values[0], length1, "%s", qbuffer );
   snprintf( values[1], length2, "%s", sbuffer );
   snprintf( values[2], 10, "%f", HspEntryIdentity(entry) );
   snprintf( values[3], 10, "%f", HspEntryExpect(entry) );
   snprintf( values[4], 10, "%f", HspEntryScore(entry) );


   /* build a tuple */
   tuple = BuildTupleFromCStrings(attinmeta, values);
   /* make the tuple into a datum */
   result = TupleGetDatum(slot,tuple);

   /* memory deallocation for tuple structure */
   for(i=0;i<5;i++)  
      pfree(values[i]);
   pfree(values);

   return (result);
}


/* ----------------
 *   blast_sequence
 *
 *   This is the routine implementing the BLAST query table function
 *   in Postgresql. It takes the necessary arguments and call 
 *   routines in blastsvc.c to retrieve hit table.
 *   A lot of Postgresql-related types and functions are used in 
 *   order to maintain memory context and manipulate tuples, please
 *   refer to Postgresql programmer manual for their definitions.
 *
 *   parameters:
 *     seq - sequence identifier or genomic sequence string (of text type)
 *     program - blast program (such as blastn, blastp, etc) (of text type)
 *     db  - blast database (such as nr, est, etc) (of text type) [optional]
 *
 *   return:
 *     returns a tuple each time this function is called by Postgresql.
 *
 *   note:
 *     this routine relies on the definition of "hit" type in Postgresql.
 *     this type must be defined in advance and its structure should be
 *     conformed to the following:
 *
 *     CREATE TYPE hit AS
 *     (
 *        query    loc,
 *        subject  loc,
 *        identity FLOAT,
 *        length   INTEGER,
 *        mismatches INTEGER,
 *        gap      INTEGER,
 *        evalue   FLOAT,
 *        bitscore FLOAT
 *      );
 *
 *   this also assumes that BLAST reports are always in the following
 *   form:
 *
 * BLASTN 2.2.10 [Oct-19-2004]
 * # Query:
 * # Database: nr
 * # Fields: Query id, Subject id, % identity, alignment length, mismatches, gap openings, q. start, q. end, s. start, s. end, e-value, bit score
 * 1_24847      gi|55818351|gb|AY688983.1|      100.00  60      0       0       1       60      301     360     5e-25    119
 * 1_24847      gi|55818471|gb|AY688964.1|      100.00  59      0       0       1       59      301     359     2e-24    117
 *
 *
 * ----------------
 */

/* ----------------
 *   local_blast_hsp_v
 *
 *   This routine is almost the same as local_blast_hsp except it accepts
 *   user-specified BLAST parameters (which is a table in Postgresql)
 *
 *   parameters:
 *     seqid - sequence identifier you assign to
 *     seq - sequence identifier or genomic sequence string (of text type)
 *     parameters  - table name in which BLAST parameters are stored.
 *
 *   return:
 *     returns a tuple each time this function is called by Postgresql.
 *
 *   note:
 *     this routine relies on the definition of "hsp" type in Postgresql.
 *     this type must be defined in advance and its structure should be
 *     conformed to the following:
 *
 *   CREATE TYPE hsp AS
 *   (
 *     queryloc    loc,             -- query location
 *     subjectloc  loc,             -- subject location
 *     identity FLOAT,              -- percent identity
 *     evalue FLOAT,                -- evalue
 *     bitscore FLOAT               -- bitscore
 *   );
 *
 *   this also depends on the assumption that the parameter table
 *   is defined to have two fields both of text type. The names of
 *   the fields do not matter. The first field serves as the parameter
 *   name and the second one serves as the parameter value.
 *
 * ----------------
 */

PG_FUNCTION_INFO_V1(local_blast_hsp_v);
Datum local_blast_hsp_v(PG_FUNCTION_ARGS)
{
  Datum result;   
  MemoryContext oldcontext;
  FuncCallContext *funcctx;
  TupleDesc tupdesc, spi_tupdesc;
  TupleTableSlot *slot;
  SPITupleTable *spi_tuptable;
  AttInMetadata *attinmeta;
  HeapTuple spi_tuple;      
  HspEntryPointer fctx;
  int call_cntr, max_calls, ret, proc, i, index;
  text *arg;
  FastaSequencePointer fasta;
  char *sequence, *paramtable, *data, buffer[500], **parameters, *seqid; 

  /* stuff done only on the first call of the function */
  if (SRF_IS_FIRSTCALL())
  {
     /* create a function context for cross-call persistence */
     funcctx = SRF_FIRSTCALL_INIT();
     /* switch to memory context appropriate for multiple function calls  */
     oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

     /* to ensure the input arguments are correct, we need to check if it's null */
     if (PG_ARGISNULL(0))  PG_RETURN_NULL();
     if (PG_ARGISNULL(1))  PG_RETURN_NULL();
     if (PG_ARGISNULL(2))  PG_RETURN_NULL();

     /* the sequence we are going to blast against */
     arg = PG_GETARG_TEXT_P(0);
     seqid = (char *) palloc(VARSIZE(arg)-VARHDRSZ);
     strcpy( seqid, GET_STR(arg) );

     arg = PG_GETARG_TEXT_P(1);
     sequence = (char*) palloc(VARSIZE(arg)-VARHDRSZ);
     strcpy( sequence, GET_STR(arg) );

     arg = PG_GETARG_TEXT_P(2);
     paramtable = (char*) palloc(VARSIZE(arg)-VARHDRSZ);
     strcpy( paramtable, GET_STR(arg) );

     /*  retrieve parameters from the designated table using SPI functions */
     if ((ret=SPI_connect()) <0)
        /* internal error */
        elog( ERROR, "local_blast_hsp_v: failed to connect SPI manager. SPI_connect returned %d",ret);

     sprintf( buffer, "SELECT * FROM %s", paramtable );
     ret = SPI_exec(buffer,0);
     proc = SPI_processed;

     spi_tuptable = NULL;
     if ((ret==SPI_OK_SELECT) && (proc>0))
     {
        spi_tuptable = SPI_tuptable;
        spi_tupdesc = spi_tuptable->tupdesc;

        /* the table must have only 2 attributes */
        if (spi_tupdesc->natts != 2)
            elog( ERROR, "Incorrect blast parameter table!" );
     }
     else
     {
        proc = 0;
        SPI_finish();
        elog( ERROR, "incorrect blast parameter table!" );

        // elog() will return immediately, but in order to avoid compiler warning, I add a return statement
        SRF_RETURN_DONE(funcctx);
     }

     /* we are about to retrieve each parameter pair from the blast parameter table */
     /* we allocate 2 extra parameter spaces for QUERY_FROM, QUERY_TO */
     parameters = (char**) palloc( (2*proc+1+4)*sizeof(char*) );
     for(i=0, index=0;i<proc;i++, index+=2)
     {
        /* get the next result */
        spi_tuple = spi_tuptable->vals[i];
        data = SPI_getvalue(spi_tuple, spi_tupdesc, 1);
        parameters[index] = data;

        data = SPI_getvalue(spi_tuple, spi_tupdesc, 2);
        parameters[index+1] = data;

     }

     /* SPI switches context on us, so reset it */
     MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
     parameters[2*proc] = NULL;

     /* call the query routine (defined in blastsvc.c) */
     if ( isSequenceFile(sequence) )
     {
        if ( (fasta=ReadFastaSequenceFromFile(sequence))!=NULL )
        {
           data = blast_with_seq( NULL, NULL, fasta->sequence, seqid, FT_PAIRWISE, parameters);
        }
        else
           data = NULL;
     }
     else
     {

        data = blast_with_seq(NULL, NULL, sequence, seqid,  FT_PAIRWISE, parameters);
     }

     pfree(sequence);
     pfree(paramtable);

     for(i=0;i<2*proc+4;i++)
       if ( parameters[i]!=NULL )  pfree(parameters[i]);

     pfree(parameters);
     SPI_finish();

     if (data==NULL)  PG_RETURN_NULL();

     /* total number of tuples to be returned */
     if ( data==NULL )  funcctx->max_calls = -1;
     else funcctx->max_calls = HspPairCount(data);

     /* build a tuple description for a BlastHitTable tuple */
     tupdesc = RelationNameGetTupleDesc(HSP_TYPE_NAME);

     /* allocate a slot for a tuple with this tupdesc */
     slot = TupleDescGetSlot(tupdesc);
     /* assign slot to function context */
     funcctx->slot = slot;
     /* generate attribute metadata needed later to produce tuples from raw c strings */
     attinmeta = TupleDescGetAttInMetadata(tupdesc);
     funcctx->attinmeta = attinmeta;

     /* allocate memory for user context */
     fctx = HspFindFirst( data, seqid );

     /* Use fctx to keep track of the blast result and unprocessed records   */
     funcctx->user_fctx = fctx;
     MemoryContextSwitchTo(oldcontext);
  }


  /* stuff done on every call of the function */
  funcctx = SRF_PERCALL_SETUP();
  call_cntr = funcctx->call_cntr;
  max_calls = funcctx->max_calls;
  slot = funcctx->slot;
  attinmeta = funcctx->attinmeta;
  fctx = funcctx->user_fctx;

  if (call_cntr < max_calls)      /* do when there is more left to send */
  {
     if ( call_cntr==0)
     {
         result = HspFormTuple( fctx, slot, attinmeta );
	 funcctx->user_fctx = fctx;
     }
     else
     {
         fctx = HspFindNext( fctx );
         result = HspFormTuple( fctx, slot, attinmeta );
         funcctx->user_fctx = fctx;
     }
     /* send the result */
     SRF_RETURN_NEXT(funcctx, result);
  }
  else
  {
    /* do when there is no more left */
    SRF_RETURN_DONE(funcctx);
  }
}

/* ----------------
 *   remote_blast_hsp_v
 *
 *   This routine is almost the same as remote_blast_hsp except it accepts
 *   user-specified BLAST parameters (which is a table in Postgresql)
 *
 *   parameters:
 *     seqid - sequence identifier you assign to
 *     seq - sequence identifier or genomic sequence string (of text type)
 *     parameters  - table name in which BLAST parameters are stored.
 *
 *   return:
 *     returns a tuple each time this function is called by Postgresql.
 *
 *   note:
 *     this routine relies on the definition of "hsp" type in Postgresql.
 *     this type must be defined in advance and its structure should be
 *     conformed to the following:
 *
 *   CREATE TYPE hsp AS
 *   (
 *     queryloc    loc,             -- query location
 *     subjectloc  loc,             -- subject location
 *     identity FLOAT,              -- percent identity
 *     evalue FLOAT,                -- evalue
 *     bitscore FLOAT               -- bitscore
 *   );
 *
 *   please note that the parameters accepted by remote BLAST server are 
 *   different than those accepted by local BLAST programs. 
 *
 *   this also depends on the assumption that the parameter table
 *   is defined to have two fields both of text type. The names of
 *   the fields do not matter. The first field serves as the parameter
 *   name and the second one serves as the parameter value.
 *
 * ----------------
 */


PG_FUNCTION_INFO_V1(remote_blast_hsp_v);
Datum remote_blast_hsp_v(PG_FUNCTION_ARGS)
{
  Datum result;    
  MemoryContext oldcontext; 
  FuncCallContext *funcctx; 
  TupleDesc tupdesc, spi_tupdesc;
  TupleTableSlot *slot;
  SPITupleTable *spi_tuptable;
  AttInMetadata *attinmeta;
  HeapTuple spi_tuple;  
  HspEntryPointer fctx;
  int call_cntr, max_calls, ret, proc, i, index;
  text *arg;
  char *sequence, *paramtable, *data, buffer[500], **parameters, *seqid;

  /* stuff done only on the first call of the function */
  if (SRF_IS_FIRSTCALL())
  {
     /* create a function context for cross-call persistence */
     funcctx = SRF_FIRSTCALL_INIT();
     /* switch to memory context appropriate for multiple function calls  */
     oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

     /* to ensure the input arguments are correct, we need to check if it's null */
     if (PG_ARGISNULL(0))  PG_RETURN_NULL();
     if (PG_ARGISNULL(1))  PG_RETURN_NULL();
     if (PG_ARGISNULL(2))  PG_RETURN_NULL();

     /* the sequence we are going to blast against */
     arg = PG_GETARG_TEXT_P(0);
     seqid = (char *) palloc(VARSIZE(arg)-VARHDRSZ);
     strcpy( seqid, GET_STR(arg) );

     arg = PG_GETARG_TEXT_P(1);
     sequence = (char*) palloc(VARSIZE(arg)-VARHDRSZ);
     strcpy( sequence, GET_STR(arg) );

     arg = PG_GETARG_TEXT_P(2);
     paramtable = (char*) palloc(VARSIZE(arg)-VARHDRSZ);
     strcpy( paramtable, GET_STR(arg) );

     /*  retrieve parameters from the designated table using SPI functions */
     if ((ret=SPI_connect()) <0)
        /* internal error */
        elog( ERROR, "local_blast_hit_v: failed to connect SPI manager. SPI_connect returned %d",ret);

     sprintf( buffer, "SELECT * FROM %s", paramtable );
     ret = SPI_exec(buffer,0);
     proc = SPI_processed;

     spi_tuptable = NULL;
     if ((ret==SPI_OK_SELECT) && (proc>0))
     {
        spi_tuptable = SPI_tuptable;
        spi_tupdesc = spi_tuptable->tupdesc;

        /* the table must have only 2 attributes */
        if (spi_tupdesc->natts != 2)
            elog( ERROR, "Incorrect blast parameter table!" );
     }
     else
     {
        proc = 0;
        SPI_finish();
        elog( ERROR, "incorrect blast parameter table!" );

        // elog() will return immediately, but in order to avoid compiler warning, I add a return statement
        SRF_RETURN_DONE(funcctx);
     }

     /* we are about to retrieve each parameter pair from the blast parameter table */
     /* we allocate 2 extra parameter spaces for QUERY_FROM, QUERY_TO */
     parameters = (char**) palloc( (2*proc+1+4)*sizeof(char*) );
     for(i=0, index=0;i<proc;i++, index+=2)
     {
        /* get the next result */
        spi_tuple = spi_tuptable->vals[i];
        data = SPI_getvalue(spi_tuple, spi_tupdesc, 1);
        parameters[index] = data;

        data = SPI_getvalue(spi_tuple, spi_tupdesc, 2);
        parameters[index+1] = data;
     }

     /* SPI switches context on us, so reset it */
     MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
     parameters[2*proc] = NULL;

     data = RemoteBlastService_V( sequence,  parameters, FT_PAIRWISE);

     pfree(sequence);
     pfree(paramtable);
     pfree(parameters);
     SPI_finish();

      /* total number of tuples to be returned */
     if ( data==NULL )
        funcctx->max_calls = -1;
     else
        funcctx->max_calls = HspPairCount(data);

      /* build a tuple description for a BlastHitTable tuple */
      tupdesc = RelationNameGetTupleDesc(HSP_TYPE_NAME);

      /* allocate a slot for a tuple with this tupdesc */
      slot = TupleDescGetSlot(tupdesc);
      /* assign slot to function context */
      funcctx->slot = slot;
      /* generate attribute metadata needed later to produce tuples from raw c strings */
      attinmeta = TupleDescGetAttInMetadata(tupdesc);
      funcctx->attinmeta = attinmeta;

      /* allocate memory for user context */
      fctx = HspFindFirst( data, seqid );

      /* Use fctx to keep track of the blast result and unprocessed records   */
      funcctx->user_fctx = fctx;
      MemoryContextSwitchTo(oldcontext);
  }
  /* stuff done on every call of the function */
  funcctx = SRF_PERCALL_SETUP();
  call_cntr = funcctx->call_cntr;
  max_calls = funcctx->max_calls;
  slot = funcctx->slot;
  attinmeta = funcctx->attinmeta;
  fctx = funcctx->user_fctx;

  if (call_cntr < max_calls)      /* do when there is more left to send */
  {
     if ( call_cntr==0)
     {
         result = HspFormTuple( fctx, slot, attinmeta );
         funcctx->user_fctx = fctx;
     }
     else
     {
         fctx = HspFindNext( fctx );
         result = HspFormTuple( fctx, slot, attinmeta );
         funcctx->user_fctx = fctx;
     }
     /* send the result */
     SRF_RETURN_NEXT(funcctx, result);
  }
  else
  {
    /* do when there is no more left */
    SRF_RETURN_DONE(funcctx);
  }
}

/* ----------------
 *   remote_blast_hit_v
 *
 *   This is the routine implementing the BLAST query table function
 *   in Postgresql. It takes the necessary arguments and call
 *   routines in blastsvc.c to retrieve hit table.
 *   A lot of Postgresql-related types and functions are used in
 *   order to maintain memory context and manipulate tuples, please
 *   refer to Postgresql programmer manual for their definitions.
 *
 *   parameters:
 *     seqid - sequence identifier you assign
 *     seq - sequence identifier or genomic sequence string (of text type)
 *     parametertable - blast parameter table 
 *
 *   return:
 *     returns a tuple each time this function is called by Postgresql.
 *
 *   note:
 *     this routine relies on the definition of "hit" type in Postgresql.
 *     this type must be defined in advance and its structure should be
 *     conformed to the following:
 *
 *     CREATE TYPE hit AS
 *     (
 *        query    loc,
 *        subject  loc,
 *        identity FLOAT,
 *        length   INTEGER,
 *        mismatches INTEGER,
 *        gap      INTEGER,
 *        evalue   FLOAT,
 *        bitscore FLOAT
 *      );
 *
 *   this also assumes that BLAST reports are always in the following
 *   form:
 *
 * BLASTN 2.2.10 [Oct-19-2004]
 * # Query:
 * # Database: nr
 * # Fields: Query id, Subject id, % identity, alignment length, mismatches, gap openings, q. start, q. end, s. start, s. end, e-value, bit score
 * 1_24847      gi|55818351|gb|AY688983.1|      100.00  60      0       0       1       60      301     360     5e-25    119
 * 1_24847      gi|55818471|gb|AY688964.1|      100.00  59      0       0       1       59      301     359     2e-24    117
 *
 *
 * ----------------
 */

PG_FUNCTION_INFO_V1(remote_blast_hit_v);
Datum remote_blast_hit_v(PG_FUNCTION_ARGS)
{
  Datum result;
  MemoryContext oldcontext;
  FuncCallContext *funcctx;
  TupleDesc tupdesc, spi_tupdesc;
  TupleTableSlot *slot;
  SPITupleTable *spi_tuptable;      
  AttInMetadata *attinmeta;
  HeapTuple spi_tuple;
  HitTableEntryPointer fctx;
  int call_cntr, max_calls, ret, proc, i, index;
  text *arg;
  //FastaSequencePointer fasta;
  char *sequence, *paramtable, *data, buffer[500], **parameters; 
  char *seqid;

  /* stuff done only on the first call of the function */
  if (SRF_IS_FIRSTCALL())
  {
     /* create a function context for cross-call persistence */
     funcctx = SRF_FIRSTCALL_INIT();
     /* switch to memory context appropriate for multiple function calls  */
     oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

     /* to ensure the input arguments are correct, we need to check if it's null */
     if (PG_ARGISNULL(0))  PG_RETURN_NULL();
     if (PG_ARGISNULL(1))  PG_RETURN_NULL();
     if (PG_ARGISNULL(2))  PG_RETURN_NULL();

     /* the sequence we are going to blast against */
     arg = PG_GETARG_TEXT_P(0);
     seqid = (char *) palloc(VARSIZE(arg)-VARHDRSZ);
     strcpy( seqid, GET_STR(arg) );

     arg = PG_GETARG_TEXT_P(1);
     sequence = (char*) palloc(VARSIZE(arg)-VARHDRSZ);
     strcpy( sequence, GET_STR(arg) );

     arg = PG_GETARG_TEXT_P(2);
     paramtable = (char*) palloc(VARSIZE(arg)-VARHDRSZ);
     strcpy( paramtable, GET_STR(arg) );

     /*  retrieve parameters from the designated table using SPI functions */
     if ((ret=SPI_connect()) <0)
        /* internal error */
        elog( ERROR, "local_blast_hit_v: failed to connect SPI manager. SPI_connect returned %d",ret);
     sprintf( buffer, "SELECT * FROM %s", paramtable );
     ret = SPI_exec(buffer,0);
     proc = SPI_processed;

     spi_tuptable = NULL;
     if ((ret==SPI_OK_SELECT) && (proc>0))
     {
        spi_tuptable = SPI_tuptable;
        spi_tupdesc = spi_tuptable->tupdesc;

        /* the table must have only 2 attributes */
        if (spi_tupdesc->natts != 2)
            elog( ERROR, "Incorrect blast parameter table!" );
     }
     else
     {
        proc = 0;
        SPI_finish();
        elog( ERROR, "incorrect blast parameter table!" );

        // elog() will return immediately, but in order to avoid compiler warning, I add a return statement
        SRF_RETURN_DONE(funcctx);
     }

     /* we are about to retrieve each parameter pair from the blast parameter table */
     /* we allocate 2 extra parameter spaces for QUERY_FROM, QUERY_TO */
     parameters = (char**) palloc( (2*proc+1+4)*sizeof(char*) );
     for(i=0, index=0;i<proc;i++, index+=2)
     {
        /* get the next result */
        spi_tuple = spi_tuptable->vals[i];
        data = SPI_getvalue(spi_tuple, spi_tupdesc, 1);
        parameters[index] = data;

        data = SPI_getvalue(spi_tuple, spi_tupdesc, 2);
        parameters[index+1] = data;

     }

     /* SPI switches context on us, so reset it */
     MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
     parameters[2*proc] = NULL;

     data = RemoteBlastService_V( sequence,  parameters, FT_TABULAR_COMMENT);

     pfree(sequence);
     pfree(paramtable);
     pfree(parameters);

     SPI_finish();
     if (data==NULL)  PG_RETURN_NULL();

      /* total number of tuples to be returned */
      funcctx->max_calls = HitTablePairCount(data);

      /* build a tuple description for a BlastHitTable tuple */
      tupdesc = RelationNameGetTupleDesc(HIT_TABLE_TYPE_NAME);

      /* allocate a slot for a tuple with this tupdesc */
      slot = TupleDescGetSlot(tupdesc);
      /* assign slot to function context */
      funcctx->slot = slot;
      /* generate attribute metadata needed later to produce tuples from raw c strings */
      attinmeta = TupleDescGetAttInMetadata(tupdesc);
      funcctx->attinmeta = attinmeta;

      /* allocate memory for user context */
      if (data)  fctx = HitTableFindFirst( data, seqid );

      /* Use fctx to keep track of the blast result and unprocessed records   */
      funcctx->user_fctx = fctx;
      MemoryContextSwitchTo(oldcontext);
  }

  /* stuff done on every call of the function */
  funcctx = SRF_PERCALL_SETUP();
  call_cntr = funcctx->call_cntr;
  max_calls = funcctx->max_calls;
  slot = funcctx->slot;
  attinmeta = funcctx->attinmeta;
  fctx = funcctx->user_fctx;

  if (call_cntr < max_calls)      /* do when there is more left to send */
  {
     if ( call_cntr==0)
         result = HitTableFormTuple( fctx, slot, attinmeta );
     else
     {
         fctx = HitTableFindNext( fctx );
         result = HitTableFormTuple( fctx, slot, attinmeta );
         funcctx->user_fctx = fctx;
     }
     /* send the result */
     SRF_RETURN_NEXT(funcctx, result);
  }
  else
  {
    /* do when there is no more left */
    SRF_RETURN_DONE(funcctx);
  }
}

/* ----------------
 *   local_blast_hit_v
 *
 *   This is the routine implementing the BLAST query table function
 *   in Postgresql. It takes the necessary arguments and call
 *   routines in blastsvc.c to retrieve hit table.
 *   A lot of Postgresql-related types and functions are used in
 *   order to maintain memory context and manipulate tuples, please
 *   refer to Postgresql programmer manual for their definitions.
 *
 *   parameters:
 *     seqid - sequence identifier you assign
 *     seq - sequence identifier or genomic sequence string (of text type)
 *     parameter table - blast parameter table
 *
 *   return:
 *     returns a tuple each time this function is called by Postgresql.
 *
 *   note:
 *     this routine relies on the definition of "hit" type in Postgresql.
 *     this type must be defined in advance and its structure should be
 *     conformed to the following:
 *
 *     CREATE TYPE hit AS
 *     (
 *        query    loc,
 *        subject  loc,
 *        identity FLOAT,
 *        length   INTEGER,
 *        mismatches INTEGER,
 *        gap      INTEGER,
 *        evalue   FLOAT,
 *        bitscore FLOAT
 *      );
 *
 *   this also assumes that BLAST reports are always in the following
 *   form:
 *
 * BLASTN 2.2.10 [Oct-19-2004]
 * # Query:
 * # Database: nr
 * # Fields: Query id, Subject id, % identity, alignment length, mismatches, gap openings, q. start, q. end, s. start, s. end, e-value, bit score
 * 1_24847      gi|55818351|gb|AY688983.1|      100.00  60      0       0       1       60      301     360     5e-25    119
 * 1_24847      gi|55818471|gb|AY688964.1|      100.00  59      0       0       1       59      301     359     2e-24    117
 *
 *
 * ----------------
 */


PG_FUNCTION_INFO_V1(local_blast_hit_v);
Datum local_blast_hit_v(PG_FUNCTION_ARGS)
{
  Datum result;
  MemoryContext oldcontext;
  FuncCallContext *funcctx;
  TupleDesc tupdesc, spi_tupdesc;
  TupleTableSlot *slot;
  SPITupleTable *spi_tuptable;
  AttInMetadata *attinmeta;
  HeapTuple spi_tuple;
  HitTableEntryPointer fctx = NULL;
  int call_cntr, max_calls, ret, proc, i, index;
  text *arg;
  FastaSequencePointer fasta;
  char *sequence, *paramtable, *data, buffer[500], **parameters, *seqid;

  /* stuff done only on the first call of the function */
  if (SRF_IS_FIRSTCALL())
  {
     /* create a function context for cross-call persistence */
     funcctx = SRF_FIRSTCALL_INIT();                                    
     /* switch to memory context appropriate for multiple function calls  */
     oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

     /* to ensure the input arguments are correct, we need to check if it's null */
     if (PG_ARGISNULL(0))  PG_RETURN_NULL();
     if (PG_ARGISNULL(1))  PG_RETURN_NULL();

     /* the sequence we are going to blast against */
     arg = PG_GETARG_TEXT_P(0);
     seqid = (char *) palloc(VARSIZE(arg)-VARHDRSZ);
     strcpy( seqid, GET_STR(arg) );

     arg = PG_GETARG_TEXT_P(1);
     sequence = (char*) palloc(VARSIZE(arg)-VARHDRSZ);
     strcpy( sequence, GET_STR(arg) );

     arg = PG_GETARG_TEXT_P(2);
     paramtable = (char*) palloc(VARSIZE(arg)-VARHDRSZ);
     strcpy( paramtable, GET_STR(arg) );

     /*  retrieve parameters from the designated table using SPI functions */
     if ((ret=SPI_connect()) <0)
        /* internal error */
        elog( ERROR, "local_blast_hit_v: failed to connect SPI manager. SPI_connect returned %d",ret);
     sprintf( buffer, "SELECT * FROM %s", paramtable );
     ret = SPI_exec(buffer,0);
     proc = SPI_processed;

     spi_tuptable = NULL;
     if ((ret==SPI_OK_SELECT) && (proc>0))
     {
        spi_tuptable = SPI_tuptable;
        spi_tupdesc = spi_tuptable->tupdesc;

        /* the table must have only 2 attributes */
        if (spi_tupdesc->natts != 2)
            elog( ERROR, "Incorrect blast parameter table!" );
     }
     else
     {
        proc = 0;
        SPI_finish();
        elog( ERROR, "incorrect blast parameter table!" );

        // elog() will return immediately, but in order to avoid compiler warning, I add a return statement
        SRF_RETURN_DONE(funcctx);
     }

     /* we are about to retrieve each parameter pair from the blast parameter table */
     /* we allocate 2 extra parameter spaces for QUERY_FROM, QUERY_TO */
     parameters = (char**) palloc( (2*proc+1+4)*sizeof(char*) );
     for(i=0, index=0;i<proc;i++, index+=2)
     {
        /* get the next result */
        spi_tuple = spi_tuptable->vals[i];
        data = SPI_getvalue(spi_tuple, spi_tupdesc, 1);
        parameters[index] = data;

        data = SPI_getvalue(spi_tuple, spi_tupdesc, 2);
        parameters[index+1] = data;

     }

     /* SPI switches context on us, so reset it */
     MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
     parameters[2*proc] = NULL;

     /* call the query routine (defined in blastsvc.c) */
     if ( isSequenceFile(sequence) )
     {
        if ( (fasta=ReadFastaSequenceFromFile(sequence))!=NULL )
        {
           data = blast_with_seq( NULL, NULL, fasta->sequence, seqid, FT_TABULAR_COMMENT, parameters);
        }
        else
           data = NULL;
     }
     else
     {

        data = blast_with_seq(NULL, NULL, sequence, seqid,  FT_TABULAR_COMMENT, parameters);
     }

     pfree(sequence);
     pfree(paramtable);

     for(i=0;i<2*proc+4;i++)
       if ( parameters[i]!=NULL )  pfree(parameters[i]);
     pfree(parameters);

     SPI_finish();

      /* total number of tuples to be returned */
      if ( data==NULL)   funcctx->max_calls = -1;
      else  funcctx->max_calls = HitTablePairCount(data);

      /* build a tuple description for a BlastHitTable tuple */
      tupdesc = RelationNameGetTupleDesc("hit");

      /* allocate a slot for a tuple with this tupdesc */
      slot = TupleDescGetSlot(tupdesc);
      /* assign slot to function context */
      funcctx->slot = slot;
      /* generate attribute metadata needed later to produce tuples from raw c strings */
      attinmeta = TupleDescGetAttInMetadata(tupdesc);
      funcctx->attinmeta = attinmeta;

      /* allocate memory for user context */
      if ( data!=NULL )
        fctx = HitTableFindFirst( data, seqid );

      /* Use fctx to keep track of the blast result and unprocessed records   */
      funcctx->user_fctx = fctx;
      MemoryContextSwitchTo(oldcontext);
  }

  /* stuff done on every call of the function */
  funcctx = SRF_PERCALL_SETUP();                                                                                               
  call_cntr = funcctx->call_cntr;
  max_calls = funcctx->max_calls;
  slot = funcctx->slot;
  attinmeta = funcctx->attinmeta;
  fctx = funcctx->user_fctx;

  if (max_calls==-1)  
    SRF_RETURN_DONE(funcctx);


  if (call_cntr < max_calls)      /* do when there is more left to send */
  {
     if ( call_cntr==0)
         result = HitTableFormTuple( fctx, slot, attinmeta );
     else
     {
         fctx = HitTableFindNext( fctx );
         result = HitTableFormTuple( fctx, slot, attinmeta );
         funcctx->user_fctx = fctx;
     }
     /* send the result */
     SRF_RETURN_NEXT(funcctx, result);
  }
  else
  {
    /* do when there is no more left */
    SRF_RETURN_DONE(funcctx);
  }
}




/*
*  The basic functionality here is, the sequence argument can be one of the followings:
*   (1) sequence file name (with .fasta or .fas file extension)
*   (2) sequence letters
*   (3) sequence identifier with range (ie. "identifier[lower..upper]" )
*  so the following section is for this purpose
*/

char* RemoteBlastService_V(char *sequence, char **parameters, int type)
{
   FastaSequencePointer ret;
   char *p, *data, **s;
   RANGE_int lower, upper;
   int i, j;

   data = NULL;      
   if ( isSequenceFile(sequence) )
   {
      /*
       *  now, we've detected the sequence argument is actually a file name,
       *  we are going to retrieve the sequence letters from the file.
       */
      ret = ReadFastaSequenceFromFile(sequence);
      if (ret!=NULL)
      {
         p = pstrdup(ret->sequence);
         FreeFastaSequence(ret);
         data = remote_blast_v(p,parameters,type);
         pfree(p);
      }
      else
      {
         /*
          *  exception here!!!
          *  we expect this to be a sequence file but could not retrieve
          *  the actual sequence letter for some reason
          */
          elog( ERROR, "unable to load sequence from fasta file %s", sequence );
      }
   }
   else
   {
      /*
       *  now, we've decided the sequence argument is not a file name,
       *  in other words, it is a sequence or sequence identifier with range
       */
      if ( (p=strchr(sequence,'['))!=NULL )
      {
         if ( sscanf(p, " [ %lld .. %lld ]", &lower, &upper) !=2 )
            ereport(ERROR,(errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),errmsg("invalid slice syntax")));

         i=0;
         s = parameters;
         while (*s++!=NULL)
         {
            i++;
         }         

        parameters = (char **) repalloc( parameters,(5+i)*sizeof(char *) );

         /* setup the slice parameters */
         parameters = (char**) palloc( 5 *sizeof(char*) );
         parameters[i+0] = (char*) palloc(11);
         strcpy(parameters[i+0], "QUERY_FROM");
         parameters[i+1] = (char*) palloc(10);
         sprintf( parameters[i+1], "%lld", lower );
         parameters[i+2] = (char*) palloc(9);
         strcpy(parameters[i+2], "QUERY_TO");
         parameters[i+3] = (char*) palloc(10);
         sprintf( parameters[i+3], "%lld", upper );
         parameters[i+4] = 0;

         *p = 0;

         data = remote_blast_v( sequence, parameters, type );
         for(j=0;j<i+4;j++)  pfree(parameters[j]);
         pfree(parameters);
      }
      else
      {
         // pure sequence letters here
         data = remote_blast_v(sequence, parameters,type);
      }
   }
   return data;
}


char* RemoteBlastService(char *sequence, char *program, char *db, int type)
{
   FastaSequencePointer ret;
   char *p, *data, **parameters;
   RANGE_int lower, upper;
   int i;


   data = NULL;
   if ( isSequenceFile(sequence) )
   {
      /*
       *  now, we've detected the sequence argument is actually a file name,
       *  we are going to retrieve the sequence letters from the file.
       */
      ret = ReadFastaSequenceFromFile(sequence);
      if (ret!=NULL)
      {
         p = pstrdup(ret->sequence);
         FreeFastaSequence(ret);
         data = remote_blast(p,program,db,type);
         pfree(p);
      }
      else
      {
         /*
          *  exception here!!!
          *  we expect this to be a sequence file but could not retrieve
          *  the actual sequence letter for some reason
          */
          elog( ERROR, "unable to load sequence from fasta file %s", sequence );
      }
   }
   else
   {
      /*
       *  now, we've decided the sequence argument is not a file name,
       *  in other words, it is a sequence or sequence identifier with range
       */
      if ( (p=strchr(sequence,'['))!=NULL )
      {
         if ( sscanf(p, " [ %lld .. %lld ]", &lower, &upper) !=2 )
            ereport(ERROR,(errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),errmsg("invalid slice syntax")));

         /* setup the slice parameters */
         parameters = (char**) palloc( 5 *sizeof(char*) );
         parameters[0] = (char*) palloc(11);
         strcpy(parameters[0], "QUERY_FROM");
         parameters[1] = (char*) palloc(10);
         sprintf( parameters[1], "%lld", lower );
         parameters[2] = (char*) palloc(9);
         strcpy(parameters[2], "QUERY_TO");
         parameters[3] = (char*) palloc(10);
         sprintf( parameters[3], "%lld", upper );
         parameters[4] = 0;

         *p = 0;

         data = remote_blast_v( sequence, parameters, type );
         for(i=0;i<4;i++)  pfree(parameters[i]);
         pfree(parameters);
      }
      else
      {
         // pure sequence letters here
         data = remote_blast(sequence,program,db,type);
      }
   }
   return data;
}


PG_FUNCTION_INFO_V1(remote_blast_hit);
Datum remote_blast_hit(PG_FUNCTION_ARGS)
{
   Datum result;
   MemoryContext oldcontext;
   FuncCallContext *funcctx;
   TupleTableSlot *slot;
   TupleDesc tupdesc;
   AttInMetadata *attinmeta;
   text *arg;
   HitTableEntryPointer fctx = NULL;
   int call_cntr, max_calls;
   char *sequence, *db, *program, *data, *seqid;


   /* stuff done only on the first call of the function */
   if (SRF_IS_FIRSTCALL())
   {
      /* create a function context for cross-call persistence */
      funcctx = SRF_FIRSTCALL_INIT();

      /* switch to memory context appropriate for multiple function  calls  */
      oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

      /* to ensure the input arguments are correct, we need to check if it's null */
      if (PG_ARGISNULL(0))  PG_RETURN_NULL();
      if (PG_ARGISNULL(1))  PG_RETURN_NULL();
      if (PG_ARGISNULL(2))  PG_RETURN_NULL();
      if (PG_ARGISNULL(3))  PG_RETURN_NULL();

      /* the sequence we are going to blast against */
      arg = PG_GETARG_TEXT_P(0);
      seqid = (char *) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( seqid, GET_STR(arg));

      arg = PG_GETARG_TEXT_P(1);
      sequence = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( sequence, GET_STR(arg) );

      arg = PG_GETARG_TEXT_P(2);
      db = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( db, GET_STR(arg) );

      arg = PG_GETARG_TEXT_P(3);
      program = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( program, GET_STR(arg) );

      /* retrieve the BLAST results here */
      data = RemoteBlastService(sequence,program,db, FT_TABULAR_COMMENT);

      pfree(sequence);
      pfree(program);
      pfree(db);


      /* 
       * now, we have to start to set up the POSTGRESQL multi-function call
       * context and parameters in order to return rows of BLAST results 
       */

      // total number of tuples to be returned 
      if ( data==NULL )
         funcctx->max_calls = -1;
      else
         funcctx->max_calls = HitTablePairCount(data);

      // build a tuple description for a BlastHitTable tuple 
      tupdesc = RelationNameGetTupleDesc(HIT_TABLE_TYPE_NAME);

      // allocate a slot for a tuple with this tupdesc 
      slot = TupleDescGetSlot(tupdesc);

      // assign slot to function context 
      funcctx->slot = slot;

      // generate attribute metadata needed later to produce tuples from raw c strings 
      attinmeta = TupleDescGetAttInMetadata(tupdesc);
      funcctx->attinmeta = attinmeta;

      // use fctx to keep track of the blast result and unprocessed records
      if (data )  fctx = HitTableFindFirst( data, seqid );
      funcctx->user_fctx = fctx;

      /* switch to memory context appropriate for multiple function  calls  */
      MemoryContextSwitchTo(oldcontext);
   }

   /* stuff done on every call of the function */
   funcctx = SRF_PERCALL_SETUP();
   call_cntr = funcctx->call_cntr;
   max_calls = funcctx->max_calls;
   slot = funcctx->slot;
   attinmeta = funcctx->attinmeta;
   fctx = funcctx->user_fctx;

   if (call_cntr < max_calls)    /* do when there is more left to send */
   {
      if ( call_cntr==0 )  result = HitTableFormTuple( fctx, slot, attinmeta );
      else
      {
         fctx = HitTableFindNext( fctx );
         result = HitTableFormTuple( fctx, slot, attinmeta );
         funcctx->user_fctx = fctx;
      }

      // send a row back to POSTGRESQL backend server
      SRF_RETURN_NEXT(funcctx, result);
   }
   else
   {
      // do when there is no more left 
      SRF_RETURN_DONE(funcctx);
   }
}

PG_FUNCTION_INFO_V1(remote_blast_hsp);
Datum remote_blast_hsp(PG_FUNCTION_ARGS)
{
   Datum result;
   MemoryContext oldcontext;
   FuncCallContext *funcctx;
   TupleTableSlot *slot;
   TupleDesc tupdesc;
   AttInMetadata *attinmeta;
   text *arg;
   HspEntryPointer fctx;
   int call_cntr, max_calls;
   char *sequence, *db, *program, *data, *seqid;

   /* stuff done only on the first call of the function */
   if (SRF_IS_FIRSTCALL())
   {
      /* create a function context for cross-call persistence */
      funcctx = SRF_FIRSTCALL_INIT();

      /* switch to memory context appropriate for multiple function  calls  */
      oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

      /* to ensure the input arguments are correct, we need to check if it's null */
      if (PG_ARGISNULL(0))  PG_RETURN_NULL();
      if (PG_ARGISNULL(1))  PG_RETURN_NULL();
      if (PG_ARGISNULL(2))  PG_RETURN_NULL();

      /* the sequence we are going to blast against */

      arg = PG_GETARG_TEXT_P(0);
      seqid = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( seqid, GET_STR(arg) );

      arg = PG_GETARG_TEXT_P(1);
      sequence = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( sequence, GET_STR(arg) );

      arg = PG_GETARG_TEXT_P(2);
      db = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( db, GET_STR(arg) );

      arg = PG_GETARG_TEXT_P(3);
      program = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( program, GET_STR(arg) );

      /* retrieve the BLAST results here */
      data = RemoteBlastService(sequence,program, db, FT_PAIRWISE);

      pfree(sequence);
      pfree(program);
      pfree(db);


      /*
       * now, we have to start to set up the POSTGRESQL multi-function call
       * context and parameters in order to return rows of BLAST results
       */

      // total number of tuples to be returned
      if (data==NULL) funcctx->max_calls = -1;
      else  funcctx->max_calls = HspPairCount(data);

      // build a tuple description for a BlastHitTable tuple
      tupdesc = RelationNameGetTupleDesc(HSP_TYPE_NAME);

      // allocate a slot for a tuple with this tupdesc
      slot = TupleDescGetSlot(tupdesc);

      // assign slot to function context
      funcctx->slot = slot;

      // generate attribute metadata needed later to produce tuples from raw c strings
      attinmeta = TupleDescGetAttInMetadata(tupdesc);
      funcctx->attinmeta = attinmeta;

      // use fctx to keep track of the blast result and unprocessed records
      fctx = HspFindFirst( data, seqid );
      funcctx->user_fctx = fctx;

      /* switch to memory context appropriate for multiple function  calls  */
      MemoryContextSwitchTo(oldcontext);
   }

   /* stuff done on every call of the function */
   funcctx = SRF_PERCALL_SETUP();
   call_cntr = funcctx->call_cntr;
   max_calls = funcctx->max_calls;
   slot = funcctx->slot;
   attinmeta = funcctx->attinmeta;
   fctx = funcctx->user_fctx;

   if (call_cntr < max_calls)    /* do when there is more left to send */
   {
      if ( call_cntr==0 )
      {
          result = HspFormTuple( fctx, slot, attinmeta );
          funcctx->user_fctx = fctx;
      }
      else
      {
         fctx = HspFindNext( fctx );
         result = HspFormTuple( fctx, slot, attinmeta );
         funcctx->user_fctx = fctx;
      }

      // send a row back to POSTGRESQL backend server
      SRF_RETURN_NEXT(funcctx, result);
   }
   else
   {
      // do when there is no more left
      SRF_RETURN_DONE(funcctx);
   }
}


PG_FUNCTION_INFO_V1(local_blast_hsp);
Datum local_blast_hsp(PG_FUNCTION_ARGS)
{
   Datum result;
   MemoryContext oldcontext;
   FuncCallContext *funcctx;
   TupleTableSlot *slot;
   TupleDesc tupdesc;
   AttInMetadata *attinmeta;
   text *arg;
   FastaSequencePointer fasta_sequence;
   HspEntryPointer fctx;
   int call_cntr, max_calls;
   char *sequence, *db, *program, *data, *seqid;

   /* stuff done only on the first call of the function */
   if (SRF_IS_FIRSTCALL())
   {
      /* create a function context for cross-call persistence */
      funcctx = SRF_FIRSTCALL_INIT();

      /* switch to memory context appropriate for multiple function  calls  */
      oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

      /* to ensure the input arguments are correct, we need to check if it's null */
      if (PG_ARGISNULL(0))  PG_RETURN_NULL();
      if (PG_ARGISNULL(1))  PG_RETURN_NULL();
      if (PG_ARGISNULL(2))  PG_RETURN_NULL();
      if (PG_ARGISNULL(3))  PG_RETURN_NULL();

      /* the sequence we are going to blast against */
      arg = PG_GETARG_TEXT_P(0);
      seqid = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( seqid, GET_STR(arg) );

      arg = PG_GETARG_TEXT_P(1);
      sequence = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( sequence, GET_STR(arg) );

      arg = PG_GETARG_TEXT_P(2);
      db = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( db, GET_STR(arg) );

      arg = PG_GETARG_TEXT_P(3);
      program = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( program, GET_STR(arg) );


      /* we have to make sure that we are able to access the specified database */
      if ( CheckFileExistence(db)==0 )
      {
          elog( ERROR, "database does not exist or not in the directory specified by BLASTDB" );
          // the following statements are kind of a moot point 
          // since elog(ERROR,...) will return immediately... I put this just in case
          pfree(sequence);
          pfree(program);
          pfree(db);
          PG_RETURN_NULL();
      }

      // added jun24
      if ( isSequenceFile(sequence) )
      {
        fasta_sequence = ReadFastaSequenceFromFile( sequence );
        if ( fasta_sequence==NULL )  elog( ERROR, "can not read fasta file %s", sequence );
        sequence = fasta_sequence->sequence;
      }


      /* retrieve the BLAST results here */
      data = blast_with_seq(db, program, sequence, seqid, FT_PAIRWISE, NULL);

      pfree(sequence);
      pfree(program);
      pfree(db);

      if (data==NULL)   PG_RETURN_NULL();

      /*
       * now, we have to start to set up the POSTGRESQL multi-function call
       * context and parameters in order to return rows of BLAST results
       */

      // total number of tuples to be returned
      if ( data==NULL)   funcctx->max_calls = -1;
      else  funcctx->max_calls = HspPairCount(data);

      // build a tuple description for a BlastHitTable tuple
      tupdesc = RelationNameGetTupleDesc(HSP_TYPE_NAME);

      // allocate a slot for a tuple with this tupdesc
      slot = TupleDescGetSlot(tupdesc);

      // assign slot to function context
      funcctx->slot = slot;

      // generate attribute metadata needed later to produce tuples from raw c strings
      attinmeta = TupleDescGetAttInMetadata(tupdesc);
      funcctx->attinmeta = attinmeta;

      // use fctx to keep track of the blast result and unprocessed records
      fctx = HspFindFirst( data, seqid );
      funcctx->user_fctx = fctx;

      /* switch to memory context appropriate for multiple function  calls  */
      MemoryContextSwitchTo(oldcontext);
   }

   /* stuff done on every call of the function */
   funcctx = SRF_PERCALL_SETUP();
   call_cntr = funcctx->call_cntr;
   max_calls = funcctx->max_calls;
   slot = funcctx->slot;
   attinmeta = funcctx->attinmeta;
   fctx = funcctx->user_fctx;

   if (call_cntr < max_calls)    /* do when there is more left to send */
   {
      if ( call_cntr==0 ) 
      {
          result = HspFormTuple( fctx, slot, attinmeta );
          funcctx->user_fctx = fctx;
      }
      else
      {
         fctx = HspFindNext( fctx );
         result = HspFormTuple( fctx, slot, attinmeta );
         funcctx->user_fctx = fctx;
      }

      // send a row back to POSTGRESQL backend server
      SRF_RETURN_NEXT(funcctx, result);
   }
   else
   {
      // do when there is no more left
      SRF_RETURN_DONE(funcctx);
   }
}


PG_FUNCTION_INFO_V1(local_blast_hit);
Datum local_blast_hit(PG_FUNCTION_ARGS)
{
   Datum result;
   MemoryContext oldcontext;
   FuncCallContext *funcctx;
   TupleTableSlot *slot;
   TupleDesc tupdesc;
   AttInMetadata *attinmeta;
   text *arg;
   FastaSequencePointer fasta_sequence;
   HitTableEntryPointer fctx = NULL;
   int call_cntr, max_calls;
   char *sequence, *db, *program, *data, *seqid;

   /* stuff done only on the first call of the function */
   if (SRF_IS_FIRSTCALL())
   {
      /* create a function context for cross-call persistence */
      funcctx = SRF_FIRSTCALL_INIT();

      /* switch to memory context appropriate for multiple function  calls  */
      oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

      /* to ensure the input arguments are correct, we need to check if it's null */
      if (PG_ARGISNULL(0))  PG_RETURN_NULL();
      if (PG_ARGISNULL(1))  PG_RETURN_NULL();
      if (PG_ARGISNULL(2))  PG_RETURN_NULL();
      if (PG_ARGISNULL(3))  PG_RETURN_NULL();

      /* the sequence we are going to blast against */
      arg = PG_GETARG_TEXT_P(0);
      seqid = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( seqid, GET_STR(arg) );

      arg = PG_GETARG_TEXT_P(1);
      sequence = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( sequence, GET_STR(arg) );

      arg = PG_GETARG_TEXT_P(2);
      db = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( db, GET_STR(arg) );

      arg = PG_GETARG_TEXT_P(3);
      program = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( program, GET_STR(arg) );

      /* we have to make sure that we are able to access the specified database */
      if ( CheckFileExistence(db)==0 )
      {
          elog( ERROR, "database does not exist or not in the directory specified by BLASTDB" );
          // the following statements are kind of a moot point
          // since elog(ERROR,...) will return immediately... I put this just in case
          pfree(sequence);
          pfree(program);
          pfree(db);
          PG_RETURN_NULL();
      }

      // added jun24
      if ( isSequenceFile(sequence) )
      {
         fasta_sequence = ReadFastaSequenceFromFile( sequence );
         if ( fasta_sequence==NULL )  elog( ERROR, "can not read fasta file %s", sequence );
         sequence = fasta_sequence->sequence;
      }


      /* retrieve the BLAST results here */
      data = blast_with_seq(db, program, sequence, seqid, FT_TABULAR_COMMENT, NULL);

      pfree(sequence);
      pfree(program);
      pfree(db);


      /*
       * now, we have to start to set up the POSTGRESQL multi-function call
       * context and parameters in order to return rows of BLAST results
       */

      // total number of tuples to be returned
      if ( data==NULL )
          funcctx->max_calls = -1;
      else
          funcctx->max_calls = HitTablePairCount(data);

      // build a tuple description for a BlastHitTable tuple
      tupdesc = RelationNameGetTupleDesc(HIT_TABLE_TYPE_NAME);

      // allocate a slot for a tuple with this tupdesc
      slot = TupleDescGetSlot(tupdesc);

      // assign slot to function context
      funcctx->slot = slot;

      // generate attribute metadata needed later to produce tuples from raw c strings
      attinmeta = TupleDescGetAttInMetadata(tupdesc);
      funcctx->attinmeta = attinmeta;

      // use fctx to keep track of the blast result and unprocessed records
      if (data )  fctx = HitTableFindFirst( data, seqid );
      funcctx->user_fctx = fctx;

      /* switch to memory context appropriate for multiple function  calls  */
      MemoryContextSwitchTo(oldcontext);
   }
   /* stuff done on every call of the function */
   funcctx = SRF_PERCALL_SETUP();
   call_cntr = funcctx->call_cntr;
   max_calls = funcctx->max_calls;
   slot = funcctx->slot;
   attinmeta = funcctx->attinmeta;
   fctx = funcctx->user_fctx;

   if (call_cntr < max_calls)    /* do when there is more left to send */
   {
      if ( call_cntr==0 )
      {
          result = HitTableFormTuple( fctx, slot, attinmeta );
      }
      else
      {
         fctx = HitTableFindNext( fctx );
         result = HitTableFormTuple( fctx, slot, attinmeta );
         funcctx->user_fctx = fctx;
      }

      // send a row back to POSTGRESQL backend server
      SRF_RETURN_NEXT(funcctx, result);
   }
   else
   {
      // do when there is no more left
      SRF_RETURN_DONE(funcctx);
   }
}

// end of newly-added

PG_FUNCTION_INFO_V1(fasta_dbinfo);
Datum fasta_dbinfo(PG_FUNCTION_ARGS)
{
   Datum retval;
   TupleDesc tupdesc;
   TupleTableSlot *slot;
   AttInMetadata *attinmeta;
   HeapTuple tuple;
   FastaDBInformation *result;
   char *dbname, **values, buffer[1024];
   text *arg;

   // argument $1 -> db path

   if (PG_ARGISNULL(0))   PG_RETURN_NULL();

   arg = PG_GETARG_TEXT_P(0);
   dbname = (char *) palloc(VARSIZE(arg)-VARHDRSZ);
   strcpy( dbname, GET_STR(arg) );

   result = FastaDBInfo( dbname );
   if ( result==NULL )  PG_RETURN_NULL();

   values = (char **) palloc(7*sizeof(char *));
   values[0] = result->dbname;
   sprintf( buffer, "%d", result->nsequences );
   values[1] = pstrdup(buffer);
   sprintf( buffer, "%d", result->nletters );
   values[2] = pstrdup(buffer);
   values[3] = result->filepath;
   values[4] = result->datetime;
   values[5] = result->version;
   sprintf( buffer, "%d", result->longest );
   values[6] = pstrdup(buffer);

   /* get the tuple description from relation name */
   tupdesc = RelationNameGetTupleDesc("fasta_db_info");
   slot = TupleDescGetSlot(tupdesc);
   attinmeta = TupleDescGetAttInMetadata(tupdesc);

   /* build tuple from strings according to the attribute metadata */
   tuple = BuildTupleFromCStrings(attinmeta, values);

   /* make tuple into datum */
   retval = TupleGetDatum(slot,tuple);

   /* free the allocated memory */
   //for (i=0;i<7;i++)  pfree(values[i]);
   //pfree( result );  /* we free FastaSequence structure, but not their content */

   return retval;
}



PG_FUNCTION_INFO_V1(fasta_sequence);
Datum fasta_sequence(PG_FUNCTION_ARGS)
{
   Datum retval;
   TupleDesc tupdesc;
   TupleTableSlot *slot;
   AttInMetadata *attinmeta;
   HeapTuple tuple;
   int start, end, i;   
   FastaSequencePointer result;
   char *dbname, *seqid, **values;
   text *arg;

   // argument $1 -> db path
   // argument $2 -> sequence id
   // argument $3 -> start (optional)
   // argument $4 -> end   (optional)

   if (PG_ARGISNULL(0)||PG_ARGISNULL(1))   PG_RETURN_NULL();

   arg = PG_GETARG_TEXT_P(0);
   dbname = (char *) palloc(VARSIZE(arg)-VARHDRSZ);
   strcpy( dbname, GET_STR(arg) );

   arg = PG_GETARG_TEXT_P(1);
   seqid = (char *) palloc(VARSIZE(arg)-VARHDRSZ);
   strcpy( seqid, GET_STR(arg) );

   if (PG_ARGISNULL(2))    start = 0;
   else                    start = PG_GETARG_INT32(2);
   if (PG_ARGISNULL(3))    end = 0;
   else                    end = PG_GETARG_INT32(3);

   result = ReadFastaSequenceFromDB( dbname, seqid, start, end );
   if ( result==NULL )  PG_RETURN_NULL();

   values = (char **) palloc(3*sizeof(char *));
   values[0] = result->identifier;
   values[1] = result->description;
   values[2] = result->sequence;

   /* get the tuple description from relation name */
   tupdesc = RelationNameGetTupleDesc("fasta_record");
   slot = TupleDescGetSlot(tupdesc);
   attinmeta = TupleDescGetAttInMetadata(tupdesc);

   /* build tuple from strings according to the attribute metadata */
   tuple = BuildTupleFromCStrings(attinmeta, values);

   /* make tuple into datum */
   retval = TupleGetDatum(slot,tuple);

   /* free the allocated memory */
   for (i=0;i<3;i++)  pfree(values[i]);
   pfree( result );  /* we free FastaSequence structure, but not their content */

   return retval;
}

PG_FUNCTION_INFO_V1(local_blast_sequence);
Datum local_blast_sequence(PG_FUNCTION_ARGS)
{
  Datum result;
  MemoryContext oldcontext;
  FuncCallContext *funcctx;
  TupleTableSlot *slot;
  TupleDesc tupdesc;
  AttInMetadata *attinmeta;
  text *arg;
  HspEntryPointer fctx;
  int call_cntr, max_calls;
  char *sequence, *db, *program, *data, *seqid;

  db = NULL;

  /* stuff done only on the first call of the function */
  if (SRF_IS_FIRSTCALL())
  {
     /* create a function context for cross-call persistence */
     funcctx = SRF_FIRSTCALL_INIT();

     /* switch to memory context appropriate for multiple function  calls  */
     oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

     /* to ensure the input arguments are correct, we need to check if it's null */
     if (PG_ARGISNULL(0))  PG_RETURN_NULL();
     if (PG_ARGISNULL(1))  PG_RETURN_NULL();
     if (PG_ARGISNULL(2))  PG_RETURN_NULL();
     if (PG_ARGISNULL(3))  PG_RETURN_NULL();

     /* the sequence we are going to blast against */
     arg = PG_GETARG_TEXT_P(0);
     seqid = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
     strcpy( seqid, GET_STR(arg) );

     arg = PG_GETARG_TEXT_P(1);
     sequence = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
     strcpy( sequence, GET_STR(arg) );

     arg = PG_GETARG_TEXT_P(2);
     db = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
     strcpy( db, GET_STR(arg) );

     arg = PG_GETARG_TEXT_P(3);
     program = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
     strcpy( program, GET_STR(arg) );

     if ( CheckFileExistence(db)==0 ) 
     {
        elog( ERROR, "database does not exist or not in the directory specified by BLASTDB" );
        pfree(sequence);
        pfree(program);
        pfree(db);
        PG_RETURN_NULL();
     }

     data = blast_with_seq(db, program, sequence, seqid, FT_PAIRWISE, NULL);

     pfree(sequence);
     pfree(program);
     if ( db!=NULL)  pfree(db);
     if (data==NULL)  PG_RETURN_NULL();

     /* total number of tuples to be returned */
     funcctx->max_calls = HspPairCount(data);

     /* build a tuple description for a BlastHitTable tuple */
     tupdesc = RelationNameGetTupleDesc(HSP_TYPE_NAME);

     /* allocate a slot for a tuple with this tupdesc */
     slot = TupleDescGetSlot(tupdesc);

     /* assign slot to function context */
     funcctx->slot = slot;

     /* generate attribute metadata needed later to produce tuples from raw c strings */
     attinmeta = TupleDescGetAttInMetadata(tupdesc);
     funcctx->attinmeta = attinmeta;

     /*
      * Use fctx to keep track of the blast result and unprocessed records
      */
     fctx = HspFindFirst( data, seqid );

     funcctx->user_fctx = fctx;
     MemoryContextSwitchTo(oldcontext);
  }

  /* stuff done on every call of the function */
  funcctx = SRF_PERCALL_SETUP();

  call_cntr = funcctx->call_cntr;
  max_calls = funcctx->max_calls;
  slot = funcctx->slot;
  attinmeta = funcctx->attinmeta;
  fctx = funcctx->user_fctx;

  if (call_cntr < max_calls)    /* do when there is more left to send */
  {
     if ( call_cntr==0 )
     {
        result = HspFormTuple( fctx, slot, attinmeta );
        funcctx->user_fctx = fctx;
     }
     else
     {
        fctx = HspFindNext( fctx );
        result = HspFormTuple( fctx, slot, attinmeta );
        funcctx->user_fctx = fctx;
     }
     /* send the result */
     SRF_RETURN_NEXT(funcctx, result);
  }
  else
  {
      /* do when there is no more left */
      SRF_RETURN_DONE(funcctx);
  }
}

PG_FUNCTION_INFO_V1(local_megablast);
Datum local_megablast(PG_FUNCTION_ARGS)
{
   Datum result;
   MemoryContext oldcontext;
   FuncCallContext *funcctx;
   TupleTableSlot *slot;
   TupleDesc tupdesc;
   AttInMetadata *attinmeta;
   text *arg;
   FastaSequencePointer fasta_sequence;
   MegaBlastEntryPointer fctx = NULL;
   int call_cntr, max_calls;
   char *sequence, *db,  *data, *seqid;


   fasta_sequence = NULL;

   /* stuff done only on the first call of the function */
   if (SRF_IS_FIRSTCALL())
   {
      /* create a function context for cross-call persistence */
      funcctx = SRF_FIRSTCALL_INIT();

      /* switch to memory context appropriate for multiple function  calls  */
      oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

      /* to ensure the input arguments are correct, we need to check if it's null */
      if (PG_ARGISNULL(0))  PG_RETURN_NULL();
      if (PG_ARGISNULL(1))  PG_RETURN_NULL();
      if (PG_ARGISNULL(2))  PG_RETURN_NULL();

      /* the sequence we are going to blast against */
      arg = PG_GETARG_TEXT_P(0);
      seqid = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( seqid, GET_STR(arg) );

      arg = PG_GETARG_TEXT_P(1);
      sequence = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( sequence, GET_STR(arg) );

      arg = PG_GETARG_TEXT_P(2);
      db = (char*) palloc(VARSIZE(arg)-VARHDRSZ+1);
      strcpy( db, GET_STR(arg) );


      /* we have to make sure that we are able to access the specified database */
      if ( CheckFileExistence(db)==0 )
      {
          elog( ERROR, "database does not exist or not in the directory specified by BLASTDB" );
          // the following statements are kind of a moot point
          // since elog(ERROR,...) will return immediately... I put this just in case
          pfree(sequence);
          pfree(db);
          PG_RETURN_NULL();
      }

      if ( isSequenceFile(sequence) )
      {
         fasta_sequence = ReadFastaSequenceFromFile( sequence );
         if ( fasta_sequence==NULL )  elog( ERROR, "can not read fasta file %s", sequence );
         sequence = fasta_sequence->sequence;
         if ( seqid[0]=='\0' ) 
            seqid = pstrdup( fasta_sequence->identifier );
      }


      /* retrieve the BLAST results here */
      data = blast_with_seq(db,"megablast",sequence,seqid, MEGABLAST, NULL);
      pfree(db);
      if ( fasta_sequence )  FreeFastaSequence(fasta_sequence);

      /*
       * now, we have to start to set up the POSTGRESQL multi-function call
       * context and parameters in order to return rows of BLAST results
       */

      // total number of tuples to be returned
      if ( data==NULL )
      {
          funcctx->max_calls = -1;
      }
      else
      {
          funcctx->max_calls = MegaBlastPairCount(data);
          if ( funcctx->max_calls==0 )  { funcctx->max_calls = -1;  data=NULL; }

      } 

      // build a tuple description for a Megablast tuple
      tupdesc = RelationNameGetTupleDesc(MEGABLAST_TYPE_NAME);

      // allocate a slot for a tuple with this tupdesc
      slot = TupleDescGetSlot(tupdesc);

      // assign slot to function context
      funcctx->slot = slot;

      // generate attribute metadata needed later to produce tuples from raw c strings
      attinmeta = TupleDescGetAttInMetadata(tupdesc);
      funcctx->attinmeta = attinmeta;

      // use fctx to keep track of the blast result and unprocessed records

      if ( data )  
      {
         fctx = MegaBlastFindFirst( data, seqid );
      }

      funcctx->user_fctx = fctx;

      /* switch to memory context appropriate for multiple function  calls  */
      MemoryContextSwitchTo(oldcontext);
   }
   /* stuff done on every call of the function */
   funcctx = SRF_PERCALL_SETUP();
   call_cntr = funcctx->call_cntr;
   max_calls = funcctx->max_calls;
   slot = funcctx->slot;
   attinmeta = funcctx->attinmeta;
   fctx = funcctx->user_fctx;

   if (call_cntr < max_calls)    /* do when there is more left to send */
   {
      if ( call_cntr==0 )
      {
          result = MegaBlastFormTuple( fctx, slot, attinmeta );
      }
      else
      {
         fctx = MegaBlastFindNext( fctx );
         result = MegaBlastFormTuple( fctx, slot, attinmeta );
         funcctx->user_fctx = fctx;
      }

      // send a row back to POSTGRESQL backend server
      SRF_RETURN_NEXT(funcctx, result);
   }
   else
   {
      // do when there is no more left
      SRF_RETURN_DONE(funcctx);
   }
}

